/* Qidfs_utils.h ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  26 November 2000
*  Apr 2001  Added Qan date routines
*  May 2001 Re-organised Qpeace
*  Added promote headers
*
*  Utility routines for generic used, including:
*
*  Setting strings, idfs lineages, etc.
*  Setting conversion tables, getting energy bin limits, ...
*  Opening and positioning idfs file
*  Manipulating time into format needed by and returned from idfs routines
*
*/


#ifndef _QIDFS_UTILS
#define _QIDFS_UTILS

#ifdef __cplusplus
extern "C"{
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include "ret_codes.h"
#include "user_defs.h"
#include "libVIDF.h"
#include "libbase_idfs.h"
/* trying to add promote stuff */
#include "libdb.h"    /* try again Jul 2002 - not in SJS's copy of IDFS */
#include "libCfg.h" 
#include "libPromote.h"
/* end of promote stuff */
#include "Qpeace_err_diag.h"  /* error codes, diagnostic print levels */
#include "PGSpectro.h"    /* for SafeLog10 */

#define MAX_LEVELS 93           /* this is upper limit; less ok */
#define MAX_TABLES_TO_APPLY 20  /* 10 should be adequate */
#define SC_NAME_LEN 30          /* big enough to hold longest sc name */
#define MAX_PSD_UNITS_STR_LEN 100
#define PEACE_SPIN_PERIOD  4.0  /*rough spin per. in secs so can backup 1spin*/

#define LINEAGE_MAX_STR 100
#define DEG2RAD  0.01745329251994329547437168
#define RAD2DEG 57.29577951308232286464772187


typedef struct IDFS_LINEAGE_S{
   char project[LINEAGE_MAX_STR];
   char mission[LINEAGE_MAX_STR];
   char experiment[LINEAGE_MAX_STR];
   char instrument[LINEAGE_MAX_STR];
   char virtual_inst[LINEAGE_MAX_STR];
   char exten[LINEAGE_MAX_STR];/* this and vsn aren't  part of the lineage*/
   SDDAS_USHORT vsn;      /* but it's useful to carry them around here */
} IDFS_LINEAGE;

typedef enum SC_E{
  CLUSTER1,
  CLUSTER2,
  CLUSTER3,
  CLUSTER4,
  C1234,
  TC1,
  TC2
} SC;

#define N_SC 7

typedef enum PSD_UNITS_E{
   PSD_UNITS_CNTS_ACCUM,
   PSD_UNITS_DIFF_ENERGY_FLUX,
   PSD_UNITS_S3_PER_KM6,
   UNKNOWN_PSD_UNITS
} PSD_UNITS;

#define N_PSD_UNITS 3   /* don't count UNKNOWN */

typedef enum BIN_UNITS_E{
   BIN_UNITS_EV,
   BIN_UNITS_KM_S
} BIN_UNITS;

#define N_BIN_UNITS 2

typedef enum INTERP_QUANTITY_E{
   DIFFERENTIAL,
   INTEGRAL
} INTERP_QUANTITY;

typedef enum PHI_ZERO_E{
   PEACE_TZERO,
   UNKNOWN
} PHI_ZERO_FRAME;

typedef enum IDFS_TIME_COMPARE_E{
  T2_LATER   = -1,
  SAME_TIME,
  T1_LATER
} IDFS_TIME_COMPARE;

typedef struct  IDFS_TIME_S{
   SDDAS_SHORT year;
   SDDAS_SHORT doy;
   SDDAS_LONG  sec;
   SDDAS_LONG  nano;
}  IDFS_TIME;



/* prototpyes */

/**********************************************************************
****  IDFS Lineage and Strings ****************************************
**********************************************************************/

char *SC_NAME(SC sc);

char *PSD_UNITS_STRING(PSD_UNITS psdu);

char *PSD_UNITS_SI_conv_STRING(PSD_UNITS psdu);

char *BIN_UNITS_STRING(BIN_UNITS bin_units);

char *BIN_UNITS_SI_conv_STRING(BIN_UNITS bin_units);

void print_idfs_lineage(IDFS_LINEAGE *lng);

void copy_lineage(IDFS_LINEAGE *lincpy, IDFS_LINEAGE *lin);


/**********************************************************************
****  IDFS Time  ******************************************************
**********************************************************************/

void to_idfs_time(int year, int doy, int hr, int min, int sec,
                                long nano, IDFS_TIME *idftp);

void sec_nano_to_msec_nano(long sec, long nanos, long *msec, long *nanoms);

void msec_nano_to_sec_nano(long msec, long nanoms, long *sec, long *nanos);

void to_hms_time(IDFS_TIME *idftp, int *hp, int *mp, int *sp, long *np);

void print_idfs_time(IDFS_TIME *idftp);

IDFS_TIME t1_minus_t2(IDFS_TIME *t1, IDFS_TIME *t2);

IDFS_TIME_COMPARE compare_idfs_time(IDFS_TIME *t1, IDFS_TIME *t2);

void QanYearDoyToDate( int year, int doy, int* month, int* day );

void QanDateToYearDoy(int year, int month, int day, int *doy);

double QEpochSecs( int year, int month, int day,
  int hour, int minute, int second, int msec );
  
double idfs_time_to_epoch_secs(IDFS_TIME *idft);

void epoch_secs_to_idfs_time(double epoch, IDFS_TIME *idftp);

void add_secs_to_idfs_time(IDFS_TIME *idftp, float secs2add);

void set_times_from_idf(IDFS_TIME *btime, IDFS_TIME *etime,
                        IDFS_TIME *sun_sen, struct idf_data *idfp);

/*********************************************************************
****** IDFS File and data structure utils ****************************
*********************************************************************/

SDDAS_SHORT init_idf_file(IDFS_TIME btime, IDFS_TIME etime, SDDAS_ULONG dkey,
     struct idf_data *idfp, SDDAS_USHORT vsn);

SDDAS_SHORT pos_idf_file_spin_start(SDDAS_ULONG dkey, struct idf_data *idfp, 
         IDFS_LINEAGE *lineage, IDFS_TIME *btime, IDFS_TIME *etime);

SDDAS_SHORT advance_idf_file_spin_start(SDDAS_ULONG dkey, struct idf_data *idfp, 
         IDFS_LINEAGE *lineage, IDFS_TIME *btime, IDFS_TIME *etime,
	 SDDAS_SHORT sensor);

void print_idf_data(struct idf_data *ip);  /* prototype for sjs diag routine */


SDDAS_SHORT  e_bin_edges(SDDAS_ULONG dkey,struct idf_data *idfp,SDDAS_USHORT vsn,
      float *bin_low,float *bin_high, BIN_UNITS bin_units, int n_lev_array_len);

SDDAS_SHORT set_psd_tables(IDFS_LINEAGE *lgp, PSD_UNITS psd_units, 
     SDDAS_CHAR *num_tbls, SDDAS_CHAR *tbls_to_apply, SDDAS_SHORT *tbl_oper);


INTERP_QUANTITY diff_or_integral(PSD_UNITS psu); 


/*********************************************************************
********* Miscellaneous Utils ****************************************
*********************************************************************/

SDDAS_SHORT invert_array(float *arr, int n_lev);

double near_nice_number(double x);

void scale_then_rotate_angles(float *orig, float *rotscale, int n,
                         float rotation, float scale);

float polar_area(float r1, float r2, float th1_deg, float th2_deg);

int find_subset_in_range(float xlo, float xhi,
          float *xloarray, float *xhiarray, int n_array,
	  int *i1, int *i2);

SDDAS_SHORT check_bins(float *lowbins, float *hibins, int nbins);
	  
void print_bins(float *low, float *high, int n);

SDDAS_SHORT set_equal_bins(float *xlobins, float *xhibins, int nbins, 
                           float xmin, float xmax);

SDDAS_SHORT set_equal_bins_linlog(float *xlobins, float *xhibins, int nbins, 
                           float xmin, float xmax, PGSpectro_linlog linlog);
			   
SDDAS_SHORT set_equal_bins_half_ends(float *xlobins, float *xhibins, int nbins, 
                           float xmin, float xmax);

float mod360(float x);

#ifdef __cplusplus
}
#endif

#endif  /* end ifndef _QIDFS_UTILS */
