 
/* Qsweep_spin.h ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*
*  May 2001 Re-organised Qpeace software and created this module from others
*  May 2002 Adding pitch angle capabilities
*           Split off Qsweep_export and Qsweep_spin_read modules, but all
*              are covered by this single header
*  June 2002 Changed PEACE_SWEEP angles to be FLOW directions
*/

#ifndef _QSWEEP_SPIN
#define _QSWEEP_SPIN

#ifdef __cplusplus
extern "C"{
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include "ret_codes.h"
#include "user_defs.h"
#include "libVIDF.h"
#include "libbase_idfs.h"

#include "Qidfs_utils.h"
#include "Qpeace_err_diag.h"

#include "qie.h"

#define QP_DEFAULT_MISSING 0.0
#define NSWEEPS_4_SWEEPS_IN_SPIN 4    
                      /* 4 sweeps gives 3 dt's, then remove max, min*/

#define N_MAX_IDFS_FILES 10		      
#define N_MAX_IDFS_START_NUDGES 5

typedef enum T_OR_F_E{
   QPFALSE = 0,
   QPTRUE
} T_OR_F;

typedef enum ARITH_OPER_E{
   NONE,
   ADD,
   SUBTRACT,
   MULTIPLY,
   DIVIDE
} ARITH_OPER;

typedef enum SWEEP_CONGRUENT_TEST_E{
   BIN_UNIT_MISMATCH = -2,
   PSD_UNIT_MISMATCH = -1,
   CONGRUENT         = 0,
   NLEV_MISMATCH     = 1,
   NSEN_MISMATCH     = 2,
   EBINS_MISMATCH    = 3,
   THBINS_MISMATCH   = 4,
   UNKNOWN_CONGRUENT_ERROR = 5
} SWEEP_CONGRUENT_TEST;

typedef struct PEACE_SWEEP_S{
/* This structure stores all(?) the info for a single sweep of PEACE data in
*  science units.
*
* ***************** ENERGY LEVEL ORDER ASCENDING ******************
*  Note that energy levels are assumed to be in ASCENDING order of energy,
*  i.e., inverted with respect to decaying energy sweeps
*
* ***************** ANGLES are FLOW DIRECTIONS *******************
*  Note also that all angles <=> directions are FLOW directions
*  Theta bins must be in ASCENDING order
*/
   IDFS_TIME btime;          /* start time of sweep                          */
   IDFS_TIME etime;          /* end time of sweep                            */
   int n_sen;                /* number of sensors <=> pol bins               */
   int n_lev;                /* energy levels - or samples in idfs-speak     */
   float phi_lo;             /* start az angle, in degrees                   */
   float phi_hi;             /* end az angle                                 */
   PHI_ZERO_FRAME phi_frame; /* at present only PEACE_TZERO                  */
   float *thlo;              /* pntr to array of n_sen val for pol bin start */
   float *thhi;              /* ditto for pol bin end                        */
   float *elo;               /* energy bin start                             */
   float *ehi;               /* energy bin end                               */
   BIN_UNITS bin_units;      /* ev, km/s, etc. from enum                     */
   float *psd;               /* pntr to 1st element in array z[n_sen][n_lev] */
                             /* so psd[jth][ilev] = psd[n_lev*jth+ilev]      */
   PSD_UNITS psd_units;      /* units for psd array                          */
   float missing_value;      /* value to be treated as missing               */
   float *pa;                /* pointer to array of [n_sen][n_lev] pitch angs*/
   IDFS_LINEAGE *lineage;    /* pointer to lineage structure of proj, miss,..*/
} PEACE_SWEEP;

typedef struct PEACE_SPIN3D_S{
/*
* This structure stores a full spin of PEACE 3D data in the form of individual
* sweeps. Most of the info is held within the sweep structures themselves,
* so this is quite simple. In particular, the azimuths are in the sweep data
*/
   IDFS_TIME btime;          /* start time of spin                */
   IDFS_TIME etime;          /* end time of spin                  */
   float tspin;              /* spin period in seconds            */
   int n_swp;                /* number of sweeps in the spin      */
   PEACE_SWEEP *sweep;       /* pointer to n_swp sweep structures */
} PEACE_SPIN3D;

typedef struct PEACE_1D_SWEEP_CUT_S{
/* Hold 1-D radial cut from a sweep simplified to centre times, energies,....*/
   IDFS_TIME starttime;      /* start time of sweep                          */
   IDFS_TIME endtime;        /* end time of sweep                          */
   int n_lev;                /* energy levels - or samples in idfs-speak     */
   float phi_cen;            /* centre az angle, in degrees                  */
   float thcen;              /* centre of polar bin in degrees               */
   float *ecen;              /* energy bin centres                           */
   BIN_UNITS bin_units;      /* ev, km/s, etc. from enum                     */
   float *psd1D;             /* pntr to 1st element in array z[n_lev]        */
   PSD_UNITS psd_units;      /* units for psd array                          */
   IDFS_LINEAGE *lineage;    /* pointer to lineage structure of proj, miss,..*/
} PEACE_1D_SWEEP_CUT;

typedef struct PEACE_PITCH_ANGLE_OPTS_S{
/* hold a few options for idfs pitch angle calcs */
   T_OR_F do_pa;             /* QPTRUE to do it; QPFALSE otherwise */
   int    n_pabins;          /* How many pitch angle bins          */
   T_OR_F mu_bins;           /* QPTRUE for equal mu=cos(th) bins   */
} PEACE_PITCH_ANGLE_OPTS;
   
/* Sweep data */

char * qpeace_version_str(void);

SDDAS_SHORT malloc_sweep_arrays(PEACE_SWEEP *swpp, T_OR_F do_pa);

void nullify_sweep_arrays(PEACE_SWEEP *swpp);       

PEACE_SWEEP  *create_sweep_full_copy(PEACE_SWEEP *swp);

void free_sweep_arrays(PEACE_SWEEP *swp);

void free_n_sweeps(int n, PEACE_SWEEP **swppp);

void print_sweep_structure(PEACE_SWEEP *swpp);

/* Sweep manipulation */

SDDAS_SHORT CombineSweeps(
        PEACE_SWEEP *swp_new, /* target; could be part full */
        float *weight,       /* 2-D array; same size as new*/
	PEACE_SWEEP *swp_old, /* old sweep */
	INTERP_QUANTITY interpq  /* differential or integ (cnts)*/
	);
	
SDDAS_SHORT fill_sweep_bin_area(float *A, PEACE_SWEEP *swpp);

SDDAS_SHORT divide_by_area(float *A, PEACE_SWEEP *swpp);

SDDAS_SHORT multiply_by_area(float *A, PEACE_SWEEP *swpp);

int test_sweeps_congruent(PEACE_SWEEP *swp1, PEACE_SWEEP *swp2);

PEACE_SWEEP *sweep_arithmetic(PEACE_SWEEP *swp_pri, PEACE_SWEEP *swp_sec,
                              ARITH_OPER oper);

PEACE_SWEEP *rebin_sweep(PEACE_SWEEP *bin_holder, PEACE_SWEEP *swp);

SDDAS_SHORT rebin_n_sweeps(PEACE_SWEEP *bin_holder, int nswps, 
     PEACE_SWEEP **swpsinpp, int *nrebinned, PEACE_SWEEP ***swpsoutppp);
     
void print_weight(float *weight, int n_th, int n_lev);

void sweeps_autoscaler(PEACE_SWEEP **swppp, int nswp, 
                      float *psdmin, float *psdmax);
    
PEACE_1D_SWEEP_CUT *sweep_1D_cut(float theta, PEACE_SWEEP *swp);

void free_1D_cut_arrays (PEACE_1D_SWEEP_CUT *swpcut);

void print_1D_cut(PEACE_1D_SWEEP_CUT *swpcut);

SDDAS_SHORT rebin_sweep_to_pa(PEACE_SWEEP **swpp, int nswp,
          PEACE_PITCH_ANGLE_OPTS *paopts, PEACE_SWEEP **paswpp);

SDDAS_SHORT Rebin_n_sweeps_to_pa(PEACE_SWEEP **swppp, int nswps, 
    int nswpgrp, PEACE_PITCH_ANGLE_OPTS *paopts, 
    PEACE_SWEEP ***paswppp_p, int *npaswp);

void set_pa_opts_defaults(PEACE_PITCH_ANGLE_OPTS *paopts);

int inbin(float value, float *lobins, float *hibins, int nbins);

/* Spin data */

SDDAS_SHORT malloc_spin_sweep(PEACE_SPIN3D *spinp, int n_sen, int n_lev,
                       T_OR_F do_pa);

SDDAS_SHORT free_spin_data (PEACE_SPIN3D *spinp);

SDDAS_SHORT free_n_spins(PEACE_SPIN3D **spinpp, int nspins);

PEACE_SPIN3D *copy_spin (PEACE_SPIN3D *spin);

void print_peace_spin_info(PEACE_SPIN3D *spinp);

int find_az_sweep ( PEACE_SPIN3D *spinp, float az );

SDDAS_SHORT Rebin_n_spins_to_pa(PEACE_SPIN3D **spinpp, int nspins, 
     PEACE_PITCH_ANGLE_OPTS *paopts, PEACE_SWEEP ***paswppp_p);


/* Qsweep_spin_read routines */

SDDAS_SHORT fill_sweep(PEACE_SWEEP *swpp , IDFS_LINEAGE *lngp,
              struct idf_data *idfp, SDDAS_ULONG dkey, 
	      T_OR_F have_bins, T_OR_F do_pa);

SDDAS_SHORT fill_single_sen_sweep(PEACE_SWEEP *swpp, int isen,
               SDDAS_ULONG dkey, struct idf_data *idfp, 
	       T_OR_F do_pa);

SDDAS_SHORT GetPeaceSpin(IDFS_LINEAGE *lineage, IDFS_TIME *btime,
   BIN_UNITS bin_units, PSD_UNITS psd_units, 
   T_OR_F do_pa, PEACE_SPIN3D *spinp);

SDDAS_SHORT GetNPeaceSpins(
   IDFS_LINEAGE *lineage, 
   IDFS_TIME *btime,
   IDFS_TIME *etime,
   BIN_UNITS bin_units, 
   PSD_UNITS psd_units, 
   T_OR_F do_pa,
   int *nspins, 
   PEACE_SPIN3D ***spinppp);

SDDAS_SHORT GetN1PeaceSpins(
   IDFS_LINEAGE *lineage, 
   IDFS_TIME *btime,
   IDFS_TIME *etime,
   BIN_UNITS bin_units, 
   PSD_UNITS psd_units, 
   T_OR_F do_pa,
   int *nspins, 
   PEACE_SPIN3D ***spinppp);


PEACE_SWEEP *PeaceGetAzSlice(
       IDFS_LINEAGE *lineage, IDFS_TIME *btime, float az,
       BIN_UNITS bin_units,
       PSD_UNITS psd_units,
       PEACE_PITCH_ANGLE_OPTS *paopts);
       
SDDAS_SHORT GetN1PeaceSweeps(
        IDFS_LINEAGE *lngp, 
	IDFS_TIME *btimep,
        IDFS_TIME *etimep,
	BIN_UNITS bin_units, 
	PSD_UNITS psd_units,
	T_OR_F do_pa,
	int *nswps, 
	PEACE_SWEEP ***swpppp);

SDDAS_SHORT GetNPeaceSweeps(
        IDFS_LINEAGE *lngp, 
	IDFS_TIME *btimep,
        IDFS_TIME *etimep,
	BIN_UNITS bin_units, 
	PSD_UNITS psd_units,
	T_OR_F do_pa,
	int *nswps, 
	PEACE_SWEEP ***swpppp);

				  
SDDAS_SHORT init_spin3d(PEACE_SPIN3D *spin_dp,
                        IDFS_LINEAGE  *lngp,
                        struct idf_data *idfp,
                        SDDAS_ULONG dkey, int *n_sens, int *n_lev,
			T_OR_F do_pa);

float sweep_duration_calculator(IDFS_TIME *swpstart, int n);

SDDAS_SHORT fill_spin3d (PEACE_SPIN3D *spinp,
                        IDFS_LINEAGE  *lngp,
                        struct idf_data *idfp,
                        SDDAS_ULONG dkey,
                        BIN_UNITS bin_units, 
			PSD_UNITS psd_units,
                        PHI_ZERO_FRAME phi_frame,
			T_OR_F do_pa );


/* Qsweep_export routines */

SDDAS_SHORT nsweeps2QiCDFContents(PEACE_SWEEP **swppp, int nswps,
                                  QiCDFContents **QiCDFp);

SDDAS_SHORT nswps2QiCDF_g_atts(QiCDFContents *QiCDF, PEACE_SWEEP **swppp);

SDDAS_SHORT nswps2QiCDF_default_attr(QiVarAttribute *attp, int iatt);

SDDAS_SHORT nswps2QiCDF_time_vars(QiCDFVariable *epo, QiCDFVariable *epo_end,
                                  PEACE_SWEEP **swppp, int nswps);

SDDAS_SHORT nswps2QiCDF_psd_var(QiCDFVariable *psdv, 
                                  PEACE_SWEEP **swppp, int nswps);

SDDAS_SHORT nswps2QiCDF_phi_vars(QiCDFVariable *phiv, QiCDFVariable *phiv_end, 
                                  PEACE_SWEEP **swppp, int nswps);

SDDAS_SHORT nswps2QiCDF_theta_vars(QiCDFVariable *thv, QiCDFVariable *thv_end,
                                  PEACE_SWEEP **swppp, int nswps);

SDDAS_SHORT nswps2QiCDF_lev_vars(QiCDFVariable *levv, QiCDFVariable *levv_end, 
                                  PEACE_SWEEP **swppp, int nswps);


#ifdef __cplusplus
}
#endif

#endif  /* end ifndef _QSWEEP_SPIN */
