/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#include "qie.h"
#include "qie_local.h"


long QiWriteCSDSgenCDF (QiCDFContents *QiSCDF, QiOptions *QiSOpt)
{

/* CDF   variables */

CDFid id;               /* ID of CDF for subsequent operations */
CDFstatus status;
char * full_name;
long check=QMW_OK;
                                      

  /* test input data structure to ensure it useable */
  
  /* ensure file name and path exist */
  QiEnsureFileName(QiSCDF, QiSOpt);

  /* remove any variables that have no data or NULL ptrs */
   if (QiMakeSafe (QiSCDF) == NULL) return BAD_STRUCT;

  check = QiTestStructure(QiSCDF);
  if (check != QMW_OK) return BAD_STRUCT;

  /* set up QSAS object specific options */
  QiSetWDependencies(QiSCDF, QiSOpt);
  
   full_name =  
     (char *) malloc(strlen(QiSCDF->io_f_path)+strlen(QiSCDF->io_f_name)+6);
      /* allow extra space for .cdf later */
   strcpy(full_name, QiSCDF->io_f_path);
   strcat(full_name, QiSCDF->io_f_name);

  /* create .cdf file */
 
  check = QiCreateCDF(full_name, QiSOpt);
  if(check != QMW_OK) return check;
  
  /* open cdf file */
  
   status = CDFlib(OPEN_, CDF_, full_name, &id, NULL_);
     if (status != CDF_OK) return CDF_OPEN_ERR;
  
  /* write structure to file */
  
  check = QiWriteContents(QiSCDF, QiSOpt);
    if (check != QMW_OK){
       /* delete badly formed CDF file and return warning */
       CDFlib(DELETE_, CDF_, NULL_);
       return check;
    }
     
   /* Close CDF and exit cleanly */
 
   CDFlib(CLOSE_, CDF_, NULL_);
   
   free(full_name);

   return QMW_OK;
   
} /* end  QiWriteCSDSgenCDF  */


 
/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiTestStructure
*
* PURPOSE: Test structure is adequate to write CSDS cdf file.
*
* DESCRIPTION: Tests that all variables have same number of records.
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            Otherwise   Fatal error detected.
*
* FUNCTIONS_USED:
*           
*
* ALGORITHM: For each variable in vardata array compare number of records
*              with that specified in top level of file object.
*        
***********************************************************************/

long QiTestStructure(QiCDFContents *QiSCDF)
{
   long check=QMW_OK;
   long n;
   
      /* test number of records is same for each variable */
      /* This is essential to match data and time in records */
      
      for (n=0; n < QiSCDF->n_vars; n++){
        if( QiSCDF->vardata[n]->rec_vary == VARY){
           if( QiSCDF->vardata[n]->max_rec_num != QiSCDF->n_recs-1){
             check = SPARSE_RECS;
           }
        }
      }
   
      /* other tests may be added later */
   
   return check;
   
 }/* end QiTestStructure */
 
/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiWriteContents
*
* PURPOSE: Write contents of structure to new cdf file.
*
* DESCRIPTION: Writes all Global attributes together with data for each variable
*              in vardata array and the variable attributes associated. 
*              Variable attributes are not written for variables that do not appear in
*              vardata array, so metadata object need not be modified for subsetting.
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            Otherwise   Fatal error detected.
*
* FUNCTIONS_USED: QiWriteGlobals
*                 QiWriteVariable
*
* ALGORITHM:  call QiWriteGlobals to write out all global attributes present
*               in metadata object.
*             for each variable in vardata array call QiWriteVariable to output
*             data and variable attributes.
*       
***********************************************************************/

 long QiWriteContents (QiCDFContents *QiSCDF,
                       QiOptions *QiSOpt)    
 {
   long check = QMW_OK;
   long n;
   
   /* write Global Attributes */
   
   check = QiWriteGlobals (QiSCDF);
   if(check != QMW_OK) {
     /* issue warning and continue */
     QiDisplayMessage(QiErrStr(check), QiSOpt);
   }
   
   /* Write Variables and their metadata */
   
   for(n=0; n< QiSCDF->n_vars; n++){
     check = QiWriteVariable (QiSCDF, n);
     if(check != QMW_OK) return check;
   }
 
   return check;

 } /* end QiWriteContents */
  
/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiWriteGlobals
*
* PURPOSE: Write Global Attributes to new cdf file.
*
* DESCRIPTION: Writes all global attribute entries from metadata object into cdf.
*              Existing entries from the skeleton are overwritten if present in
*              metadata object.
*
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            Otherwise   Fatal error detected.
*
* FUNCTIONS_USED: CDFlib
*           
*
* ALGORITHM:  For each global attribute entry in metadata object:
*               If entry present.
*                 Try to select attribute of that name
*                 if unsuccessful create attribute of that name.
*                 For each entry in turn:
*                    select entry number and write entry.
*             Append software version number of this software to Software_version
*
***********************************************************************/

long QiWriteGlobals ( QiCDFContents *QiSCDF)
{
 
 long check=QMW_OK;
 long n, n_entry; 
 long attr_num;
 CDFstatus status;
 
 /* loop through entries in Global attributes in structure */
 
 for (n=0; n < QiSCDF->num_g_attrs; n++){
  
    /* select or create attribute */
 
    status = CDFlib(SELECT_, ATTR_NAME_, QiSCDF->g_attr[n]->name, 
                    NULL_);
    if(status != CDF_OK){
      status = CDFlib(CREATE_, ATTR_, QiSCDF->g_attr[n]->name, 
                                      GLOBAL_SCOPE, &attr_num, 
                      NULL_);
      if(status != CDF_OK){
        check = NO_GLOBAL_ATTR;
        continue;               /* skip this as we can't select or create it */
      }
    }
    
    /* write each entry in turn */
    
    for (n_entry=0; n_entry < QiSCDF->g_attr[n]->num_entries; 
         n_entry++){
   
      /* test if entry present in structure, if not go on to next */
    
      if(QiSCDF->g_attr[n]->entry[n_entry].exists == 0) continue;

      status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);
        if(status != CDF_OK){
          QiStatusHandler(status,"on next global entry");
          check = FAIL_TO_SELECT_ENTRY;
          continue;
        }
      
      status = CDFlib(PUT_, gENTRY_DATA_, 
                      QiSCDF->g_attr[n]->entry[n_entry].data_type,
                      QiSCDF->g_attr[n]->entry[n_entry].num_elems, 
                      QiSCDF->g_attr[n]->entry[n_entry].data, 
                      NULL_);
      if (status != CDF_OK) {
         check = FAIL_G_ATTR_PUT;
      }  
      
    } /* end for over entries */
    
 } /* end for over G attributes */
 
    /* Append the software version of this module */
    
    status = CDFlib(SELECT_, ATTR_NAME_, "Software_version", NULL_);
    if(status != CDF_OK){
      status = CDFlib(CREATE_, ATTR_, "Software_version", 
                                      GLOBAL_SCOPE, &attr_num, 
                      NULL_);
      if(status != CDF_OK){
        return NO_GLOBAL_ATTR;
      }
    }
    
    status = CDFlib(GET_, ATTR_MAXgENTRY_, &n_entry, NULL_);
    if (status != CDF_OK){
      return FAIL_GET_MAX_ENTRY;
    }
    n_entry++;
    status = CDFlib(SELECT_, gENTRY_, n_entry, NULL_);

    status = CDFlib(PUT_, gENTRY_DATA_, CDF_CHAR,
                                        strlen(QIE_VERSION),
                                        QIE_VERSION, 
                    NULL_); 
    if(status != CDF_OK) return FAIL_VERSION_WRITE;
    
    return check;
     
} /* end QiWriteGlobals */
 
 
/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiWriteVariables
*
* PURPOSE: Write Variables and Variable Attributes to new cdf file.
*
* DESCRIPTION: Create variable of appropriate name. Write data. Find matching
*              metadata variable attribute  (same name and data type) and write.
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            Otherwise   Fatal error detected.
*
* FUNCTIONS_USED:  CDFlib
*                  QiTestMetadata
*                  QiWriteVarAttrs
*
* ALGORITHM: call CDFlib to extend records in cdf to take data.
*            call CDFlib to select intervals and sizes for variable.
*            call CDFlib to write all data via hyperput
*            call QiTestMetadata to locate matching attributes in metadata
*            call QiWriteVarAttrs to write variable attributes
*
***********************************************************************/

long QiWriteVariable ( QiCDFContents *QiSCDF, 
                       long n)
{
  long check;
  CDFstatus status;
  long var_num;
  long j;
  long dim_indices[CDF_MAX_DIMS];
  long dim_intervals[CDF_MAX_DIMS];
  
 /* try to select variable as label and Half_interval exist */

  status = CDFlib(SELECT_, zVAR_NAME_, QiSCDF->vardata[n]->name, NULL_);

  if (status != CDF_OK) {

    /* create variable in CDF file */
  
    status = CDFlib(CREATE_, zVAR_, QiSCDF->vardata[n]->name,
                                    QiSCDF->vardata[n]->data_type,
                                    QiSCDF->vardata[n]->num_elems,
                                    QiSCDF->vardata[n]->num_dims,
                                    QiSCDF->vardata[n]->dim_sizes,
                                    QiSCDF->vardata[n]->rec_vary,
                                    QiSCDF->vardata[n]->dim_varys,
                                    &var_num,
                    NULL_);
  
    QiStatusHandler(status, " on create");
  }

  /* allocate space in cdf */
  
  status = CDFlib(PUT_, zVAR_EXTENDRECS_, QiSCDF->vardata[n]->max_rec_num, 
                  NULL_);
  
  /* select parameters for hyper write */
  
  for (j=0; j < QiSCDF->vardata[n]->num_dims; j++){        
    dim_indices[j] = 0;
    dim_intervals[j] = 1;
  }
  
    /* write out data */
  status = CDFlib (SELECT_, zVAR_RECNUMBER_, 0,
                            zVAR_RECCOUNT_, QiSCDF->vardata[n]->max_rec_num + 1,
                            zVAR_RECINTERVAL_, 1,
                            zVAR_DIMCOUNTS_, QiSCDF->vardata[n]->dim_sizes, 
                            zVAR_DIMINDICES_, dim_indices,
                            zVAR_DIMINTERVALS_, dim_intervals,
                   NULL_);
  QiStatusHandler(status, " on select");

  if (status != CDF_OK) return CDF_VAR_SELECT_ERR;
    
  status = CDFlib(PUT_, zVAR_HYPERDATA_, QiSCDF->vardata[n]->data, NULL_);
    if(status != CDF_OK) return FAIL_WRITE_DATA;
  
  /* write variable attributes */
  
  check = QiWriteVarAttrs(QiSCDF, n); 
  
  return check;
  

}  /* end QiWriteVariables */


/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiWriteVarAttrs
*
* PURPOSE: Write out the attributes associated with variable.
*
* DESCRIPTION: writes out all variable attributes associated with current variable
*              in metadata, creating it if not already present in cdf file.
*
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            Otherwise   Fatal error detected.
*
* FUNCTIONS_USED: CDFlib
*           
*
* ALGORITHM: Select V attr entry for variable.
*            For each V attribute entry in metadata at current variable:
*              Select attrbute by name, if not present create it.
*              write entry.
*         
***********************************************************************/
 
 long QiWriteVarAttrs( QiCDFContents *QiSCDF,
                       long var_num)
{
 long n;
 long check = QMW_OK;
 long attr_num;
 CDFstatus status;

 /* select entry appropriate for variable */

    status = CDFlib(SELECT_, zENTRY_NAME_, 
                          QiSCDF->vardata[var_num]->name, 
                   NULL_);
      if(status != CDF_OK) return BAD_VARIABLE_NAME;

 /* loop through entries in var attributes in structure */
 
 for (n=0; n < QiSCDF->vardata[var_num]->num_v_attrs; n++){
 
  if(QiSCDF->vardata[var_num]->attribute[n].num_elems > 0){
    
    /* select or create attribute */
 
    status = CDFlib(SELECT_, ATTR_NAME_, 
                          QiSCDF->vardata[var_num]->attribute[n].name, 
                   NULL_);
    if(status != CDF_OK){
      status = CDFlib(CREATE_, ATTR_,  
                          QiSCDF->vardata[var_num]->attribute[n].name,
                                      VARIABLE_SCOPE, &attr_num, 
                      NULL_);
      if(status != CDF_OK){
        check = NO_VAR_ATTR;
        continue;              /* skip this as we can't select or create it */
      }
    }
    
    /* write  entry */ 
      
        status = CDFlib(PUT_, zENTRY_DATA_,  
                        QiSCDF->vardata[var_num]->attribute[n].data_type, 
                        QiSCDF->vardata[var_num]->attribute[n].num_elems,
                        QiSCDF->vardata[var_num]->attribute[n].data, 
                      NULL_);
      
      if (status != CDF_OK) {
         check = FAIL_V_ATTR_PUT;
      }  
  } /* end if on elements */   
    
 } /* end for over V attributes */
    
    return check;

} /* end QiWriteVarAttrs */

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  Qie
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiStatusHandler
*
* PURPOSE: Process response status of CDF library function calls.
*
* DESCRIPTION: Determine status returned by CDF library calls. Use returned 
*              status to determine nature of warnings/errors and initiate 
*              error processing when appropriate.
*
* ARGUMENTS:
*    IN:  CDFstatus  status
*    IN:  char       *note
*
* GLOBALS:  FILE *logfile
*
* RETURN:
*    TYPE:   void
*    VALUES: none
*
* FUNCTIONS_USED:
*         CDFerror
*         QiErrorHandler
*
* ALGORITHM:
*       If status is OK return. 
*       Otherwise output note and message returned from CDFerror to log file. 
*       Pass warning or error to QiErrorHandler.
*
***********************************************************************/

long QiStatusHandler (CDFstatus status, char *note)

{
  char message[CDF_STATUSTEXT_LEN+1];
  long check;

  if (status == CDF_OK){
    /* do nothing */
    return QMW_OK;
  }
  
  /* else print location note then fetch CDF message and print */
  
  fprintf(stderr,"%s\n", note);

  if (status < CDF_WARN){
     check = QMW_ERROR;
  }
  else
     if (status < CDF_OK){   
        fprintf (stderr,"Warning message from CDF...\n");
        check = QMW_WARNING;
     }
     else
     {
        fprintf (stderr, "CDF function executed OK, but...\n");
        check = QMW_OK;
     }      
   
     CDFerror (status, message);
     fprintf (stderr, " CDF> %s\n",message);
     
     return check;

}  /* end QiStatusHandler */

/**********************************************************************
************************ START OF CDF DATA IMPORT FUNCTIONS ***********
***********************************************************************/

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiGetCSDSgenCDF
*
* PURPOSE: Get data for whole CDF file. Metadata are retrieved separately.
*
* DESCRIPTION: Opens named file and reads each variable in turn.
*              r-variables are coerced into z-variables on the fly.
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   int
*    VALUES: QMW_OK         Executed successfully.
*            CDF_OPEN_ERR   Unable to open named file.
*            CDF_N_VAR_ERR  Unable to get number of variables.
*
* FUNCTIONS_USED:
*           
*
* ALGORITHM:
*           
*
***********************************************************************/

int QiGetCSDSgenCDF(const char *file_name,     /* name of cdf file with path */
                 QiCDFContents *Scdf_data)  /* ptr to struct for data return */

{

  CDFstatus status;
  CDFid id;
  long n_vars, n;
  long max_rec;
  long check;
  long time_var;
  
   /* open named cdf file */
   
   status = CDFlib(OPEN_, CDF_, file_name, &id, NULL_);
     if (status != CDF_OK) return CDF_OPEN_ERR; 
   
   /* set to z mode to convert r to z variables */
   
    CDFlib(SELECT_, CDF_zMODE_, zMODEon2, NULL_);

   /* For each variable in file get data */
   
   status = CDFlib(GET_, CDF_NUMzVARS_, &n_vars, 
                         zVARs_MAXREC_, &max_rec,
                   NULL_);
     if (status != CDF_OK) return CDF_N_VAR_ERR;  

   /* keep record of number of variables and number of records */

   Scdf_data->n_vars = n_vars;
   Scdf_data->n_recs = max_rec + 1;

   /* allocate memory for variable pointers */
 
   Scdf_data->vardata = (QiCDFVariable **) QiMakeQiVariablePtrs( n_vars);
    
   for(n=0; n<n_vars; n++){
   
   status = CDFlib (SELECT_, zVARs_RECNUMBER_, 0, NULL_);
     if (status != CDF_OK) return CDF_SELECT_ERR;  
 
   /* allocate memory for variable structure */

   Scdf_data->vardata[n] = (QiCDFVariable *) QiMakeQiVariable( );
 
     check = QiGetCDFVar( Scdf_data->vardata[n], n, &time_var);

     if(check != CDF_OK) {
        Scdf_data->vardata[n] = NULL; /* lose partially filled object */
        return check;
     }
     else{
       if ((time_var == KEEP) && (Scdf_data->time_var_num == -1)) 
                  Scdf_data->time_var_num = n;
     }
   }  
    
   /* retrieve metadata */
   
   check = QiGetCDFmetadata(Scdf_data );
     if(check != QMW_OK) return check;
   
    
   status = CDFlib(CLOSE_, CDF_, NULL_);
     if (status != CDF_OK) return QMW_WARNING;  

   return QMW_OK;
  

} /* end QiGetCSDSgenCDF */

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiGetCDFmetadata
*
* PURPOSE: Fills metadata into QiCDFContents structure.
*
* DESCRIPTION:  
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK         Executed successfully.
*
* FUNCTIONS_USED:
*           
*
* ALGORITHM:
*        
*
***********************************************************************/

long QiGetCDFmetadata (QiCDFContents *Scdf_data)  /* ptr to struct for data return */
                  
{
long num_attrs;
long attr_scope;
long i, j;

  /* find number of G attrs and malloc pointers */
  
  if( CDFlib( GET_, CDF_NUMgATTRS_, &(Scdf_data->num_g_attrs), NULL_) != CDF_OK )
    return FAIL_ON_CDF_READ;
    
  Scdf_data->g_attr =
          (QiGlobalAttribute **) QiMakeQiGAttrPtrs( Scdf_data->num_g_attrs);
  
  /* loop through all attributes and get G attrs */
     
  if( CDFlib( GET_, CDF_NUMATTRS_, &num_attrs, NULL_) != CDF_OK )
    return FAIL_ON_CDF_READ;

  j=0;
  for( i=0 ; i < num_attrs ; i++ )
  {
    if( CDFlib( SELECT_, ATTR_, i, NULL_) != CDF_OK )
      return FAIL_ON_SELECT_ATTR;

    if( CDFlib( GET_, ATTR_SCOPE_,  &attr_scope, NULL_) != CDF_OK )
      return FAIL_ON_CDF_READ;

    if( attr_scope == GLOBAL_SCOPE )
    {

    Scdf_data->g_attr[j] = (QiGlobalAttribute *) QiMakeQiGAttr();

      if( QiGetGlobalAttr( j, Scdf_data->g_attr[j] ) != QMW_OK )
        return FAIL_ON_G_ATTR_READ;
      j++;
    }   /* else variable scope and ignore here */

  } /* end loop over attributes */

  return QMW_OK;
  
}  /* end QiGetCDFmetadata */

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE 
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiGetGlobalAttr
*
* PURPOSE: Read in global attribute entries for attribute with specified number.
*
* DESCRIPTION:  
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   CDFstatus
*    VALUES: QMW_OK         Executed successfully.
*
* FUNCTIONS_USED:
*           
*
* ALGORITHM:
*        
*
***********************************************************************/

CDFstatus QiGetGlobalAttr( long attr_num, 
                           QiGlobalAttribute *g_attr )
{

char attribute_name[ CDF_ATTR_NAME_LEN +1 ];
CDFstatus status;

long i;
long check;

g_attr->number = attr_num;

if(  CDFlib( GET_,
             ATTR_NAME_, attribute_name,
             ATTR_MAXgENTRY_, &(g_attr->num_entries),
             NULL_ )!= CDF_OK ){
     QieAlertBox("warning","Failed to get name for attribute");
     return FAIL_ON_G_ATTR_READ;
 }

/* actually what comes back is the maximum gEntry number, but we want
** to use the number of gEntries _including_ any gEntries which 
** have no data (ie ones which "don't exist")
** This is why we do not use ATTR_NUMgENTRIES_
**
** therefore increment ..
**/

g_attr->num_entries++;

g_attr->name = (char *) malloc(strlen(attribute_name) +1);
strcpy(g_attr->name, attribute_name);

g_attr->entry = ( QiGAttrEntry *) QiMakeEntries(g_attr->num_entries);

for( i = 0; i < g_attr->num_entries; i++ )
{
  if( CDFlib( CONFIRM_, gENTRY_EXISTENCE_, i, NULL_) == CDF_OK )
  {
    g_attr->entry[i].exists = 1;
    status = CDFlib( SELECT_,
                       gENTRY_, i,
                     GET_,
                       gENTRY_DATATYPE_, &(g_attr->entry[i].data_type),
                       gENTRY_NUMELEMS_, &(g_attr->entry[i].num_elems),
                     NULL_ );

    if( status != CDF_OK ){
      QieAlertBox(g_attr->name, "- Failed to get entry info for attribute");
      return FAIL_ON_G_ATTR_READ;
    }

    g_attr->entry[i].data = QiAttrEntryMalloc( g_attr->entry[i].num_elems, 
                              g_attr->entry[i].data_type );

    status = CDFlib( GET_, gENTRY_DATA_, g_attr->entry[i].data, NULL_ );


  } else
  {
    g_attr->entry[i].exists = 0;
  }

} /* end loop over entries */

check = QiStatusHandler(status,g_attr->name);
return check;

}  /* end QiGetGlobalAttr */
/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE 
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiGetCDFVar
*
* PURPOSE: Fills data into QiCDFVariable structure.
*
* DESCRIPTION:  
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK         Executed successfully.
*
* FUNCTIONS_USED:
*           
*
* ALGORITHM:
*        
*
***********************************************************************/
                                                                        
long QiGetCDFVar(QiCDFVariable *vardata, /* ptr to var data struct */
                long number,             /* var number in cdf & struct array */
                long * time_flag)        /* flag for time line id */
{
  CDFstatus status;
  char name[CDF_VAR_NAME_LEN+1];
  size_t length;
  int size_of_datatype;
  long dim_indices[CDF_MAX_DIMS];
  long dim_counts[CDF_MAX_DIMS];
  long n;

  *time_flag = IGNORE;

  /* select variable */
  
  status = CDFlib(SELECT_, zVAR_, number,
                           zENTRY_, number,
                           NULL_);
  if (status != CDF_OK) return CDF_VAR_N_ERR;  
 

  /* fill in info on variable */
  
  vardata->number = number;
  
  status = CDFlib(GET_, zVAR_NAME_, &name[0], NULL_);
  if (status != CDF_OK) return CDF_VAR_NAME_ERR;
  vardata->name = (char *) malloc( sizeof(char)*(strlen(name)+1) );
  strcpy(vardata->name, name);
  
  status = CDFlib(GET_, zVAR_DATATYPE_, &(vardata->data_type), NULL_);
  if (status != CDF_OK) return CDF_VAR_TYPE_ERR;

  status = CDFlib(GET_, zVAR_RECVARY_, &(vardata->rec_vary), NULL_);
  if (status != CDF_OK) return CDF_VAR_RVARY_ERR;
  
  status = CDFlib(GET_, zVAR_NUMELEMS_, &(vardata->num_elems), NULL_);
  if (status != CDF_OK) return CDF_VAR_ELEMS_ERR;
  
  status = CDFlib(GET_, zVAR_NUMDIMS_, &(vardata->num_dims), NULL_);
  if (status != CDF_OK) return CDF_VAR_DIMS_ERR;
  
  status = CDFlib(GET_, zVAR_MAXREC_, &(vardata->max_rec_num), NULL_);
  if (status != CDF_OK) return CDF_VAR_DIMS_ERR;
  
  if (vardata->max_rec_num < 0) {
     vardata->max_rec_num = 0;
     /* -1 returned for virtual records! */
    QieAlertBox( &name[0], ": variable is virtual"); 
  }

  if (vardata->num_dims > 0 ){
  
    status = CDFlib(GET_, zVAR_DIMSIZES_, &(vardata->dim_sizes[0]), NULL_);
    if (status != CDF_OK) return CDF_VAR_SIZES_ERR;

    status = CDFlib(GET_, zVAR_DIMVARYS_, &(vardata->dim_varys[0]), NULL_);
    if (status != CDF_OK) return CDF_VAR_DVARY_ERR;
    
  }
  
  /* Allocate memory for data and indices and counts for reading */
  
  switch(vardata->data_type){
    case CDF_CHAR: 
    case CDF_UCHAR: 
    case CDF_UINT1:
    case CDF_INT1:
    case CDF_BYTE:
    {
      size_of_datatype = 1;
      break;
    }
    case CDF_INT2:
    case CDF_UINT2:
    {
      size_of_datatype = 2;
      break;
    }
    case CDF_INT4:
    case CDF_UINT4:
    case CDF_REAL4:
    case CDF_FLOAT:
    {
      size_of_datatype = 4;
      break;
    }
    case CDF_REAL8:
    case CDF_DOUBLE:
    {
      size_of_datatype = 8;
      break;
    }
    case CDF_EPOCH:
    {
      size_of_datatype = 8;
      *time_flag = KEEP;
      break;
    }
    default:
    {
    return BAD_CDF_DATATYPE;
    } 
  }
 
  length = (size_t) vardata->num_elems * (vardata->max_rec_num +1)
                    * size_of_datatype;
                    
     
  for (n=0; n < vardata->num_dims; n++){    
     
    dim_indices[n] = 0;

    if(vardata->dim_varys[n] == VARY){
      length = length * vardata->dim_sizes[n];
      dim_counts[n] = vardata->dim_sizes[n];
    }
    else{
      dim_counts[n] = 1;
    }   
  }
  vardata->data = (void *) malloc (length);
  
  /* get data array in same ordering as cdf file */
    
  status = CDFlib (SELECT_, zVAR_DIMCOUNTS_, dim_counts, 
                            zVAR_DIMINDICES_, dim_indices,
                            zVAR_RECCOUNT_, vardata->max_rec_num + 1,
                   NULL_);
  if (status < CDF_WARN) {
     fprintf(stderr,"CDFlib err code %ld", (long)status);
     return CDF_VAR_SELECT_ERR;
  }

  status = CDFlib (GET_, zVAR_HYPERDATA_, vardata->data, NULL_);
  if (status < CDF_WARN)  {
     fprintf(stderr, "CDFlib err code %ld", status);
     return CDF_VAR_GET_DATA_ERR;
  }
  
  /* get V attrs */
  
  if( QiGetVariableAttr( number, vardata ) != QMW_OK )  return FAIL_ON_V_ATTR_READ;

  return QMW_OK;
  
} /* end QiGetCDFVar */

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE 
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiAttrEntryMalloc
*
* PURPOSE: Allocate memory for attribute entry
*
* DESCRIPTION:  
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   void *
*    VALUES: pointer to new memory allocated for entry.
*
* FUNCTIONS_USED:
*           
*
* ALGORITHM:
*        
*
***********************************************************************/

void *QiAttrEntryMalloc( long num_elems, long data_type )
{
void *p;
size_t size;

if( num_elems < 1 ) num_elems = 1;   /* some defensive programming ... */

switch (data_type) {
  case CDF_BYTE: case CDF_INT1: case CDF_UINT1: 
  {
    size = num_elems;
    break;
  }
  case CDF_INT2: case CDF_UINT2:
  {
    size = 2 * num_elems;
    break;
  }
  case CDF_INT4: case CDF_UINT4: case CDF_REAL4: case CDF_FLOAT: 
  {
    size = 4 * num_elems;
    break;
  }

  case CDF_REAL8:  case CDF_EPOCH: case CDF_DOUBLE:
  {
    size = 8 * num_elems;
    break;
  }

  case CDF_CHAR: case CDF_UCHAR: 
  {
    size = num_elems + 1;
    break;
  }
  default: size = 8 * num_elems;
}

p = (void *)malloc( size );

if( data_type == CDF_CHAR || data_type == CDF_UCHAR )
  ((char *)p)[num_elems] = '\0' ;

return p;

}   /* end QiAttrEntryMalloc  */

/***********************************************************************
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QSAS
* 
* MODULE:  QIE
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiGetVariableAttr
*
* PURPOSE: Read all variable attribute entries for specified variable
*
* DESCRIPTION: 
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            Otherwise   Fatal error detected.
*
* FUNCTIONS_USED:
*         
*         
*
* ALGORITHM:  
*
*
***********************************************************************/

long QiGetVariableAttr( long var_num,
                        QiCDFVariable * var)
{

CDFstatus status;

long i;
long max_num_attrs;
long attr_num;
long attr_scope;
char attribute_name[ CDF_ATTR_NAME_LEN +1 ];


  if(CDFlib( GET_, CDF_NUMATTRS_, &max_num_attrs, NULL_ ) != CDF_OK )
    return FAIL_ON_SELECT_ATTR;


 /* first  test we have enough space */
  if(max_num_attrs > MAX_N_ATTRS) max_num_attrs = MAX_N_ATTRS;
  
  /* count entries as we go, zENTRY_ already selected by calling routine */
                 
  attr_num = 0;

 /* read attribute entries: if we fail, try next in list */

for( i=0 ; i < max_num_attrs ; i++ )
{
  if( CDFlib( SELECT_, ATTR_, i, NULL_) != CDF_OK ) continue;

  if( CDFlib( GET_, ATTR_SCOPE_,  &attr_scope, NULL_) != CDF_OK ) continue;
  
  if( CDFlib( CONFIRM_, zENTRY_EXISTENCE_, var_num, NULL_) != CDF_OK ) continue;

  if( attr_scope == VARIABLE_SCOPE ){
    
    if( CDFlib( GET_, ATTR_NAME_, attribute_name, NULL_ ) 
        != CDF_OK ) continue;

    var->attribute[attr_num].number = attr_num;
    var->attribute[attr_num].name = (char *) malloc(strlen(attribute_name)+1);
    strcpy(var->attribute[attr_num].name, attribute_name);

    if( CDFlib( GET_,
                zENTRY_DATATYPE_, &(var->attribute[attr_num].data_type),
                zENTRY_NUMELEMS_, &(var->attribute[attr_num].num_elems),
                NULL_ ) 
       != CDF_OK ) continue;
    

/* allocate entry space */

      var->attribute[attr_num].data =
        QiAttrEntryMalloc( var->attribute[attr_num].num_elems, 
                           var->attribute[attr_num].data_type );

      if( ( status = CDFlib(
                  GET_,  zENTRY_DATA_, var->attribute[attr_num].data,
                  NULL_ ) )
          != CDF_OK ) continue;
       
       /* if we get here the attribute is read OK, so increment counter */
          
       attr_num++;
     
  }  /* end if on VARIABLE_SCOPE */

} /* end for loop over attributes */
       
  var->num_v_attrs = attr_num;

  return QMW_OK;

} /* end QiGetVariableAttr */

