/* PGSpectro 2-D spectrogram and PGWheel routines using Pgplot
*  Also simple XY traces
*
*  Handles both Y-X and Y-time arrays for spectrograms
*
*  Language:  C
*
*  History:
*  July 2000 SJS First Draft
*  25 April 2001 - up-dated
*  1 Oct 2001 - fixed bug in calculating colour from zval and zrange by
*               adding cast to float and extra parenthesis to prevent
*               early truncating cast to int
*    Oct 2001 - added XYtrace routines
*  Jan 2002   - minor bug fix (double -> int for cpgslw)
*  Mar 2002   - add missing values for PGSpectro and time for traces
* 
*  See header file for description of parameters and defaults
*
*  Use:
*
*    Allocate PGSpectro or PGWheel structure
*    Fill with defaults
*    Fill with actual arrays, strings for labels
*    Set options
*    Pass to plot routine
*
*    Notes: Pgplot works entirely in floats
*           Time values should be in seconds of the day
*/

#ifdef __cplusplus
extern "C"{
#endif

#include "PGSpectro.h"

/*****************************************************************/
int set_pgp_rainbow(PGSpectro_col color_table){
/*****************************************************************/
/* set pgplot colours to a rainbow, with red at max of col index,
*  purple at min, or to gray with white as either low or high.
*  Returns number of colors in table
*
*  Need to do more than once if cpggray is used, as it re-defines
*  the col table
*
*/

int i;
int icollow, icolhigh, ncolors;
float huestart, hueend, huestep;
float lightstart, lightend, lightstep;
float hue, lightness, saturation;

cpgqcir(&icollow, &icolhigh);

/* printf(" device colour range is %d - %d \n",icollow, icolhigh); */

ncolors = icolhigh - icollow +1;
ncolors = (ncolors > NCOLOURS_MAX) ? NCOLOURS_MAX : ncolors;

cpgscir(icollow,icollow+ncolors-1); /* use only available color indices */

switch (color_table){
   case PGSPECTRO_GRAY_WHITE_HIGH:
       lightstart = 0.;
       lightend  = 1.;
       lightstep = 1./(float)(ncolors-1);
       hue = 0;
       saturation = 0;
       for (i = icollow, lightness = lightstart;
            i<icolhigh+1;
            i++, lightness+=lightstep){
         cpgshls(i, hue, lightness, saturation);
       }
       break;
   case PGSPECTRO_GRAY_WHITE_LOW:
       lightstart = 1.;
       lightend  = 0.;
       lightstep = -1./(float)(ncolors-1);
       hue = 0;
       saturation = 0;
       for (i = icollow, lightness = lightstart;
            i<icolhigh+1;
            i++, lightness+=lightstep){
         cpgshls(i, hue, lightness, saturation);
       }
       break;

   case PGSPECTRO_RAINBOW: default:
       huestart = 420.;  /* <=> 60 deg; this is kind of purple */
       hueend = 110.;    /* this is red */
       huestep = -(huestart - hueend)/(float)(ncolors-1);  /* step backward */
       lightness = 0.5;
       saturation = 1.0;
       for (i = icollow, hue=huestart; i<icolhigh+1; i++, hue+= huestep){
         /* printf(" i = %d, hue = %g\n",i,hue);  */
         cpgshls( i, hue, lightness, saturation);
       }
       break;
}

return ncolors;
}

/*****************************************************************/
void spectro_wedge(PGSpectro_wedge wedge, float zscalemin,
           float zscalemax, PGSpectro_linlog lin_log) {
/*****************************************************************/
/* sjs routine to draw spectrogram key to the right of the current viewport */

float x1vp_save, x2vp_save, y1vp_save, y2vp_save;
float x1win_save, x2win_save, y1win_save, y2win_save; /* world coords */
int ci;  /* color index */
float x1,x2,y1,y2;
float chsize_ndc,dummy;

int i,col, icollow, icolhigh, ncolors;  /* see set_rainbow for algorithm */
float zstep, zrange;

cpgqwin(&x1win_save, &x2win_save, &y1win_save, &y2win_save); /* save window */
cpgqci(&ci);

cpgqvp(0, &x1vp_save, &x2vp_save, &y1vp_save, &y2vp_save);/* get vport in ndc*/
cpgqcs(0,&chsize_ndc,&dummy);               /* get curr char size in ndc for */
                                                       /*text with vert base */
x1 = x2vp_save + wedge.disp*chsize_ndc;
x2=x1 + wedge.width*chsize_ndc;

cpgsvp(x1,x2,y1vp_save, y2vp_save);
cpgswin(0.,1.,zscalemin, zscalemax);

cpgqcir(&icollow, &icolhigh);                 /* get colour range of device. */

ncolors = icolhigh - icollow +1;                  /* copied from set_rainbow */
ncolors = (ncolors > NCOLOURS_MAX) ? NCOLOURS_MAX : ncolors;

zrange = zscalemax - zscalemin;
zstep = zrange/(ncolors - 1);
for (i = 0; i<ncolors; i++){
   x1=0.0; x2=1.0;
   y1= zscalemin + ((float)i-0.5)*zstep;
   y2=y1+zstep;
   col = icollow + i;
   cpgsci(col);
   cpgrect(x1,x2,y1,y2);
}
cpgsci(1);
switch (lin_log) {
    case PGSPECTRO_LOG:
      cpgtbox("BC",0.0, 0, "BCMTSLV", 0.0, 0);
      break;
    case PGSPECTRO_LIN:
      cpgtbox("BC",0.0, 0, "BCMTSV", 0.0, 0);
      break;
    default:
      printf("Unknown PGSpectro_linlog in spectro_wedge\n");
      break;
}
cpgmtxt("L",0.5, 0.5, 0.5, wedge.label);

/* restore pgsettings for user */

cpgsvp(x1vp_save, x2vp_save, y1vp_save, y2vp_save); /* restore vport in ndc */
cpgswin(x1win_save, x2win_save, y1win_save, y2win_save);  /* restore window */
cpgsci(ci);                                          /* restore color index */


return ;
}
/*****************************************************************/
void PGSpectro_set_defaults(PGSpectro *spectro){
/*****************************************************************/

/* set default values for as much as possible */

spectro->nx = 0;            spectro->ny = 0;
spectro->zscalemin = 1.0;   spectro->zscalemax = 10.0;
spectro->skip_missing = NO; spectro->zmissing = -1.e31;
spectro->xlabel = " ";      spectro->ylabel = " ";
spectro->title = " ";       spectro->subtitle = " ";

spectro->color_table = PGSPECTRO_RAINBOW;
PGSpectro_domain_set_defaults(  &(spectro->domain) );
spectro->wedge.disp = 1.5;
spectro->wedge.width = 1.5;
spectro->wedge.label = " ";
spectro->xlinlog =  PGSPECTRO_LIN;  spectro->ylinlog =  PGSPECTRO_LIN;
spectro->zlinlog =  PGSPECTRO_LIN;
spectro->xnums = YES;               spectro->ynums = YES;
spectro->ylabel_vert = NO;
spectro->draw_wedge = YES;
spectro->timex = NO;
spectro->datestart = " ";
spectro->dateend = " ";
spectro->char_size = PGSPECTRO_DEFAULT_CHAR_SIZE;

return;
}


void PGSpectro_domain_set_defaults (PGSpectro_domain *domainp){

domainp->xworldmin = 1.0;   domainp->xworldmax = 10.0;
domainp->yworldmin = 1.0;   domainp->yworldmax = 10.0;
domainp->xndcmin   = 0.1;   domainp->xndcmax   = 0.85;
domainp->yndcmin   = 0.1;   domainp->yndcmax   = 0.9;

return;
}

/*****************************************************************/
PGSpectro_ret PGSpectro_Plot(PGSpectro *spectro){
/*****************************************************************/

int ncolours,ncolminusone;
int col, icollow=LOWEST_COLOR_INDEX;
int pgid;
int i, j, k;
float zrange,zmin,zmax, zval;
int skip;
float skip_val;
float xworld1, xworld2, yworld1, yworld2;
float ztemp;
float *xlo, *xhi, *ylo, *yhi, *z;  /* local pointers */
int nx, ny;  /* local copies */
char xopt[20], yopt[20];
int iopt=0, jopt=0;

cpgqid(&pgid);
if (pgid == 0) return PGSPECTRO_NO_DEVICE;   /* bale out if no pgplot device */

cpgiden();
cpgsitf(0);  /* 0=lin, 1 = log, 2 = sqrt; do log manually, hence always lin */
cpgsvp(spectro->domain.xndcmin, spectro->domain.xndcmax,
          spectro->domain.yndcmin, spectro->domain.yndcmax);

ncolours = set_pgp_rainbow(spectro->color_table);
ncolminusone = ncolours -1;
if (spectro->zscalemax > spectro->zscalemin ){       /* no inverted scales */
   zmin = spectro->zscalemin;
   zmax = spectro->zscalemax;
} else {
   zmin = spectro->zscalemax;
   zmax = spectro->zscalemin;
}

nx = spectro->nx;
ny = spectro->ny;
skip = (spectro->skip_missing == YES);
skip_val = spectro->zmissing;

if (spectro->zlinlog == PGSPECTRO_LOG){
   zmin = SafeLog10 (zmin);
   zmax = SafeLog10 (zmax);
   z = (float *)malloc(nx*ny*sizeof(float));
   for (i=0; i<nx; i++){
     for(j=0; j<ny; j++){
       k=ny*i+j;
       ztemp = spectro->z[k];
       if (skip && ztemp == skip_val){
           z[k]=SKIP_LOG_VALUE;
       } else {
           z[k]=SafeLog10(ztemp);
       }
     }
   }
   skip_val = SKIP_LOG_VALUE; /* now tests are done, re-define to plot */
} else {
   z = spectro->z;
}

zrange = zmax- zmin;

/* x-axis. Ignore log if time axes to be used */
if(spectro->xlinlog == PGSPECTRO_LOG && spectro->timex == NO){
   xlo = (float *)malloc(spectro->nx * sizeof(float));
     if (xlo == NULL) return PGSPECTRO_MALLOC_FAILURE;
   xhi = (float *)malloc (spectro->nx * sizeof(float));
     if (xhi == NULL) return PGSPECTRO_MALLOC_FAILURE;
   for (i=0; i<nx; i++){
     xlo[i]  = SafeLog10(spectro->xbinlow[i]);
     xhi[i]  = SafeLog10(spectro->xbinhigh[i]);
   }
   xworld1 = SafeLog10(spectro->domain.xworldmin);
   xworld2 = SafeLog10(spectro->domain.xworldmax);
} else {
   xlo  = spectro->xbinlow;
   xhi  = spectro->xbinhigh;
   xworld1 = spectro->domain.xworldmin;
   xworld2 = spectro->domain.xworldmax;
}

if(spectro->ylinlog == PGSPECTRO_LOG){
   ylo = (float *)malloc(spectro->ny * sizeof(float));
     if (ylo == NULL) return PGSPECTRO_MALLOC_FAILURE;
   yhi = (float *)malloc (spectro->ny * sizeof(float));
     if (yhi == NULL) return PGSPECTRO_MALLOC_FAILURE;
   for (i=0; i<ny; i++){
     ylo[i] = SafeLog10(spectro->ybinlow[i]);
     yhi[i]  = SafeLog10(spectro->ybinhigh[i]);
   }
   yworld1 = SafeLog10(spectro->domain.yworldmin);
   yworld2 = SafeLog10(spectro->domain.yworldmax);

} else {
   ylo  = spectro->ybinlow;
   yhi  = spectro->ybinhigh;
   yworld1 = spectro->domain.yworldmin;
   yworld2 = spectro->domain.yworldmax;
}

cpgswin(xworld1, xworld2, yworld1, yworld2);


for (i=0; i<nx; i++){
  for(j=0; j<ny; j++){
    k = ny*i+j;
    zval = z[k];
    if (skip && zval == skip_val){  /* fill with background */
        cpgsci(0);
    } else {
	zval = (zval< zmin)? zmin : zval;
	zval = (zval > zmax) ? zmax : zval;
	col = icollow+(int)(((float)ncolminusone*(zval- zmin) )/zrange);
	cpgsci(col);
    }
    cpgrect(xlo[i],xhi[i],ylo[j],yhi[j]);
/* printf("i = %d, j = %d, xlo = %g, xhi = %g, ylo = %g, yhi = %g, col = %d\n",
          i, j, xlo[i], xhi[i], ylo[j], yhi[j], col);
*/
  }
}

cpgsci(1);  /* reset default colour */
cpgsch(spectro->char_size);

strcpy(xopt,"BCTS");   strcpy(yopt, "BCTS");
iopt = strlen(xopt);   jopt = strlen(yopt);

if (spectro->xlinlog ==PGSPECTRO_LOG && spectro->timex == NO)xopt[iopt++] = 'L';
if (spectro->ylinlog == PGSPECTRO_LOG) yopt[jopt++] = 'L';
if (spectro->xnums   == YES          ) xopt[iopt++] = 'N';
if (spectro->ynums   == YES          ) yopt[jopt++] = 'N';

if (spectro->timex == YES){                     /* time axis labelling, etc. */
  xopt[iopt++] = 'Z';     xopt[iopt++] = 'H';
  xopt[iopt++] = 'X';     xopt[iopt++] = 'O';
}

xopt[iopt] = '\0';                              /* terminate options strings */
yopt[jopt] = '\0';

cpgtbox(xopt, 0.0, 0, yopt , 0.0, 0);  /* box after to see ticks! */
if (spectro->draw_wedge != NO){
  spectro_wedge(spectro->wedge,zmin, zmax, spectro->zlinlog);
}
if(spectro->ylabel_vert ==NO){
  cpglab(spectro->xlabel,spectro->ylabel,spectro->title);
} else {
  cpglab(spectro->xlabel, "", spectro->title);
  cpgmtxt("LV", 1.0, 0.5, 1.0, spectro->ylabel);
}
cpgsch(0.75*spectro->char_size);
cpgmtxt("T", 0.5, 0.5,0.5, spectro->subtitle);
cpgsch(spectro->char_size);


if (spectro->timex == YES){
   cpgmtxt("B", 2.5, 0.0, 0.5, spectro->datestart);
   if ( strcmp(spectro->datestart,spectro->dateend) ){
     cpgmtxt("B", 2.5, 1.0, 1.0, spectro->dateend);
   }
}

if (spectro->xlinlog == PGSPECTRO_LOG && spectro->timex == NO) {
   free(xlo); free (xhi);
}
if (spectro->ylinlog == PGSPECTRO_LOG) {
   free(ylo); free (yhi);
}
if (spectro->zlinlog == PGSPECTRO_LOG) free (z);

return PGSPECTRO_OK;
}

/*****************************************************************/
void PGWheel_set_defaults(PGWheel *wheel){
/*****************************************************************/

/* set default values for as much as possible */

wheel->nr = 0;          wheel->nth = 0;
wheel->ncircles = 0;
wheel->zscalemin = 1.0; wheel->zscalemax = 10.0;
wheel->zmissing = -1.e31; wheel->skip_missing = NO;
wheel->xlabel = " "; wheel->ylabel = " "; wheel->title = " ";
wheel->subtitle = " "; wheel->subsubtitle = " ";
wheel->suptitle = " "; wheel->supsuptitle = " ";

wheel->color_table = PGSPECTRO_RAINBOW;
PGWheel_domain_set_defaults(  &(wheel->domain) );
wheel->wedge.disp = 1.5;
wheel->wedge.width = 1.5;
wheel->wedge.label = " ";
wheel->rlinlog =  PGSPECTRO_LIN;
wheel->zlinlog =  PGSPECTRO_LIN;
wheel->rnums = YES;
wheel->title_just = 0.5;

return;
}


void PGWheel_domain_set_defaults (PGWheel_domain *domainp){

domainp->rmin = 0.0;   domainp->rmax = 10.0;
domainp->xndcmin   = 0.15;   domainp->xndcmax   = 0.8;
domainp->yndcmin   = 0.15;   domainp->yndcmax   = 0.8;

return;
}

/*****************************************************************/
PGSpectro_ret PGWheel_Plot(PGWheel *wheel){
/*****************************************************************/

int ncolours,ncolminusone;
int col, icollow=LOWEST_COLOR_INDEX;
int pgid;
int i, j, k;
float zrange,zmin,zmax, zval;
int skip;
float skip_val;
float r1, r2, r2a;
float ztemp;
float *rlo, *rhi, *thlo, *thhi, *z;  /* local pointers */
int nr, nth;  /* local copies */
char ropt[20];
float rcirc;

cpgqid(&pgid);
if (pgid == 0) return PGSPECTRO_NO_DEVICE;   /* bale out if no pgplot device */

cpgiden();
cpgsitf(0);  /* 0=lin, 1 = log, 2 = sqrt; do log manually, hence always lin */
cpgsvp(wheel->domain.xndcmin, wheel->domain.xndcmax,
          wheel->domain.yndcmin, wheel->domain.yndcmax);

ncolours = set_pgp_rainbow(wheel->color_table);
ncolminusone = ncolours -1;
if (wheel->zscalemax > wheel->zscalemin ){       /* no inverted scales */
   zmin = wheel->zscalemin;
   zmax = wheel->zscalemax;
} else {
   zmin = wheel->zscalemax;
   zmax = wheel->zscalemin;
}

nr  = wheel->nr;
nth = wheel->nth;
skip = (wheel->skip_missing == YES);
skip_val = wheel->zmissing;

if (wheel->zlinlog == PGSPECTRO_LOG){
   zmin = SafeLog10 (zmin);
   zmax = SafeLog10 (zmax);
   z = (float *)malloc(nr*nth*sizeof(float));
   for (i=0; i<nr; i++){
     for(j=0; j<nth; j++){
       k=nth*i+j;
       ztemp = wheel->z[k];
       if (skip && ztemp == skip_val){
          z[k] = SKIP_LOG_VALUE;
       } else {
          z[k]=SafeLog10(ztemp);
       }
     }
   }
   skip_val = SKIP_LOG_VALUE;  /* re-define now tests are done */
} else {
   z = wheel->z;
}

zrange = zmax- zmin;

/* r-arrays. */
if(wheel->rlinlog == PGSPECTRO_LOG ){
   r1  = SafeLog10(wheel->domain.rmin);     /* "origin" in log r scale */
   r2a =  SafeLog10(wheel->domain.rmax) ;   /* for axes */
   r2  = r2a- r1;                           /* for dist from origin */
   rlo = (float *)malloc(wheel->nr * sizeof(float));
     if (rlo == NULL) return PGSPECTRO_MALLOC_FAILURE;
   rhi = (float *)malloc (wheel->nr * sizeof(float));
     if (rhi == NULL) return PGSPECTRO_MALLOC_FAILURE;
   for (i=0; i<nr; i++){
     rlo[i]  = SafeLog10(wheel->rbinlow[i]) - r1;
     rhi[i]  = SafeLog10(wheel->rbinhigh[i])  - r1;
   }
} else {
   rlo  = wheel->rbinlow;
   rhi  = wheel->rbinhigh;
   r1 = wheel->domain.rmin;
   r2 = wheel->domain.rmax;
}

thlo = wheel->thbinlow;
thhi = wheel->thbinhigh;

cpgwnad(-r2, r2, -r2, r2);   /* window with equal x and y axes */


for (i=0; i<nr; i++){
  for(j=0; j<nth; j++){
    k = nth*i+j;
    zval = z[k];
    if (skip && zval == skip_val){  /* fill with background */
       cpgsci(0);
    } else {
       zval = (zval< zmin)? zmin : zval;
       zval = (zval > zmax) ? zmax : zval;
       col = icollow+(int)( ((float)ncolminusone*(zval- zmin) )/zrange);
       cpgsci(col);
    }
    if (rhi[i] > 0.0){   /* at least part of arc has positive r */
       if (rlo[i] < 0.0) rlo[i] = 0.0;   /* clip any negative portion */
       arc(rlo[i],rhi[i],thlo[j],thhi[j]);/* draw an arc;*/
    }
  }
}

cpgsci(1);  /* reset default colour */

spectro_wedge(wheel->wedge,zmin, zmax, wheel->zlinlog);
/* cpglab(wheel->xlabel,wheel->ylabel," "); */
cpgmtxt("B",4.0, 0.5, 0.5, wheel->xlabel);
cpgmtxt("L",4.0, 0.5, 0.5, wheel->ylabel);
cpgsch(1.0);
cpgmtxt("T",4.0, 0.5,wheel->title_just, wheel->supsuptitle);
cpgsch(0.75);
cpgmtxt("T",3.5, 0.5,wheel->title_just, wheel->suptitle);
cpgmtxt("T",2.5, 0.5,wheel->title_just, wheel->title);
cpgmtxt("T", 1.5, 0.5,wheel->title_just, wheel->subtitle);
cpgmtxt("T", 0.5, 0.5,wheel->title_just, wheel->subsubtitle);
cpgsch(1.0);

if (wheel->rlinlog ==PGSPECTRO_LIN )  {

   if (wheel->rnums   == YES )  strcpy(ropt,"ABCTSN"); else strcpy(ropt,"ABCTS");

   cpgtbox(ropt, 0.0, 0, ropt , 0.0, 0);  /* box after to see ticks!      */
                                          /* Then zero axes:              */
   cpgaxis("", 0.0, -r2, 0.0, r2, -r2, r2, 0.0, 1, 0.5, 0.5, 0.5, 1.0, 0);
   cpgaxis("", -r2, 0.0,  r2,0.0, -r2, r2, 0.0, 1, 0.5, 0.5, 0.5, 1.0, 0);

} else if (wheel->rlinlog == PGSPECTRO_LOG ) {
                                   /* for log r need to draw axis segments*/
    if (wheel->rnums == YES) strcpy(ropt,"LN"); else strcpy(ropt, "L");
/* bottom axis */
    cpgaxis(ropt, -r2, -r2, 0.0, -r2, r2a, r1, 0.0, 1, 0.5, 0.0, 0.5, 1.0,320);
    cpgaxis(ropt, 0.0, -r2, r2,  -r2, r1, r2a, 0.0, 1, 0.5, 0.0, 0.5, 1.0, 320);
/* top axis */
    cpgaxis("L" , -r2,  r2, 0.0,  r2, r2a, r1, 0.0, 1, 0.0, 0.5, 0.5, 1.0, 0);
    cpgaxis("L" , 0.0,  r2,  r2,  r2, r1, r2a, 0.0, 1, 0.0, 0.5, 0.5, 1.0, 0);
/* horiz zero axis */
    cpgaxis("L" , -r2, 0.0, 0.0, 0.0, r2a, r1, 0.0, 1, 0.5, 0.5, 0.5, 1.0, 0);
    cpgaxis("L" , 0.0, 0.0,  r2, 0.0, r1, r2a, 0.0, 1, 0.5, 0.5, 0.5, 1.0, 0);
/* left axis */
    cpgaxis(ropt, -r2, -r2, -r2, 0.0, r2a, r1, 0.0, 1, 0.0, 0.5, 0.5, -1.0, 40);
    cpgaxis(ropt, -r2, 0.0, -r2,  r2, r1, r2a, 0.0, 1, 0.0, 0.5, 0.5, -1.0, 40);
/* right axis */
    cpgaxis("L" ,  r2, -r2,  r2, 0.0, r2a, r1, 0.0, 1, 0.5, 0.0, 0.5, 1.0, 0);
    cpgaxis("L" ,  r2, 0.0,  r2,  r2, r1, r2a, 0.0, 1, 0.5, 0.0, 0.5, 1.0, 0);
/* vertical zero axis */
    cpgaxis("L" , 0.0, -r2, 0.0, 0.0, r2a, r1, 0.0, 1, 0.5, 0.5, 0.5, 1.0, 0);
    cpgaxis("L" , 0.0, 0.0, 0.0,  r2, r1, r2a, 0.0, 1, 0.5, 0.5, 0.5, 1.0, 0);
}

/* Draw reference circles */
cpgsave();   /* save attributes */
cpgsci(0);   /* circles in background colour */
cpgsls(2); /* dashed */
cpgsfs(2); /* outline circles */

for (i=0; i< wheel->ncircles; i++){
   if (wheel->rlinlog == PGSPECTRO_LOG){
                rcirc = SafeLog10(wheel->rcircles[i]) - r1;
   }
   else rcirc = wheel->rcircles[i];
   if (rcirc >= 0.0) cpgcirc(0.0, 0.0, rcirc); /* manual clipping at origin */
}

cpgunsa(); /* restore attributes */

/* Cleanup */
if (wheel->rlinlog == PGSPECTRO_LOG) {
   free(rlo); free (rhi);
}
if (wheel->zlinlog == PGSPECTRO_LOG) free (z);

return PGSPECTRO_OK;
}

/******************************************************************/
void arc(float rlo, float  rhi, float thlo,float thhi){
/******************************************************************/

/* draw a pseudo-arc with limits as given using current fill properties
*     r in units of plot, e.g., already converted to logs if log scales
*     th in degrees
*
*     present routine uses one theta at same radii in addition to limits */

float x[6], y[6];

x[0] = rlo*cos(QPGDEG2RAD*thlo); y[0] = rlo*sin(QPGDEG2RAD*thlo);
x[1] = rhi*cos(QPGDEG2RAD*thlo); y[1] = rhi*sin(QPGDEG2RAD*thlo);
x[5] = rlo*cos(QPGDEG2RAD*0.5*(thlo+thhi));
                              y[5] = rlo*sin(QPGDEG2RAD*0.5*(thlo+thhi));
x[2] = rhi*cos(QPGDEG2RAD*0.5*(thlo+thhi));
                              y[2] = rhi*sin(QPGDEG2RAD*0.5*(thlo+thhi));
x[4] = rlo*cos(QPGDEG2RAD*thhi); y[4] = rlo*sin(QPGDEG2RAD*thhi);
x[3] = rhi*cos(QPGDEG2RAD*thhi); y[3] = rhi*sin(QPGDEG2RAD*thhi);

cpgpoly(6,x,y);

return;

}

/****************************************************************/
PGSpectro_ret PGXYtrace_plot(PGXYtrace *t){
/****************************************************************/
/* plot a single XY trace, including (optionally) box and legend
*  For multiple traces, make multiple calls to this routine, changing
*  options as necessary. In particular, setting YoNbox to NO skips all
*  axes, labels, etc. apart from in the legend */

char xopt[50],yopt[50], xlabopt[5], ylabopt[5];
int iopt, jopt;
float xworld1, xworld2, yworld1, yworld2;
float *x, *y;
float datestrcoord;
int i;

cpgsave();  /* save pgstate and restore on return */ 
cpgsci(1);  /* set color to default foreground */

/* check if either axis log; if so, calc logs of limits and data */
if (t->xlinlog != PGSPECTRO_LOG || t->timex == YES) { /*ignore log for time */
  x =  t->x;
  xworld1 = t->plot_domain.xworldmin;  xworld2 = t->plot_domain.xworldmax;
} else {
    x = (float *)malloc (t->npts * sizeof (float) );
    for (i=0; i<t->npts; i++) x[i] = SafeLog10(t->x[i]);
    xworld1 = SafeLog10(t->plot_domain.xworldmin);
    xworld2 = SafeLog10(t->plot_domain.xworldmax);
}

if (t->ylinlog != PGSPECTRO_LOG) {
  y =  t->y;
  yworld1 = t->plot_domain.yworldmin;  yworld2 = t->plot_domain.yworldmax;
} else {
    y = (float *)malloc (t->npts * sizeof (float) );
    for (i=0; i<t->npts; i++) y[i] = SafeLog10(t->y[i]);
    yworld1 = SafeLog10(t->plot_domain.yworldmin);
    yworld2 = SafeLog10(t->plot_domain.yworldmax);
}   

/* viewport etc for plot */

cpgsvp(t->plot_domain.xndcmin, t->plot_domain.xndcmax,
       t->plot_domain.yndcmin, t->plot_domain.yndcmax);

cpgswin(xworld1, xworld2, yworld1, yworld2);

/* box and axes  - we make 2 calls to cpgbox. The first draws the box and
*  axes and ticks, so we can control their length and thickness. The 2nd
*  only puts the numerica labels, so these can be different size and
*  line thickness */

if (t->YoNbox == YES){
   cpgslw(t->box_thickness);
   cpgsch(t->tick_length);
   strcpy(xopt,"BCTS");     strcpy(yopt,"BCTS");
   iopt = strlen(xopt);     jopt = strlen(yopt);
   if (t->xlinlog     == PGSPECTRO_LOG && t->timex == NO) xopt[iopt++] = 'L';
   if (t->ylinlog     == PGSPECTRO_LOG) yopt[jopt++] = 'L';
   if (t->timex == YES){                  /* time axis labelling, etc. */
     xopt[iopt++] = 'Z';     xopt[iopt++] = 'H';
     xopt[iopt++] = 'X';     xopt[iopt++] = 'O';
   }
   xopt[iopt] = '\0';       yopt[jopt]     = '\0';   
   cpgtbox(xopt,0.0,0,yopt,0.0,0);  /* bare box with ticks, no tick labels */
   cpgslw(t->text_line_width);
   cpgsch(t->label_char_size); 
   if (t->YoNXnumlab  == YES) {
      if (t->Xnumlab_bot == NO ) {
        xopt[iopt++] = 'M';} else {xopt[iopt++] = 'N';}
   }
   if (t->YoNYnumlab  == YES) {
      if (t->Ynumlab_left == NO ) {
        yopt[jopt++] = 'M';} else {yopt[jopt++] = 'N';}
   }
   xopt[iopt] = '\0';       yopt[jopt]     = '\0';
   cpgtbox(&(xopt[2]), 0.0, 0,&(yopt[2]), 0.0, 0); /* chop "BC" => tick lab */
}

/* titles and labels */

if (t->YoNbox == YES){
    cpgsch(t->title_char_size);
    cpgslw(t->text_line_width);
    cpgmtxt("T", t->title_disp, 0.5, t->title_just, t->title);
    cpgsch(t->label_char_size);
    cpgmtxt("T", t->subtitle_disp, 0.5, t->title_just, t->subtitle);
    cpgmtxt("T", t->subsubtitle_disp, 0.5, t->title_just, t->subsubtitle);
    if (t->YoNXnumlab == YES){
	if (t->Xtextlab_bot == NO){ strcpy(xlabopt,"T");
	} else { strcpy(xlabopt,"B");
	}
	cpgmtxt(xlabopt,2.0, 0.5, 0.5, t->xlabel);  /* 2 char heights */
    }
    if (t->YoNYnumlab == YES){
	if (t->Ytextlab_left == NO){ strcpy(ylabopt,"R");
	} else { strcpy(ylabopt,"L");
	}
	cpgmtxt(ylabopt,2.0, 0.5, 0.5, t->ylabel);
    }
    if (t->timex == YES){
       datestrcoord = (xworld1<=xworld2) ? 0.0 : 1.0; /* decr time axes ok */
       cpgmtxt("B", 2.5, datestrcoord, 0.5*(1.0+datestrcoord), t->datestart);
       if ( strcmp(t->datestart,t->dateend) ){
	 cpgmtxt("B", 2.5, 1.0-datestrcoord,1.0-0.5*datestrcoord,t->dateend);
       }
    }
}
/* data - this is the easy bit! */

cpgslw(t->line_width);
cpgsls(t->linestyle);
cpgsci(t->colour);
cpgsch(t->symbol_size);
if (t->YoNline == YES)cpgline(t->npts, x, y);
if (t->YoNsymbol == YES) cpgpt(t->npts, x, y, t->symbol);

/* legend */

if (t->YoNlegend == YES) PGXYtrace_legend(t);

if (t->xlinlog == PGSPECTRO_LOG && t->timex == NO) free(x);
if (t->ylinlog == PGSPECTRO_LOG) free(y);
cpgunsa();  /* restore state before call */
return PGSPECTRO_OK;
}

/*******************************************************************/
PGSpectro_ret PGXYtrace_legend(PGXYtrace *t){
/*******************************************************************/
/* routine to draw the legend for an XY trace plot */

float x[2], y[2], marg, len, xch, ych;

cpgsave();

cpgsci(1);; /* black */
cpgsls(QPGSOLID);
cpgsch(t->legend_text_char_size);

cpgsvp(t->legend_domain.xndcmin, t->legend_domain.xndcmax,
       t->legend_domain.yndcmin, t->legend_domain.yndcmax);

cpgswin(0.0, 1.0,0.0, 1.0);

if (t->YoNlegendbox == YES) {
   cpgslw(t->box_thickness);
   cpgbox("BC",0.0,0,"BC",0.0,0);
}

y[0]= y[1] = t->legend_just;
len = t->legend_line_length_frac;
marg = 0.1*len;
x[0] = marg;  x[1]=marg+len;

cpgslw(t->line_width);
cpgsls(t->linestyle);
cpgsci(t->colour);
if (t->YoNline == YES)cpgline(2, x, y);
if (t->YoNsymbol == YES) {
  cpgsch(t->symbol_size);
  cpgpt1( 0.5*(x[0]+x[1]), y[0], t->symbol);
}
cpgsci(1);
cpgslw(t->text_line_width);
cpgsch(t->legend_text_char_size);
cpgqcs(4,&xch,&ych);    /* get ch height in units of world coords */
cpgptxt(x[1]+marg, y[1]- 0.4*ych, 0.0, 0.0, t->legend_text);

cpgunsa();
return PGSPECTRO_OK;
}

/********************************************************************/
PGSpectro_ret PGXYtrace_set_defaults(PGXYtrace *trace){
/********************************************************************/
/* set defaults for as much as possible. Should be safe enough so that
*  will not crash if nothing is set other than data */

float ndcwidth;   /* will need this to make room for legend */

trace->npts=0;
PGSpectro_domain_set_defaults(&(trace->plot_domain));
PGSpectro_domain_set_defaults(&(trace->legend_domain));
ndcwidth =trace->plot_domain.xndcmax - trace->plot_domain.xndcmin;
trace->plot_domain.xndcmax = trace->plot_domain.xndcmin + 0.6*ndcwidth;
trace->legend_domain.xndcmin = trace->plot_domain.xndcmax + 0.1*ndcwidth;

trace->colour = QPGBLACK;
trace->linestyle = QPGSOLID;
trace->line_width = 1;
trace->YoNline = YES;
trace->YoNsymbol = NO;
trace->symbol = QPGFILLEDCIRCLE;
trace->symbol_size = 1;

trace->subtitle = " ";
trace->subsubtitle = " ";
trace->xlabel = " ";
trace->ylabel = " ";
trace->title = " ";
trace->Xtextlab_bot = YES;
trace->Ytextlab_left = YES;
trace->title_just = 0.5;
trace->title_disp = 2.0;
trace->subtitle_disp = 1.5;
trace->subsubtitle_disp = 0.5;

trace->label_char_size = 1.0;
trace->title_char_size = 1.5;
trace->text_line_width = 1;     /* cpgslw expects an int */

trace->YoNXnumlab = YES;
trace->YoNYnumlab = YES;
trace->Ynumlab_left = YES;
trace->Xnumlab_bot = YES;
trace->YoNbox = YES;
trace->xlinlog = PGSPECTRO_LIN;
trace->ylinlog = PGSPECTRO_LIN;
trace->tick_length = 1.0;
trace->box_thickness = 1;       /* cpgslw expects an int */

trace->timex = NO;
trace->datestart = " ";
trace->dateend   = " ";

trace->legend_text_char_size = 0.7;
trace->YoNlegend = YES;
trace->YoNlegendbox = YES;
trace->legend_just  = 0.5;
trace->legend_line_length_frac = 0.2;
trace->legend_text = " ";

return PGSPECTRO_OK;
}

/*****************************************************************/
double SafeLog10 (double x) {  /* from QppLog10 in qpp_panel.cc */
/*****************************************************************/
  return (x < 1.0e-33) ? -33.0 : log10(x);
}

#ifdef __cplusplus
}
#endif


/* end of PGSpectro.c file */
