/* Qidfs_utils.c ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  26 November 2000
*  6 March 2001 added set_times_from_idf routine
*  May 2001 - re-organised Qpeace routines
*  July 2001 - added stuff for sweep export
*  May 2002  - tidied PSD_UNITS_STRING with statics for safety
*
*  Utility routines for generic used, including:
*
*  Setting strings, idfs lineages, etc.
*  Setting conversion tables, getting energy bin limits, ...
*  Opeining and positioning idfs file
*  Manipulating time into format needed by and returned from idfs routines
*
*/

#ifdef __cplusplus
extern "C"{
#endif

#include "Qidfs_utils.h"

/**********************************************************************
****  IDFS Lineage and Strings ****************************************
**********************************************************************/

char *SC_NAME(SC sc){

/* these could be done in an array, but the following avoids the
*  need to be sure the enum order matches the order of an array */
static char *c1      = "CLUSTER-1";  /* statics: see K&R 5.8 */
static char *c2      = "CLUSTER-2";
static char *c3      = "CLUSTER-3";
static char *c4      = "CLUSTER-4";
static char *c1234   = "C1234";
static char *tc1     = "TC-1";
static char *tc2     = "TC-2";
static char *unknown = "UNKNOWN";

 switch(sc){
   case CLUSTER1: return c1;      break;
   case CLUSTER2: return c2;      break;
   case CLUSTER3: return c3;      break;
   case CLUSTER4: return c4;      break;
   case C1234:    return c1234;   break;
   case TC1:     return tc1;     break;
   case TC2:     return tc2;     break;
   default:       return unknown; break;
 }
}

char *PSD_UNITS_STRING(PSD_UNITS psdu){

static char *cnt_acc = "Counts / Accumm";
static char *def     = "D e-flux ergs/(cm**2 s sr eV)";
static char *psds3k6 = "Phase space density s**3/km**6";
static char *unknown = "UNKNOWN PSD UNITS";

 switch(psdu){
   case PSD_UNITS_CNTS_ACCUM:       return cnt_acc; break;
   case PSD_UNITS_DIFF_ENERGY_FLUX: return def;     break;
   case PSD_UNITS_S3_PER_KM6:       return psds3k6; break;
   default:                         return unknown; break;
 }
}

char *PSD_UNITS_SI_conv_STRING(PSD_UNITS psdu){

static char *cnt_acc = "1>(counts/accum)";
static char *def     = "6.2418e15>m^-2 s^-1 sr^-1";
static char *psds3k6 = "1.e-18>s^3 m^-6";
static char *unknown = "1>(unknown psd units)";

 switch(psdu){
   case PSD_UNITS_CNTS_ACCUM:       return cnt_acc; break;
   case PSD_UNITS_DIFF_ENERGY_FLUX: return def;     break;
   case PSD_UNITS_S3_PER_KM6:       return psds3k6; break;
   default:                         return unknown; break;
 }
}


char *BIN_UNITS_STRING(BIN_UNITS bin_units){

static char *ev       = "Energy eV";
static char *kms      = "Speed km/s";
static char *unknown  = "Unknown Energy Units";

 switch (bin_units){
   case BIN_UNITS_EV:   return ev;      break;
   case BIN_UNITS_KM_S: return kms;     break;
   default:             return unknown; break;
 }
}

char *BIN_UNITS_SI_conv_STRING(BIN_UNITS bin_units){

static char *ev       = "1.6021e-19>J";
static char *kms      = "1.e3>m s^-1";
static char *unknown  = "1>(UNKNOWN ENERGY BIN UNITS)";

 switch (bin_units){
   case BIN_UNITS_EV:   return ev;      break;
   case BIN_UNITS_KM_S: return kms;     break;
   default:             return unknown; break;
 }
}


/************************************************************************/
void print_idfs_lineage(IDFS_LINEAGE *lng){
/************************************************************************/

printf("IDFS Lineage Strings are:\n");
   printf("project |%s|\n",lng->project);
   printf("mission |%s|\n",lng->mission);
   printf("experiment |%s|\n",lng->experiment);
   printf("instrument |%s|\n",lng->instrument);
   printf("virtual_inst |%s|\n",lng->virtual_inst);
   printf("exten |%s|\n",lng->exten);
return;
}

/*****************************************************************/
void copy_lineage(IDFS_LINEAGE *lincpy, IDFS_LINEAGE *lin){
/*****************************************************************/

strcpy(lincpy->project,lin->project);
strcpy(lincpy->mission,lin->mission);
strcpy(lincpy->experiment,lin->experiment);
strcpy(lincpy->instrument,lin->instrument);
strcpy(lincpy->virtual_inst,lin->virtual_inst);
strcpy(lincpy->exten,lin->exten);
lincpy->vsn = lin->vsn;

return;
}


/**********************************************************************
****  IDFS Time  ******************************************************
**********************************************************************/

/**********************************************************************/
void to_idfs_time(int year, int doy, int hr,
               int min, int sec, long nano, IDFS_TIME *idftp){
/*************************************************************************/
/* convert a time to format expected by idfs routines */
/* at present just a dumb routine which takes int
* year, doy, hr, min, sec*/

idftp->year = year;
idftp->doy  = doy;
idftp->sec  = 60*(60*hr + min)+sec;
idftp->nano = nano;

return;
}

/*****************************************************************************/
void sec_nano_to_msec_nano(long sec, long nanos, long *msec, long *nanoms){
/*****************************************************************************/

*msec =1000*sec + (nanos/1000000);
*nanoms = nanos % 1000000;
return;
}

/***************************************************************************/
void msec_nano_to_sec_nano(long msec, long nanoms, long *sec, long *nanos){
/**************************************************************************/

*sec = msec/1000;
*nanos = 1000000*(msec % 1000) + nanoms;
return;
}

/***************************************************************/
void to_hms_time(IDFS_TIME *idftp, int *hp, int *mp, int *sp, long *np){
/***************************************************************/
/* and an inverse dumb routine. Will use QDOS Time soon */
*hp = idftp->sec/3600;
*mp = idftp->sec/60 - *hp*60;
*sp = idftp->sec - 60*(*mp + 60**hp);
*np = idftp->nano;

return ;
}

void print_idfs_time(IDFS_TIME *idftp){
int hr, min, sec;
long nano;
to_hms_time(idftp, &hr, &min, &sec, &nano);
printf("%4d/%3d %2d:%2d:%2d (%ld secs in day) and %ld nanosecs",
         idftp->year, idftp->doy, hr, min, sec, idftp->sec, nano);
return;
}

/******************************************************************/
IDFS_TIME t1_minus_t2(IDFS_TIME *t1, IDFS_TIME *t2){
/******************************************************************/

/* find difference between two IDFS_TIMES
*  since no field can overflow, there is no
*  need to do anything other than difference the
*  independent members
*******************************************/

IDFS_TIME dt_idfs, *dtp=&dt_idfs;

dtp->year = t1->year - t2->year;
dtp->doy  = t1->doy  - t2->doy;
dtp->sec  = t1->sec  - t2->sec;
dtp->nano = t1->nano - t2->nano;

return  dt_idfs;
}

/******************************************************************/
IDFS_TIME_COMPARE compare_idfs_time(IDFS_TIME *t1, IDFS_TIME *t2){
/******************************************************************/

IDFS_TIME dt_idfs;

dt_idfs = t1_minus_t2(t1,t2);

if (dt_idfs.year > 0) return T1_LATER;
   else if (dt_idfs.year < 0 ) return T2_LATER;
/* years same; check days */
if (dt_idfs.doy > 0) return T1_LATER;
   else if (dt_idfs.doy < 0 ) return T2_LATER;
/* days same; check seconds */
if (dt_idfs.sec > 0) return T1_LATER;
   else if (dt_idfs.sec < 0 ) return T2_LATER;
/* secs same; check nanosecs */
if (dt_idfs.nano > 0) return T1_LATER;
   else if (dt_idfs.nano < 0 ) return T2_LATER;
/* nanos same; must be same time */

return SAME_TIME;
}

/* Qan-based routines follow. SJS added () around leap year && || ops
*  and constructed QanDateToYearDoy from QanEpochToYearDoy */ 


/* daytab: number of days in month for normal and leap years 
**         index 1 is Jan 
*/

int qidf_days_in_month[2][13] = {
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};


void QanYearDoyToDate( int year, int doy, int* month, int* day )
{
  int i, leap;
  leap = ( (year%4 == 0 && year%100 != 0) || year%400 == 0 );
/*
** check that doy is in range for supplied year
*/
  if( (leap && doy > 366 ) || (!leap && doy > 365) || doy < 1 )
  {
    *month = -1; *day = -1;
    return;
  }
  
  for(i=1; doy > qidf_days_in_month[leap][i]; i++ )
     doy -= qidf_days_in_month[leap][i];
  *month = i;
  *day = doy;
}


void QanDateToYearDoy(int year, int month, int day, int *doy)
{
  int leap_year, i;
  leap_year = (( year % 4 == 0 ) && ( year % 100!=0 )) || ( year % 400==0 );
  if (month < 1 || month > 12){
     *doy = -1;
     return;
  }
  for( *doy=0, i=0; i < month ; i++)
    *doy += qidf_days_in_month[leap_year][i];
  *doy += day;
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++  QEpochSecs  ++
*** Based on:
*** QanComputeEpoch: source taken from CDF 2.4.11 source (epochu.c) 
***/
/******************************************************************************
* Local function prototypes.
******************************************************************************/

static long JulianDay( long year, long month, long day );

/******************************************************************************
* JulianDay.
*    The year, month, and day are assumed to have already been validated.  This
* is the day since 0 AD (julian day may not be the proper term).
******************************************************************************/

static long JulianDay( long y, long m, long d )
{
  long jd;
  jd = (long)
       (367*y-7*(y+(m+9)/12)/4-3*((y+(m-9)/7)/100+1)/4+275*m/9+d+1721029);
  return (jd);
}

/******************************************************************************
* computeEPOCH.
*   Computes (and returns) an EPOCH in **SECS** since 0AD 
* based on its component parts.  -1.0
* is returned if an illegal component part is detected.
double computeEPOCH (year, month, day, hour, minute, second, msec)
long year, month, day, hour, minute, second, msec;
******************************************************************************/

double QEpochSecs( int year, int month, int day,
  int hour, int minute, int second, int msec )
{
  double epoch;
/*  printf("QEpochSecs got %d %d %d %d %d %d %d\n", year, month, day, hour,
           minute, second, msec);
*/
  if ((day < 1 || day > 31) ||
      (month < 1 || month > 12) ||
      (year < 0 || year > 9999) ||
      (hour < 0 || hour > 23) ||
      (minute < 0 || minute > 59) ||
      (second < 0 || second > 59) ||
      (msec < 0 || msec > 999))
    epoch = -1.0;
  else {
    epoch = (double) JulianDay (year, month, day) - 1721060;
    epoch = epoch * 24.0 + hour;
    epoch = epoch * 60.0 + minute;
    epoch = epoch * 60.0 + second;
    epoch = epoch + 0.001*msec;
  }
  return epoch;
}


/******************************************************************/
double idfs_time_to_epoch_secs(IDFS_TIME *idft){
/******************************************************************/
/* convert an idfs time to secs since 0AD; can't get to nano accuracy*/

int month, day, hr, min, sec, msec;
long ns;
QanYearDoyToDate(idft->year, idft->doy, &month, &day);
to_hms_time(idft, &hr, &min, &sec, &ns);
msec = (int) ((double) idft->nano)/((double) 1000000.0);
return QEpochSecs(idft->year, month, day, hr, min, sec, msec);
} 

/*****************************************************************/
void epoch_secs_to_idfs_time(double epoch, IDFS_TIME *idftp){
/****************************************************************/
/* convert epoch secs since 0AD to idfs time; based on cdf
*  EPOCHbreakdown, but works in secs
*
*  Only accurate down to msec level */
  long jd,i,j,k,l,n;
  double msec_AD, second_AD, minute_AD, hour_AD, day_AD;
  int year, doy, hour, min, sec;
  long nano;

  if (epoch<=0.0) {   /* just give up here */
    to_idfs_time(0,0,0,0,0,0L,idftp);
    return;
  }

  msec_AD    = 1000.0 * epoch;
  second_AD  = epoch;
  minute_AD  = epoch/60.0;
  hour_AD    = minute_AD/60.0;
  day_AD     = hour_AD / 24.0;

  jd = (long) (1721060 + day_AD);
  l  =  jd+68569;
  n  =  4*l/146097;
  l  =  l-(146097*n+3)/4;
  i  =  4000*(l+1)/1461001;
  l  =  l-1461*i/4+31;
  j  =  80*l/2447;
  k  =  l-2447*j/80;                   /* day   */
  l  =  j/11;
  j  =  j+2-12*l;                      /* month */
  i  =  100*(n-49)+i+l;                /* year  */

  year = (int)i;
  QanDateToYearDoy(year,j,k,&doy);
  
  hour   = (int) fmod (hour_AD, (double) 24.0);
  min    = (int) fmod (minute_AD, (double) 60.0);
  sec    = (int) fmod (second_AD, (double) 60.0);
  nano   = (long) ((double) 1000000.0* fmod (msec_AD, (double) 1000.0));
  
  to_idfs_time(year, doy, hour, min, sec, nano, idftp);

  return;
}

/****************************************************************/
void add_secs_to_idfs_time(IDFS_TIME *idftp, float secs2add){
/****************************************************************/
/* use epoch routines to add (or subtract) secs2add from idftp */

epoch_secs_to_idfs_time( idfs_time_to_epoch_secs(idftp) + secs2add,
                   idftp);
return;
}


/*****************************************************************/
void set_times_from_idf(IDFS_TIME *btime, IDFS_TIME *etime,
                        IDFS_TIME *sun_sen, struct idf_data *idfp){
/*****************************************************************/
/* set start time, end time and sun sensor times by looking at
*  a (filled) idf_data structure
*
*  returned in the individual IDFS_TIME structures */

btime->year = idfp->byear;   etime->year = idfp->eyear;
btime->doy  = idfp->bday;    etime->doy  = idfp->eday;
msec_nano_to_sec_nano(idfp->bmilli, idfp->bnano,
                         &(btime->sec), &(btime->nano));
msec_nano_to_sec_nano(idfp->emilli, idfp->enano,
                         &(etime->sec), &(etime->nano));
sun_sen->year = btime->year;
sun_sen->doy = btime->doy;
msec_nano_to_sec_nano(idfp->sun_sen, 0L, &(sun_sen->sec), &(sun_sen->nano));
return;
}

/*********************************************************************
****** IDFS File and data structure utils ****************************
*********************************************************************/

/**********************************************************************/
SDDAS_SHORT init_idf_file(IDFS_TIME btime, IDFS_TIME etime, SDDAS_ULONG dkey,
     struct idf_data *idfp, SDDAS_USHORT vsn){
/*********************************************************************/
/* open and position an idfs file */
SDDAS_SHORT ret_val;
SDDAS_CHAR exten[5];

strcpy(exten,"");

ret_val = file_open(dkey,exten, vsn,btime.year,btime.doy, btime.sec, btime.nano,
              etime.year, etime.doy, etime.sec, etime.nano, 0);
if (ret_val != ALL_OKAY){
   printf ("PEACE: Error opening idfs file\n"); return QUTILS_ERROR;
}

ret_val = file_pos ( dkey,exten, vsn,idfp,btime.year, btime.doy, btime.sec, btime.nano,
              etime.year,etime.doy, etime.sec, etime.nano);
if (ret_val != ALL_OKAY){
   printf ("PEACE: Error positioning idfs file\n"); return QUTILS_ERROR;
}

if(DIAG_PRINT){
   printf("init_idf_file ok. Start and stop times sent to file_pos are\n");
   print_idfs_time(&btime);printf("\n");
   print_idfs_time(&etime);printf("\n");
}

return ALL_OKAY;
}

/************************************************************************/
SDDAS_SHORT pos_idf_file_spin_start(SDDAS_ULONG dkey, struct idf_data *idfp, 
         IDFS_LINEAGE *lineage, IDFS_TIME *btime, IDFS_TIME *etime){
/*************************************************************************/
/* Use information in idf_data structure to re-position the file associated
*  with dkey to its spin start and return the spin start/stop times.
*/

float tspin;
SDDAS_SHORT ret_val;

tspin = idfp->spin_rate/1000.;
btime->year = idfp->byear;
btime->doy  = idfp->bday;
msec_nano_to_sec_nano(idfp->sun_sen, 0L, &(btime->sec), &(btime->nano));
etime->year = btime->year;
etime->doy  = btime->doy;
etime->sec  = btime->sec + 1.e-9*btime->nano + tspin ; /*trunc tspin to int*/
etime->nano = 1000000000*(btime->sec +1.e-9*btime->nano+tspin -etime->sec);

ret_val = reset_experiment_info(dkey,lineage->exten,lineage->vsn,
               btime->year, btime->doy, btime->sec, btime->nano,
	       etime->year, etime->doy, etime->sec, etime->nano);
if (ret_val != ALL_OKAY){
   printf("pos_idf_file_spin_start reset_exp returned %d\n", ret_val);
   return ret_val;
}

ret_val = file_pos (dkey, lineage->exten, lineage->vsn, idfp,
               btime->year, btime->doy, btime->sec, btime->nano,
	       etime->year, etime->doy, etime->sec, etime->nano);
if (ret_val != ALL_OKAY){
   printf("pos_idf_file_spin_start file_pos returned %d\n", ret_val);
   return ret_val;
}
 
return ALL_OKAY;
}       

/************************************************************************/
SDDAS_SHORT advance_idf_file_spin_start(SDDAS_ULONG dkey, struct idf_data *idfp, 
         IDFS_LINEAGE *lineage, IDFS_TIME *btime, IDFS_TIME *etime,
	 SDDAS_SHORT sensor){
/*************************************************************************/
/* Use information in idf_data structure to ADVANCE the file associated
*  with dkey to the NEXT spin start and return the sweep start/stop times
*  for the 1st sweep in the spin.
*  Reads data for "sensor", so pass a sensor number which is always filled
*  (or which you need to be filled)
*
*  Do a non-forwarding read 1st to see if file was forwarded on last
*  read.
*
*  Guard against files which have (some) records out of order
*/

SDDAS_SHORT ret_val;
IDFS_TIME suntime1,suntime2;
IDFS_TIME_COMPARE compare;

set_times_from_idf(btime,etime,&suntime1,idfp);

ret_val = read_drec (dkey, lineage->exten, lineage->vsn, 
	           idfp, sensor,0, 1); /* read sens , no fwd time */
if(ret_val != ALL_OKAY){
  printf("1st read_drec in advance_idf_file_spin_start failed returning %d\n",
                     ret_val);
  return ret_val;
}
set_times_from_idf(btime,etime,&suntime2,idfp);

compare = compare_idfs_time(&suntime1,&suntime2);

switch(compare){
     case T2_LATER:
          return ALL_OKAY;/* file was forwarded at last read and is already */
                          /* at beginning of next, so return */
          break;
     case T1_LATER:  
          printf("advance_idf_file_spin_start confused. Idf time on entry = ");
          print_idfs_time(&suntime1);
          printf("\nTime after initial read is EARLIER= ");
          print_idfs_time(&suntime2);printf("\nTrying to advance anyway\n");
	  compare = SAME_TIME; /* fool it, to cope with non-monotonic files*/
	  break;
      case SAME_TIME: default:
          break;
}

while ( compare != T2_LATER ){
    
   ret_val = read_drec (dkey, lineage->exten, lineage->vsn, 
	           idfp, sensor,1, 1); /* read sens , fwd time */
   if(ret_val != ALL_OKAY){
       printf("read_drec fwd advance_idf_... failed returning %d\n",
                     ret_val);
       return ret_val;
   }
   ret_val = read_drec (dkey, lineage->exten, lineage->vsn, 
	           idfp, sensor,0, 1); /* read sens , no fwd time */
   if(ret_val != ALL_OKAY){
       printf("read_drec no fwd advance_idf... failed returning %d\n",
                     ret_val);
       return ret_val;
   }
   set_times_from_idf(btime,etime,&suntime2,idfp);
   compare = compare_idfs_time(&suntime1,&suntime2);
}
 
return ALL_OKAY;

}       

/**********************************************************/
void print_idf_data(struct idf_data *ip){
/*********************************************************/

/* print out a bunch of info from idf_data structure */

int i;

printf("\nInformation from idf_data structure for Sensor %d\n", ip->sensor);
printf("Start year day milli nano End year day milli nano\n");
printf("%d %d %ld %ld %d %d %ld %ld \n",
              ip->byear, ip->bday, ip->bmilli, ip->bnano,
              ip->eyear, ip->eday, ip->emilli, ip->enano);

printf("num_sample %d, num_angle %d, cal_len %ld num_swp_steps %d num_pitch %d\n",
              ip->num_sample, ip->num_angle, ip->cal_len, 
	      ip->num_swp_steps, ip->num_pitch);

printf("sun_sen %ld spin_rate %ld\n", ip->sun_sen, ip->spin_rate);
printf("start_theta = %g stop_theta = %g for sensor %d\n",
            ip->start_theta, ip->stop_theta, ip->sensor);

printf("\n i_angle Start Az   Stop Az\n");
for (i=0; i<ip->num_angle; i++){
    printf("%d %g %g\n",i, ip->start_az[i], ip->stop_az[i]);
}

printf("\n i_swp_step swp_data pairs:");
for (i=0; i<ip->num_swp_steps; i++){
   if(i % 6 == 0) printf("\n");
   printf(" (%d %ld)",i,ip->swp_data[i]);
}

printf("\n i_cal cal_data\n");
for (i=0; i<ip->cal_len; i++){
    printf("%d  %ld\n",i, ip->cal_data[i]);
}

printf("\nRaw Sweep Data\n");
for (i=0; i<ip->num_sample; i++){
    if (i%10 == 0)printf("\n");
    printf("%8ld",ip->sen_data[i]);
}

return;
}

/**********************************************************************************/
SDDAS_SHORT  e_bin_edges(SDDAS_ULONG dkey,struct idf_data *idfp,SDDAS_USHORT vsn,
      float *bin_low,float *bin_high, BIN_UNITS bin_units, int n_lev_array_len){
/********************************************************************************/
/* get the energy bin lower and upper edges; convert to requested bin_units
*
*  see set_psd_tables below for table requirements copied from VIDF*/

SDDAS_SHORT ret_val;
SDDAS_SHORT tbl_oper[4];
SDDAS_CHAR num_tbls, tbls_to_apply[4];
int j;

if (idfp->num_swp_steps > n_lev_array_len){            /* make sure enough room */
   printf("More sweep steps than size of array to hold them\n");
   return QUTILS_ERROR;
}

if (sizeof (SDDAS_FLOAT) != sizeof (float)){
   printf("SDDAS_FLOAT and float different sizes; bin_edges failure\n");
   return QUTILS_ERROR;
}

tbl_oper[0] = 0;  /* first operation always 0 */

switch (bin_units){
  case BIN_UNITS_KM_S:
     num_tbls = 4;
     tbls_to_apply[1] = 5;    tbl_oper[1] = 3;
     tbls_to_apply[2] = 11;   tbl_oper[2] = 63;
     tbls_to_apply[3] = 13;   tbl_oper[3] = 3;
     break;
  case BIN_UNITS_EV:
  default:
     num_tbls = 1;
     break;
}


tbls_to_apply[0] = 9; /* lower sweep edges */
tbl_oper[0] = 0;

ret_val = convert_to_units (dkey, "", vsn, idfp,idfp->sensor, SWEEP_STEP, 0,
               num_tbls, tbls_to_apply, tbl_oper, bin_low, 0, 0);
if (ret_val != ALL_OKAY){
   printf("Error getting bin low edge.\n"); return QUTILS_ERROR;
}

tbls_to_apply[0] = 10;  /* upper sweep edges */

ret_val = convert_to_units (dkey, "", vsn, idfp,idfp->sensor, SWEEP_STEP, 0,
               num_tbls, tbls_to_apply, tbl_oper, bin_high, 0, 0);
if (ret_val != ALL_OKAY){
   printf("Error getting bin upper edge.\n"); return QUTILS_ERROR;
}

if (DIAG_PRINT_EBINS){
  printf("sample   swp_step   elow   ehigh within e_bin_edges\n");
  for (j=0; j< idfp->num_swp_steps; j++){
    printf("%4d     %4ld  %10g   %10g\n",j, idfp->swp_data[j],
                   bin_low[j], bin_high[j]);
  }
}


return ALL_OKAY;
}


/******************************************************************/
SDDAS_SHORT set_psd_tables(IDFS_LINEAGE *lgp, PSD_UNITS psd_units, 
     SDDAS_CHAR *num_tbls,  SDDAS_CHAR *tbls_to_apply, SDDAS_SHORT *tbl_oper){
/*******************************************************************/

/* Double star tables to apply differ in that there is no table zero
*  to convert from telemtery to counts */

/*
From VIDF for CPPADL [Cluster II original VIDF]
Copied 22 May 2001

t  The following are some of the units which can be derived from the   
t  included tables. The format is to give the tables applied followed 
t  by the operations and unit definition.  Note that for brevity we    
t  represent the table sequence 0,2,3,4 by T_S1 and the operation     
t  sequence 0,154,4,4 by O_S1.                                        
t                                                                     
t  DATA  SEN  TABLES           OPERS             UNITS                
t  Scan  all  1                0                 eV                    
t  Scan  all  1,5,11           0,3,63            cm/sec                
t  Scan  all  1,5,11,12        0,3,63,3          m/sec                 
t  Scan  all  1,5,11,13        0,3,63,3          km/sec                
t  Scan  all  9                0                 eV (lower edge)       
t  Scan  all  10               0                 eV (upper edge)       
t  Sen   all  0                0                 cnts/accum            
t  Sen   all  0,2              0,4               cnts/accum (eff. cor) 
t  Sen   all  0,2              0,154             cnts/sec              
t  Sen   all  T_S1             O_S1              cnts/(cm**2-str-s)    
t  Sen   all  T_S1,1           O_S1,4            cnts/(cm**2-str-s-eV) 
t  Sen   all  T_S1,5           O_S1,3            ergs/(cm**2-str-s-eV) 
t  Sen   all  T_S1,6,5,1,5,1   O_S1,3,4,4,4,4    sec**3/km**6          
t  Sen   all  T_S1,6,5,1,5,1,8 O_S1,3,4,4,4,4,3  sec**3/m**6           
t  Sen   all  T_S1,6,5,1,5,1,7 O_S1,3,4,4,4,4,3  sec**3/cm**6          
*/


if (strncmp(lgp->project, "CLUST",5)==0){  /* cluster */

    switch (psd_units){

       case PSD_UNITS_S3_PER_KM6:
	  *num_tbls = 9;
	  tbls_to_apply[0] = 0;    tbl_oper[0] = 0;
	  tbls_to_apply[1] = 2;    tbl_oper[1] = 154;
	  tbls_to_apply[2] = 3;    tbl_oper[2] = 4;
	  tbls_to_apply[3] = 4;    tbl_oper[3] = 4;
	  tbls_to_apply[4] = 6;    tbl_oper[4] = 3;
	  tbls_to_apply[5] = 5;    tbl_oper[5] = 4;
	  tbls_to_apply[6] = 1;    tbl_oper[6] = 4;
	  tbls_to_apply[7] = 5;    tbl_oper[7] = 4;
	  tbls_to_apply[8] = 1;    tbl_oper[8] = 4;
	  break;

       case PSD_UNITS_DIFF_ENERGY_FLUX:
	  *num_tbls = 5;
	  tbls_to_apply[0] = 0;    tbl_oper[0] = 0;
	  tbls_to_apply[1] = 2;    tbl_oper[1] = 154;
	  tbls_to_apply[2] = 3;    tbl_oper[2] = 4;
	  tbls_to_apply[3] = 4;    tbl_oper[3] = 4;
	  tbls_to_apply[4] = 5;    tbl_oper[4] = 3;
	  break;

       case PSD_UNITS_CNTS_ACCUM:   /* uncorrected */
       default:
	  *num_tbls = 1;
	  tbls_to_apply[0] = 0;    tbl_oper[0] = 0;
	  break;
    }

} else if(strncmp(lgp->project, "DOUBLEST", 8)==0){ /* double star */

    switch (psd_units){

       case PSD_UNITS_S3_PER_KM6:
	  *num_tbls = 8;
	  tbls_to_apply[0] = 2;    tbl_oper[0] = 154;
	  tbls_to_apply[1] = 3;    tbl_oper[1] = 4;
	  tbls_to_apply[2] = 4;    tbl_oper[2] = 4;
	  tbls_to_apply[3] = 6;    tbl_oper[3] = 3;
	  tbls_to_apply[4] = 5;    tbl_oper[4] = 4;
	  tbls_to_apply[5] = 1;    tbl_oper[5] = 4;
	  tbls_to_apply[6] = 5;    tbl_oper[6] = 4;
	  tbls_to_apply[7] = 1;    tbl_oper[7] = 4;
	  break;

       case PSD_UNITS_DIFF_ENERGY_FLUX:
	  *num_tbls = 4;
	  tbls_to_apply[0] = 2;    tbl_oper[0] = 154;
	  tbls_to_apply[1] = 3;    tbl_oper[1] = 4;
	  tbls_to_apply[2] = 4;    tbl_oper[2] = 4;
	  tbls_to_apply[3] = 5;    tbl_oper[3] = 3;
	  break;

       case PSD_UNITS_CNTS_ACCUM:   /* uncorrected */
       default:
	  *num_tbls = 0;
	  break;
    }


} else {
   printf ("set_psd_tables project = %s not recognised\n",lgp->project);
   return QUTILS_ERROR;
}

return ALL_OKAY;
}


/**************************************************************************/
INTERP_QUANTITY diff_or_integral(PSD_UNITS psu){
/**************************************************************************/

switch(psu){
   case PSD_UNITS_CNTS_ACCUM:
     return INTEGRAL;
     break;
   case PSD_UNITS_S3_PER_KM6:
   case PSD_UNITS_DIFF_ENERGY_FLUX:
     return DIFFERENTIAL;
     break;
   default:
     printf("Unknown PSD_UNITS:diff_or_integral defaults to differential\n");
     return DIFFERENTIAL;
     break;
}

return DIFFERENTIAL; /* shouldn't get here! */
}


/************************************************************************
*************** Miscellaneous Utilities *********************************
************************************************************************/

/*************************************************************************/
SDDAS_SHORT invert_array(float *arr, int n_lev){
/**************************************************************************/

/* invert order of array of floats */

float temp;
int i,n;

n=n_lev/2;

for (i=0,n_lev--; i<n; i++,n_lev--){
   temp = arr[i];
   arr[i] = arr[n_lev];
   arr[n_lev] = temp;
}

return ALL_OKAY;
}

/*****************************************************************/
double near_nice_number(double x){
/*****************************************************************/

/* find a nice number near to x. Nice means 1,2, or 5 times power of 10
*  The Nice number is always larger in magnitude than x
*
*  This mimics the pgplot routine pgrnd, which doesn't seem to work
*  on all machines across the c interface */

double xlog, r, temp;
int pow_of_10, pos,inum;

    pos = (x>0);                /* for x < 0, need to strip and keep sign info */
    xlog = SafeLog10(fabs(x) );
    pow_of_10 = (int)xlog;      /* find nearest power of 10; if < 0, dec by 1 */
    if (pow_of_10 <= 0) pow_of_10--;
    temp = xlog - pow_of_10;    /* temp is now log of number part only */
    if (temp > 1){temp -= 1; pow_of_10++;}  /* fudge for vals 0.1-0.9*/
    r = pow(10,temp);
    inum = (int) r; /* inum is the integer number part */
    if ( r/inum > 1.0) inum++;
    if (inum > 5) inum = 10;
    else if (inum > 2) inum = 5;
    else if (inum > 1) inum = 2;
    else inum = 1;
    r = inum*pow(10, pow_of_10);/* use integer number times power of 10 */
    if (!pos) r *= -1;          /* restore sign */

 return r;

}

/*****************************************************************/
void scale_then_rotate_angles(float *orig, float *rotscale, int n,
                         float rotation, float scale){
/*****************************************************************/

/* first scale, then rotate angles:
*   rotscale[i] = (scale * orig ) + rotation */

int i;

for (i=0; i< n; i++){
   rotscale[i] =  (scale * orig[i]) + rotation;
}

return;

}

/***********************************************************************/
float polar_area(float r1, float r2, float th1_deg, float th2_deg){
/***********************************************************************/
return 0.5*(r2*r2 - r1*r1)*(cos(DEG2RAD*th1_deg) - cos(DEG2RAD*th2_deg));

}

/***********************************************************************/
int find_subset_in_range(float xlo, float xhi,
          float *xloarray, float *xhiarray, int n_array,
	  int *i1, int *i2){
/***********************************************************************/
/* find the array indices i1 and i2 such that segments
*  ( xloarray[i1], xhiarray[i1] )  -> ( xloarray[i2], xhiarray[i2] )
*  all overlap with the range (xlo, xhi)
*
*  The input arrays xloarray, xhiarray are assumed to be sorted in
*  ascending order and xlo <= xhi
*
*  Returns the number of overlapping ranges found and, via i1 and i2,
*  the indices. Note: Calling routine MUST check number returned != 0
*  prior to using i1, i2;
*/

int ii1, ii2, nm1;

nm1 = n_array-1;

if (xloarray[0] >= xhi || xhiarray[nm1] <= xlo) return 0;

ii1 = 0; /* work up from bottom until top of bin > xlo */
while (xhiarray[ii1] <= xlo && ii1 < nm1) ii1++;

ii2 = nm1; /* work down from top until bottom of bin < xhi */
while (xloarray[ii2] >= xhi && ii2 > 0) ii2--;

*i1 = ii1;
*i2 = ii2;
return ii2 -ii1 + 1; /* nb: if (xlo,xhi) lies entirely inbetween two bins */
                   /* above algorithm results in i2 = i1 -1 so routine  */
		   /* returns 0 as required */

}

/*************************************************************************/
SDDAS_SHORT check_bins(float *lowbins, float *hibins, int nbins){
/*************************************************************************/

/* check bins. If lowbin[i] = hibin[i] then set mid-values */
/* this is a fudge to cope with pathological VIDFs, e.g., initial
*  ones. Note that bins are assumed to be given in ascending order,
*  i.e., lowbins[i] < lowbins[i+1]*/

int i;

int ndiff = 0;
float mid;

for (i=0; i<nbins; i++){
  if (lowbins[i] == hibins[i]) ndiff++;
}
if (!ndiff) return ALL_OKAY;

/* some bins identical */
if(DIAG_PRINT)printf("Identical bins found; shifting to mid-values\n");

for (i=0; i< nbins-1; i++){
   mid = (hibins[i] + hibins[i+1])/2.0;
   lowbins[i+1] = mid;
   hibins[i]  = mid;
}
return ALL_OKAY;
}

/************************************************************************/
void print_bins(float *low, float *high, int n){
/***********************************************************************/

int i;
printf("-------------------------------------------\n");
printf("bin      low   -> high\n");
for (i=0; i<n; i++){
   printf("%d    %g    %g\n",i,low[i],high[i]);
}
return;
}


/************************************************************************/
SDDAS_SHORT set_equal_bins(float *xlobins, float *xhibins, int nbins, 
                           float xmin, float xmax){
/************************************************************************/
/* set nbins equal size bins from xmin to xmax */

float bin_size;
int i;

if (nbins < 1) return PEACE3D_ERROR;

bin_size = (xmax - xmin)/(float)nbins;

xlobins[0] = xmin;
xhibins[0] = xlobins[0] + bin_size;
for (i=1; i<nbins; i++){
  xlobins[i] = xhibins[i-1];
  xhibins[i] = xlobins[i]+bin_size;
}
xhibins[nbins -1] = xmax;  /* not necessary, but forces to eliminate any */ 
                           /* accumulated rounding error                 */
return ALL_OKAY;
}

/************************************************************************/
SDDAS_SHORT set_equal_bins_linlog(float *xlobins, float *xhibins, int nbins, 
                           float xmin, float xmax, PGSpectro_linlog linlog){
/************************************************************************/
/* set nbins equal size bins from xmin to xmax */

float bin_size,bin_factor;
int i;

if (nbins < 1) return PEACE3D_ERROR;

xlobins[0] = xmin;

switch(linlog){
   case PGSPECTRO_LOG:
       bin_factor = (SafeLog10(xmax)-SafeLog10(xmin))/(float)nbins;
       bin_factor = pow(10.0,bin_factor);
       xhibins[0] = xlobins[0]*bin_factor;
       for (i=1; i<nbins; i++){
          xlobins[i] = xhibins[i-1];
	  xhibins[i] = xlobins[i]*bin_factor;
       }
       break;
   case PGSPECTRO_LIN: default:
       bin_size = (xmax - xmin)/(float)nbins;

       xhibins[0] = xlobins[0] + bin_size;
       for (i=1; i<nbins; i++){
	 xlobins[i] = xhibins[i-1];
	 xhibins[i] = xlobins[i]+bin_size;
       }
       break;
}

xhibins[nbins -1] = xmax;  /* not necessary, but forces to eliminate any */ 
                           /* accumulated rounding error                 */
return ALL_OKAY;
}

/************************************************************************/
SDDAS_SHORT set_equal_bins_half_ends(float *xlobins, float *xhibins, int nbins, 
                           float xmin, float xmax){
/************************************************************************/
/* set equal size bins from xmin to xmax with end bins half size of rest */

float bin_size;
int i;

if (nbins < 1) return PEACE3D_ERROR;
if (nbins == 1) return set_equal_bins(xlobins, xhibins, nbins, xmin, xmax);

bin_size = (xmax - xmin)/(float)(nbins-1);

xlobins[0] = xmin; xhibins[0] = xlobins[0] + 0.5*bin_size;
for (i=1; i<nbins; i++){
  xlobins[i] = xhibins[i-1];
  xhibins[i] = xlobins[i]+bin_size;
}
xhibins[nbins -1] = xmax;  /* forces top bin to be half size */
return ALL_OKAY;
}

/*************************************************************************/
float mod360(float x){
/*************************************************************************/
/* return number modulo 360 */
float y;
y= x - 360.0*( (int)(x/360.0));   /* get to -360 < y < +360 */
if (y >= 0) return y;
else return y+360.0;
}


#ifdef __cplusplus
}
#endif
