/* Qsweep_spin.c ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  May 2001 - re-organised Qpeace routines and created this module from others
*  16 June 2001 Added GetNSweeps routine for qjas and other apps
*  20 June 2001 Re-think on lineage; now lineage treated together with
*     bin boundaries, so that:
*
*     spin structures: 1 copy of bins + lineage structure per spin
*     separate sweep structures: hold indepedent copy of bins + lineage
* 
*  27 July 2001 add nswps exporting routines to dump to QIE cdf structure
*     Jan  2002 up-dated for Cluster Exchange Format
*     Jan   2002 Add promote; incomplete as new IDFS with libdbSQL... needed
*                This affects dbInitialise here, in Qpeace_pad, and includes
*                in Qidfs_utils.h
*     May 2002  Re-order theta bins for Spinpad to be look directions, which
*                also makes theta bins ascending in theta, as needed by
*                CombineSweeps, etc. routines. Move read routines to separate
*                file. Add rebinning to pitch angles.
*
*  Routines for  holding, and manipulating PEACE sweeps and spins
*
*/

#ifdef __cplusplus
extern "C"{
#endif

#include "Qidfs_utils.h"
#include "Qsweep_spin.h"
#include "Qpeace_pad.h"
#include <float.h>     /* for FLT_MIN */

/********************************************************************/
char * qpeace_version_str(void){
/********************************************************************/
 static char str[80];
  strcpy(str, QPEACE_VERSION);
  return &(str[0]);
}

/********************************************************************/
SDDAS_SHORT malloc_sweep_arrays(PEACE_SWEEP *swpp, T_OR_F do_pa){
/********************************************************************/
/* Malloc:
*
*  thlo, thhi angle bins
*  elo, ehi   energy bins
*  psd, pa bins
*  lineage structure
*
*  Dimensions should already be in sweep structure */

int bad_malloc = 0;
int ns, nl;

ns = swpp->n_sen;
nl = swpp->n_lev;

nullify_sweep_arrays(swpp);

if((swpp->thlo = (float *)malloc (ns * sizeof(float))) == NULL)bad_malloc++;
if((swpp->thhi = (float *)malloc (ns * sizeof(float))) == NULL)bad_malloc++;
if((swpp->elo  = (float *)malloc (nl * sizeof(float))) == NULL)bad_malloc++;
if((swpp->ehi  = (float *)malloc (nl * sizeof(float))) == NULL)bad_malloc++;
if((swpp->psd  = (float *)malloc (nl*ns*sizeof(float)))== NULL)bad_malloc++;
if((swpp->lineage = (IDFS_LINEAGE *)malloc( sizeof(IDFS_LINEAGE))) == NULL){
     bad_malloc++;
} 
if (do_pa){
  if((swpp->pa  = (float *)malloc (nl*ns*sizeof(float)))== NULL)bad_malloc++;
}

if (bad_malloc){
   printf("Malloc failed %d times in malloc_sweep_arrays\n",bad_malloc);
   free_sweep_arrays(swpp);
   return PEACE3D_ERROR;
} else {
   return ALL_OKAY;
}
}

/****************************************************************/
void nullify_sweep_arrays(PEACE_SWEEP *swpp){
/***************************************************************/
/* initialise sweep array pointers to NULL */
swpp->elo = NULL;
swpp->ehi = NULL;
swpp->thlo= NULL;
swpp->thhi= NULL;
swpp->psd = NULL;
swpp->pa  = NULL;
swpp->lineage = NULL;
return;
}

/*****************************************************************/
PEACE_SWEEP  *create_sweep_full_copy(PEACE_SWEEP *swp){
/*****************************************************************/

/* make a complete copy, including bin arrays, of a PEACE SWEEP */

PEACE_SWEEP *swpcpy;
SDDAS_SHORT ret_val;
int n_sen, n_lev;
T_OR_F do_pa;

int i, n_psd;

swpcpy = (PEACE_SWEEP *)malloc (sizeof (PEACE_SWEEP));

*swpcpy = *swp; /* copy all the contents; NB missing val here too. */

n_sen  = swpcpy->n_sen;
n_lev  = swpcpy->n_lev;
n_psd  = n_sen * n_lev;

do_pa = (swp->pa == NULL)? QPFALSE : QPTRUE; /* if pa not NULL copy them */

ret_val = malloc_sweep_arrays(swpcpy, do_pa);
if (ret_val != ALL_OKAY){
    printf("Create_sweep_full_copy malloc_sweep_arrays returned %d. Oops\n",
                ret_val);
    return NULL;
}   /* really should return error message, but instead let's return NULL */

for (i=0; i<n_sen; i++){
   swpcpy->thlo[i] = swp->thlo[i];
   swpcpy->thhi[i] = swp->thhi[i];
}

for (i=0; i<n_lev; i++){
   swpcpy->elo[i] = swp->elo[i];
   swpcpy->ehi[i] = swp->ehi[i];
}

for (i=0; i<n_psd; i++) {
   swpcpy->psd[i] = swp->psd[i];
}

if (do_pa){
  for (i=0; i<n_psd; i++) {
    swpcpy->pa[i] = swp->pa[i];
  }
}

copy_lineage(swpcpy->lineage, swp->lineage);

return swpcpy;
}

/*****************************************************************/
void free_sweep_arrays(PEACE_SWEEP *swp){
/*****************************************************************/
/* free bin, psd arrays and lineage. Careful, bin arrays may be shared, e.g.,
* with other sweeps in a spin structure */

if (swp->thlo    != NULL) free (swp->thlo);
if (swp->thhi    != NULL) free (swp->thhi);
if (swp->elo     != NULL) free (swp->elo);
if (swp->ehi     != NULL) free (swp->ehi);
if (swp->psd     != NULL) free (swp->psd);
if (swp->pa      != NULL) free (swp->pa);
if (swp->lineage != NULL) free(swp->lineage);

nullify_sweep_arrays(swp);

return;
}

/*************************************************************************/
void free_n_sweeps(int n, PEACE_SWEEP **swppp){
/*************************************************************************/
/* free up n PEACE sweep structures, the pointers to which are in the
*  array swppp[]. For safety, the initial pointers should have been
*  initialised to NULL. Note that malloc_sweep_arrays is structured
*  to ensure that the bins and psd arrays are not NULL. Note, however
*  that free_sweep_arrays tries to free bins for each sweep, so they
*  must each have their own copy. This is not guaranteed, e.g., sweeps
*  in a spin structure may SHARE bin arrays.
*/
int i;

for (i=0; i<n; i++){
   if (swppp[i] != NULL){
       free_sweep_arrays(swppp[i]);
       free(swppp[i]);
   }
}

return;
}

/******************************************************************/
void print_sweep_structure(PEACE_SWEEP *swpp){
/*****************************************************************/

int i,isen,ilev;

printf("\n\nPEACE SWEEP STRUCTURE DIAGNOSTIC PRINTOUT\n");
printf("-----------------------------------------\n\n");
printf("%s %s %s %s %s %s version: %d\n",swpp->lineage->project,
    swpp->lineage->mission,swpp->lineage->experiment,swpp->lineage->instrument,
    swpp->lineage->virtual_inst, swpp->lineage->exten, swpp->lineage->vsn);
printf("Sweep Start time:  "); print_idfs_time(&swpp->btime);printf("\n");
printf("Sweep End time:    "); print_idfs_time(&swpp->etime); printf("\n");
printf("Start and Stop az angles (deg): %g, %g\n",swpp->phi_lo, swpp->phi_hi);
printf("Number of sensors = %d, number of energy levels = %d\n",
         swpp->n_sen, swpp->n_lev);
printf("PHI_ZERO_FRAME %d, BIN_UNITS %d, PSD_UNITS %d\n",
        swpp->phi_frame, swpp->bin_units, swpp->psd_units);
printf("Theta bins: i, theta low, theta high\n");
for (i=0; i<swpp->n_sen; i++){
   printf("%d    %6g  %6g\n", i, swpp->thlo[i], swpp->thhi[i]);
}
printf("Energy bins: i, E low, E high\n");
for (i=0; i<swpp->n_lev; i++){
   printf("%d    %8g   %8g\n", i, swpp->elo[i], swpp->ehi[i]);
}

printf("\n");
printf("Phase Space Density for all Sensors (Polar Bins)\n");
printf("---------------------------------------------------------\n");

for ( isen = 0; isen < swpp->n_sen; isen++ ){
   printf("\nSensor %d for %g < theta < %g from lowest energy to highest\n",
              isen, swpp->thlo[isen],swpp->thhi[isen]);
   i = swpp->n_lev * isen;
   for (ilev = 0; ilev < swpp->n_lev; ilev++){
      if (ilev % 5 == 0) printf("\n");
      printf("%10g  ",swpp->psd[i+ilev]);
   }
   printf("\n");
}

if(swpp->pa == NULL){
  printf("No pitch angles held in sweep\n");
} else {
   printf("\n");
   printf("Pitch Angles for all Sensors (Polar Bins)\n");
   printf("---------------------------------------------------------\n");

   for ( isen = 0; isen < swpp->n_sen; isen++ ){
      printf("\nSensor %d for %g < theta < %g from lowest energy to highest\n",
        	 isen, swpp->thlo[isen],swpp->thhi[isen]);
      i = swpp->n_lev * isen;
      for (ilev = 0; ilev < swpp->n_lev; ilev++){
	 if (ilev % 5 == 0) printf("\n");
	 printf("%10g  ",swpp->pa[i+ilev]);
      }
      printf("\n");
   }
}
return;
}

/***********************************************************************/
SDDAS_SHORT CombineSweeps(
        PEACE_SWEEP *swp_new, /* target; could be part full */
        float *weight,       /* 2-D array; same size as new*/
	PEACE_SWEEP *swp_old, /* old sweep */
	INTERP_QUANTITY interpq  /* differential or integ (cnts)*/
	){
/***********************************************************************/
/* Combine the contents of swp_old onto swp_new.
*  On return, only psd values in swp_new are changed.
*
*  Note: A sweep is a slice at ~fixed azimuth from polar angle >= to <=180
*    like an orange segment. Interpolation assumes polar area bins of areas
*    0.5 * (r2^2 - r1^2)(cos theta1 - cos theta2)
*
*  Algorithms and use:
* 
*  Prior to calling this routine:
*     If re-griding, dimension arrays in swp_new as desired, assign
*        theta and level bin values, and initialise weights all zero.
*
*     If combining onto existing sweep, assign all weights to be area of bin
*     This is done by calling fill_sweep_bin_area(weight,swp_new);
*
*  Algorithm
*     If bin units are an integral measure (e.g., counts) divide psd
*        arrays in swp_new and swp_old by bin area to get density
*     Multiply swp_new by weight for each bin
*     Find overlap of an old bin on a new bin
*     Add swp_old->psd * overlap area to new sweep
*     Add overlap area to weight
*     Divide swp_new->psd by weight
*     If bin units are an integral measure
*         Multiply swp_old-> by bin area to restore to counts
*         Multiply swp_new->psd by bin area to get to counts
*
*     Mathematically let d <=> data, i.e., swp_old
*                        g <=> grid, i.e., swp_new
*                        Ao = Area of overlap between bin d and bin g
*
*      f_g = sum f_d * Ao / sum Ao     where sum Ao = weight for bin g
*
*      Cg  = (Ag / sum Ao) * sum [ (Cd/Ad)*Ao ] 
*
*      swp->psd is array z[n_sen][n_lev] 
*               so psd[jth][ilev] = psd[n_lev*jth+ilev]
*
*****************************************************************************/

float *Ad, *Ag;   /* to hold areas of old and new grid, if integ */
int ilevg, ithg, nlevg, nthg,  kg;
int ilevd, ithd ,nlevd, nthd,  kd;
int ilevdov1, ilevdov2, ithdov1, ithdov2, n_overlap_lev, n_overlap_th;

float *elog, *ehig, *elod, *ehid, *thlog, *thhig, *thlod, *thhid;
float *psdg, *psdd;
float thovlo, thovhi, levovlo, levovhi; /* to hold overlap region */
float add_weight, add_psd, missingold, missingnew;

/* make local copies of lots of things */

nlevg = swp_new->n_lev;     nthg  = swp_new->n_sen;
nlevd = swp_old->n_lev;     nthd  = swp_old->n_sen;
elog  = swp_new->elo;       ehig  = swp_new->ehi;
elod  = swp_old->elo;       ehid  = swp_old->ehi;
thlog = swp_new->thlo;      thhig = swp_new->thhi;
thlod = swp_old->thlo;      thhid = swp_old->thhi;
psdg  = swp_new->psd;       psdd  = swp_old->psd;
missingold = swp_old->missing_value;
missingnew = swp_new->missing_value;

switch (interpq){
   case INTEGRAL:
      Ad = (float *) malloc (nlevd * nthd * sizeof(float));
      Ag = (float *) malloc (nlevg * nthg * sizeof(float));
      fill_sweep_bin_area(Ad, swp_old); fill_sweep_bin_area(Ag, swp_new);
      if(DIAG_PRINT_INTERP){
        printf("New bin thetas\n");
	print_bins(thlog,thhig,nthg);
	printf("New bin levels\n");
	print_bins(elog,ehig,nlevg);
        printf("Old bin thetas\n");
	print_bins(thlod,thhid,nthd);
	printf("Old bin levels\n");
	print_bins(elod,ehid,nlevd);
        printf("Integral quantity, e.g., counts\n New Bin Areas:\n");
	print_weight(Ag, nthg, nlevg);
	printf("Old Bin Areas\n");
	print_weight(Ad, nthd, nlevd);
      }
      divide_by_area(Ad, swp_old);      divide_by_area(Ag, swp_new);
      break;
   case DIFFERENTIAL: default:
      break;
}

multiply_by_area(weight, swp_new); /* get ready */

for (ithg = 0; ithg<nthg; ithg++){ /* go through all bins in new sweep grid */
   kg = nlevg*ithg;
   for (ilevg = 0;ilevg<nlevg;ilevg++){/*for each new bin, find all overlaps*/
      n_overlap_th = find_subset_in_range(
         thlog[ithg], thhig[ithg],thlod,thhid,nthd,&ithdov1, &ithdov2);
      n_overlap_lev = find_subset_in_range(
         elog[ilevg], ehig[ilevg],elod,ehid,nlevd,&ilevdov1, &ilevdov2);
      if(DIAG_PRINT_INTERP){
         printf("For new bin (theta,lev) = (%d, %d) ",ithg,ilevg);
         printf("%d theta bins (%d->%d); %d levels (%d->%d) overlap\n",
	    n_overlap_th, ithdov1, ithdov2,n_overlap_lev, ilevdov1,ilevdov2);
      }
      if (n_overlap_th && n_overlap_lev){
         for ( ithd = ithdov1; ithd <= ithdov2; ithd++){
	   kd = nlevd*ithd;
	     /* overlap limits are max(low bounds) and min(high bounds) */
	   thovlo = ( thlog[ithg] > thlod[ithd] ) ? thlog[ithg] : thlod[ithd] ;
	   thovhi = ( thhig[ithg] < thhid[ithd] ) ? thhig[ithg] : thhid[ithd] ;
	   for (ilevd = ilevdov1; ilevd <=ilevdov2; ilevd++){
	     if ( (add_psd = psdd[kd+ilevd])  !=  missingold){
	       levovlo = (elog[ilevg] > elod[ilevd] ) ? elog[ilevg]:elod[ilevd];
	       levovhi = (ehig[ilevg] < ehid[ilevd] ) ? ehig[ilevg]:ehid[ilevd];
	       add_weight = polar_area(levovlo,levovhi,thovlo,thovhi);
	       weight[kg+ilevg]+= add_weight;
	       if (psdg[kg+ilevg]==missingnew){
	          psdg[kg+ilevg]  = add_psd*add_weight;
	       } else {
	          psdg[kg+ilevg] += add_psd*add_weight;
	       }
	       if(DIAG_PRINT_INTERP){
		 printf("Ovlp th(%g->%g) lev(%g->%g) wt %g x %g",
	             thovlo,thovhi,levovlo,levovhi,add_weight,psdd[kd+ilevd]);
		 printf(" old (t,l) = (%d,%d) to (%d,%d)\n", 
	                    ithd,ilevd,ithg,ilevg);
	       }
	     } /* end if (add_psd != missingold ... skips if missing val */ 	        
	   }
	 }
       }
       if(DIAG_PRINT_INTERP){
          printf("New bin (%d, %d) weight = %g, psd x weight = %g\n",
	       ithg, ilevg, weight[kg+ilevg], psdg[kg+ilevg]);
       }      
   }
}

divide_by_area(weight,swp_new);   /* divide by total weight */

switch (interpq){
   case INTEGRAL:                 /* convert back to, e.g., cnts */
      multiply_by_area(Ad, swp_old);
      multiply_by_area(Ag, swp_new);
      if (Ad != NULL) free(Ad);
      if (Ag != NULL) free(Ag);
      break;
   case DIFFERENTIAL: default:
      break;
}

if(DIAG_PRINT_INTERP){
  printf("CombineSweeps finished. swp_new is:\n");
  print_sweep_structure(swp_new);
}

return ALL_OKAY;

}

/***********************************************************************/
SDDAS_SHORT fill_sweep_bin_area(float *A, PEACE_SWEEP *swpp){
/***********************************************************************/
int ilev,jth,k;
int n_lev, n_th;
float *thlo, *thhi, *elo, *ehi;
float th1,th2;

n_lev = swpp->n_lev;
n_th  = swpp->n_sen;
thlo  = swpp->thlo;
thhi  = swpp->thhi;
elo   = swpp->elo;
ehi   = swpp->ehi;

for (jth = 0; jth < n_th; jth++){
   k = n_lev * jth;
   th1 = thlo[jth];
   th2 = thhi[jth];
   for (ilev = 0; ilev < n_lev; ilev++){
      A[k + ilev] = polar_area(elo[ilev],ehi[ilev],th1,th2);
   }
}

return ALL_OKAY;
}

/***********************************************************************/
SDDAS_SHORT divide_by_area(float *A, PEACE_SWEEP *swpp){
/***********************************************************************/
int k, kmax;
float missing;

missing = swpp->missing_value;

kmax = swpp->n_lev * swpp->n_sen;
for (k=0; k<kmax; k++){
   if ( A[k] == 0.0 || swpp->psd[k] == missing) swpp->psd[k]=missing;
   else swpp->psd[k] /= A[k];
}

return ALL_OKAY;
}

/***********************************************************************/
SDDAS_SHORT multiply_by_area(float *A, PEACE_SWEEP *swpp){
/***********************************************************************/
int k, kmax;
float missing;

missing = swpp->missing_value;

kmax = swpp->n_lev * swpp->n_sen;
for (k=0; k<kmax; k++){
   if (swpp->psd[k] != missing) swpp->psd[k] *= A[k] ;
}

return ALL_OKAY;
}


/**************************************************************************/
int test_sweeps_congruent(PEACE_SWEEP *swp1, PEACE_SWEEP *swp2){
/**************************************************************************/
/* see if sweeps have same bins, units, etc. 
*  returns:
*
* typedef enum SWEEP_CONGRUENT_TEST_E{
*   BIN_UNIT_MISMATCH = -2,
*   PSD_UNIT_MISMATCH = -1,
*   CONGRUENT         = 0,
*   NLEV_MISMATCH     = 1,
*   NSEN_MISMATCH     = 2,
*   EBINS_MISMATCH    = 3,
*   THBINS_MISMATCH   = 4
*   UNKNOWN_CONGRUENT_ERROR = 5
} SWEEP_CONGRUENT_TEST;
*/
int i, nlev;

if (swp1->bin_units != swp2->bin_units) return BIN_UNIT_MISMATCH;
if (swp1->psd_units != swp2->psd_units)return PSD_UNIT_MISMATCH;
if (swp1->n_sen != swp2->n_sen) return NSEN_MISMATCH;
if (swp1->n_lev != swp2->n_lev) return NLEV_MISMATCH;

i=0;
while (swp1->thlo[i] == swp2->thlo[i] && 
       swp1->thhi[i] == swp2->thhi[i] &&
       i < swp1->n_sen)  i++;
if (i != swp1->n_sen) return THBINS_MISMATCH;

i=0;
nlev = swp1->n_lev;
while (swp1->elo[i] == swp2->elo[i] && 
       swp1->ehi[i] == swp2->ehi[i] &&
       i < nlev)  i++;
if (i != swp1->n_lev) return EBINS_MISMATCH;

return CONGRUENT;
}

/*****************************************************************/
PEACE_SWEEP *sweep_arithmetic(PEACE_SWEEP *swp_pri, PEACE_SWEEP *swp_sec,
                              ARITH_OPER oper){
/*****************************************************************/
/* perform swp_pri oper swp_sec using primary sweep for bins, etc. 
*
*  Create full copy of primary sweep; note psd must be initialised,
*    even if overwriting, to ensure that psd x weight is a number;
*    else if psd = nan, then psd x 0.0 = nan !!
*
*  If swp_pri and swp_sec have same bins, etc.
*     Do requested operation
*
*  Else 
*     Create 2nd copy of primary sweep (for bins, etc.)
*     Combine secondary sweep onto this copy
*     Do requested operation
*        Missing values trapped. 
*             For + or -, skip
*             For * or /, set result to missing
*     Delete 2nd copy
*
*  Return pointer to resultant sweep (nb this is a full copy, with
*     its own private copies of bins, etc.)
************************************************************************/   

PEACE_SWEEP *swp2=NULL, *swp1_cpy=NULL, *swp2_congr = NULL;
float *weights = NULL;
int i, n_psd;
float temp,missing1,missing2;

const float APPROX_ZERO = 100. * FLT_MIN;

n_psd = swp_pri->n_sen * swp_pri->n_lev;

switch(test_sweeps_congruent(swp_pri,swp_sec)){
   case NLEV_MISMATCH:     /* bins not identical (but passed units tests) */
   case NSEN_MISMATCH:
   case EBINS_MISMATCH:
   case THBINS_MISMATCH:
     if (DIAG_PRINT_INTERP)printf("Sweeps not congruent; new one created\n");
     swp2_congr = create_sweep_full_copy(swp_pri); 
     weights = (float *)malloc(swp_pri->n_lev * swp_pri->n_sen * sizeof(float));
     for (i = 0; i<n_psd; i++) weights[i] = 0.0; /*psd will be overwritten */
     CombineSweeps(swp2_congr, weights, swp_sec,
                      diff_or_integral(swp_pri->psd_units));		      
     swp2 = swp2_congr;
     break;
   case CONGRUENT:     /* both units and bins identical */
     swp2 = swp_sec;
     break;
   case BIN_UNIT_MISMATCH: /* units not same or confusion*/
   case PSD_UNIT_MISMATCH:
   case UNKNOWN_CONGRUENT_ERROR:
   default:
     return NULL;
     break;
}
     
swp1_cpy = create_sweep_full_copy(swp_pri); /*cpy, init bins, */
missing2 = swp2->missing_value;
missing1 = swp1_cpy->missing_value;

switch (oper){

  case ADD:
     for (i=0; i<n_psd; i++) {
        if( (temp=swp2->psd[i]) != missing2 && swp1_cpy->psd[i] != missing1) {
	     swp1_cpy->psd[i] += temp;
	}
     }
     break;

  case SUBTRACT:
     for (i=0; i<n_psd; i++) {
        if( (temp=swp2->psd[i]) != missing2 && swp1_cpy->psd[i] != missing1) {
           swp1_cpy->psd[i] -= temp;
	}
     }
     break;

  case MULTIPLY:
     for (i=0; i<n_psd; i++) {
        if( (temp=swp2->psd[i]) != missing2 && swp1_cpy->psd[i] != missing1) {
              swp1_cpy->psd[i] *= temp;
	} else swp1_cpy->psd[i] = missing1;
     }
     break;

  case DIVIDE:
     for (i=0; i<n_psd; i++) {
        temp = swp2->psd[i];
	if (fabs(temp) < APPROX_ZERO || 
	      temp == missing2 || swp1_cpy->psd[i]==missing1) {
	   swp1_cpy->psd[i] = missing1;
	}
	else swp1_cpy->psd[i] /= temp;
     }
     break;
  default:
     printf("Unknown Arithmetic Operation for Sweep; missing vals returned\n");
     for (i=0; i<n_psd; i++) swp1_cpy->psd[i] = missing1;
     break;
}

if (weights != NULL) free(weights);
if (swp2_congr != NULL) {
    free_sweep_arrays(swp2_congr);
    free(swp2_congr);
}

return swp1_cpy;

}

/*******************************************************************/
PEACE_SWEEP *rebin_sweep(PEACE_SWEEP *bin_holder, PEACE_SWEEP *swp){
/*******************************************************************/
/* Re-bin swp using the bins in bin_holder, which needs the following:
*
*  n_sen, n_lev, thlo[], thhi[], elo[], ehi[], bin_units
*
*  All other bits are copied from *swp except pitch angles
*
*  Returns a complete sweep structure (including copies of bins),
*  or NULL if BIN_UNITS don't match;
*/

PEACE_SWEEP *swp_rebin;
float *weights=NULL, *psd_local;
int i, n_sen, n_lev, n_psd;
SDDAS_SHORT ret_val;

if (bin_holder->bin_units != swp->bin_units){
   printf("rebin_sweep failed; bin-units don't match\n");
   return NULL;
}

/* create new sweep and copy contents of old. Note that the bins and
*  psd are held as pointers, so no array memory allocated here */
swp_rebin = (PEACE_SWEEP *)malloc (sizeof (PEACE_SWEEP));
if (swp_rebin == NULL){
   printf("rebin_sweep failed to malloc ptr\n");
   return NULL;
}

*swp_rebin = *swp;  /* copy contents; arrays will be replaced */

/* prepare for new arrays */

n_sen = swp_rebin->n_sen = bin_holder->n_sen;
n_lev = swp_rebin->n_lev = bin_holder->n_lev;
n_psd = swp_rebin->n_lev * swp_rebin->n_sen;
swp_rebin->pa = NULL;

ret_val=malloc_sweep_arrays(swp_rebin, QPFALSE);
if(ret_val != ALL_OKAY){
  printf("rebin_sweep malloc_sweep_arrays returned %d - Oops!\n",ret_val);
  free(swp_rebin);
  return NULL;
}

copy_lineage(swp_rebin->lineage,swp->lineage);

psd_local = swp_rebin->psd;   /* local copy of pointer */

for (i=0; i<n_sen; i++){
   swp_rebin->thlo[i] = bin_holder->thlo[i];
   swp_rebin->thhi[i] = bin_holder->thhi[i];
}

for (i=0; i<n_lev; i++){
   swp_rebin->elo[i] = bin_holder->elo[i];
   swp_rebin->ehi[i] = bin_holder->ehi[i];
}

weights = (float *)malloc(n_psd * sizeof(float));
for (i = 0; i<n_psd; i++) {
    weights[i] = psd_local[i] = 0.0; /*psd will be overwritten */
}

CombineSweeps(swp_rebin, weights, swp, diff_or_integral(swp_rebin->psd_units));		      

free(weights);
return swp_rebin;

}

/**************************************************************************/
SDDAS_SHORT rebin_n_sweeps(PEACE_SWEEP *bin_holder, int nswps, 
     PEACE_SWEEP **swpsinpp, int *nrebinned, PEACE_SWEEP ***swpsoutppp){
/**************************************************************************/
/* Re-bin nswps sweeps using the bins in bin_holder, which needs the following:
*
*  n_sen, n_lev, thlo[], thhi[], elo[], ehi[], bin_units
*
*  All other bits are copied from **swpsinpp except pitch angles
*
*  Returns nrebinned sweep structures in swpsoutpp (including copies of bins
*  for EACH sweep - this is an overkill, but consistent with rest of QPEACE),
*  or NULL if BIN_UNITS don't match)
*
*  Note: swpsoutppp holds nswps pointers, but it is conceivable that
*  the number rebinned is less than this; the unused pointers are set
*  to NULL.
*
*  This routine mallocs the sweep pointer array swpsoutppp and all internal
*  arrays.
*/

int iswp;
SDDAS_SHORT ret;
PEACE_SWEEP **swppp;

swppp = (PEACE_SWEEP **)malloc(nswps * sizeof(PEACE_SWEEP *));
if (swppp == NULL){
   printf("Malloc error in rebin_n_sweeps\n");
   return SWEEP_REBIN_ERROR;
}
for (iswp = 0; iswp < nswps; iswp++) swppp[iswp] = NULL;
*swpsoutppp = swppp; /* return pointer */
*nrebinned = 0;
for (iswp = 0, ret = ALL_OKAY; iswp < nswps && ret == ALL_OKAY; iswp++){
   swppp[iswp] = rebin_sweep(bin_holder, swpsinpp[iswp]);
   if (swppp[iswp] == NULL) ret = SWEEP_REBIN_ERROR;  else (*nrebinned)++;
}

if (*nrebinned > 0){
    return ALL_OKAY;
} else {
    free_n_sweeps(nswps, swppp); /* safe since all swps initialised to NULL */
    free(swppp);
    printf("rebin_n_sweeps failed to rebin even 1 sweep\n");
    return SWEEP_REBIN_ERROR;
}


} 

/**************************************************************************/
void print_weight(float *weight, int n_th, int n_lev){
/**************************************************************************/

int ith,ilev, k;

printf("Weights[theta][level] array is:\n");
printf("-------------------------------\n");
for (ith = 0; ith<n_th; ith++){
  k = n_lev*ith;
  printf("\nTheta bin %d; levels commence at 0\n",ith);
  for (ilev = 0; ilev < n_lev; ilev++){
    if (ilev % 6 == 6)printf("\n");
    printf("%8.3g  ",weight[k+ilev]);
  }
}
printf("\n");
}

/******************************************************************/
void sweeps_autoscaler(PEACE_SWEEP **swppp, int nswp,
                       float *psdmin, float *psdmax){
/******************************************************************/
/*  Inputs:   Array swppp of pointers to nswp sweep structures
*  
*   Output:  min and max values of psd array; missing vals trapped
*
*   Code borrowed from Sweep_wheel_plot
*/

int i, iswp, idummy, n_sen, n_lev, j;
PEACE_SWEEP *curswp;
float pmin, pmax,ptemp;
float psd_abs_min, missing;

psd_abs_min = 100. * FLT_MIN;
pmin = pmax = psd_abs_min;

iswp=0;    /* cope with 1st sweep all zeroes by looping while */
while(fabs(pmin) < 10.0* psd_abs_min && iswp < nswp){

  curswp = swppp[iswp];   /* use ith sweep to initialise; usually 1st ! */
  n_sen = curswp->n_sen;
  n_lev = curswp->n_lev;
  missing = curswp->missing_value;

  idummy = n_sen * n_lev;        /* find non-zero psd to initialise min max */
  for( i=0; i<idummy; i++){
     if (fabs(curswp->psd[i]) > psd_abs_min && curswp->psd[i]!=missing){
	pmin = pmax = curswp->psd[i];
	break;
     }
  }
  iswp++;
}

for (iswp = 0; iswp < nswp; iswp++){
  curswp = swppp[iswp];
  n_sen = curswp->n_sen;
  n_lev = curswp->n_lev;
  missing = curswp->missing_value;

  j=n_sen * n_lev;
  for (i=0; i<j; i++){
      ptemp = curswp->psd[ i ];
        if (ptemp != missing){    /* change to use missing val in sweep */
	pmin = (pmin < ptemp ) ? pmin : ptemp;
	pmax = (pmax > ptemp ) ? pmax : ptemp;
      }
  }
}

if (fabs(pmax)> 0.1*FLT_MAX){  /* remove inf values */
   pmax = (pmax > 0.0) ? 0.1*FLT_MAX : -0.1*FLT_MAX;
}
if (fabs(pmin)> 0.1*FLT_MAX){  /* remove inf values */
   pmin = (pmin > 0.0) ? 0.1*FLT_MAX : -0.1*FLT_MAX;
}
if(pmax < pmin){ptemp=pmin; pmin = pmax; pmax=ptemp;}
if (fabs(pmax - pmin)<FLT_MIN){  /* make psd range finite */
   pmax = (fabs(pmin)<FLT_MIN) ? 1.0 : 1.1*fabs(pmin);
}

*psdmin = pmin;
*psdmax = pmax;
return;
}

/******************************************************************/
PEACE_1D_SWEEP_CUT *sweep_1D_cut(float theta, PEACE_SWEEP *swp){
/***************************************************************/
/* Select the 1-D radial cut correponding to the angle theta from
*  a sweep and fill the PEQCE_1D_SWEEP_CUT structure. THis is
*  intended primarily for plotting, so the structure has only
*  centre values for energies, ... but carries all the
*  necessary metadata inside the structure */

int inbin, i,k, nlev;
PEACE_1D_SWEEP_CUT *swpcut;

nlev = swp->n_lev;

inbin = 0;
while (swp->thhi[inbin] < theta && inbin < swp->n_sen-1 )inbin++;
if (inbin == swp->n_sen){
   printf("1D CUT: Theta bins exhausted\n");
   return NULL;
}


swpcut = (PEACE_1D_SWEEP_CUT *)malloc(sizeof(PEACE_1D_SWEEP_CUT));
swpcut->ecen = (float *)malloc(swp->n_lev * sizeof(float));
swpcut->psd1D = (float *)malloc(swp->n_lev * sizeof(float));
swpcut->lineage = (IDFS_LINEAGE *)malloc(sizeof(IDFS_LINEAGE));

swpcut->starttime = swp->btime;
swpcut->endtime   = swp->etime;
swpcut->n_lev     = nlev;
swpcut->phi_cen   = 0.5*(swp->phi_lo + swp->phi_hi);
swpcut->thcen     = 0.5*(swp->thlo[inbin] + swp->thhi[inbin]);

k= nlev*inbin;

for (i=0; i< nlev; i++){
    swpcut->ecen[i] = 0.5*(swp->elo[i]+swp->ehi[i]);
    swpcut->psd1D[i] = swp->psd[k+i];
}

swpcut->bin_units = swp->bin_units;
swpcut->psd_units = swp->psd_units;
copy_lineage(swpcut->lineage , swp->lineage);

return swpcut;

}

/*****************************************************************/
void free_1D_cut_arrays (PEACE_1D_SWEEP_CUT *swpcut){
/*****************************************************************/
/*  free up all the memory associated with the arrays of a 1D Sweep cut
*   To completely free a 1D cut use this, then free the sweep cut structure*/

free(swpcut->ecen);
free(swpcut->psd1D);
free(swpcut->lineage);

return;
}

/********************************************************************/
void print_1D_cut(PEACE_1D_SWEEP_CUT *swpcut){
/********************************************************************/
int i,nlev;

nlev=swpcut->n_lev;

printf("PEACE 1D SWEEP CUT\n====================\n");
printf("Sweep Start Time = ");print_idfs_time(&(swpcut->starttime));
print_idfs_lineage(swpcut->lineage);
printf("Az Centre = %g, Theta (polar) Centre = %g\n",
      swpcut->phi_cen, swpcut->thcen);
printf("\n%s, %s\n=========================\n", 
    BIN_UNITS_STRING(swpcut->bin_units), PSD_UNITS_STRING(swpcut->psd_units));
for (i=0; i<nlev; i++){
  printf("%11.5g  %11.5g\n", swpcut->ecen[i], swpcut->psd1D[i]);
}

return;
}



/**************************************************************/
SDDAS_SHORT malloc_spin_sweep(PEACE_SPIN3D *spinp, int n_sen, int n_lev,
               T_OR_F do_pa){
/*************************************************************/
/* Malloc spin and sweeps for a single spin. Specifically, malloc:
*
*  n_swp PEACE_SWEEP structures
*  thlo, thhi      arrays for theta limits: only 1 set for all sweeps
*  elo,  ehi       arrays for energy limits; only 1 set for all sweeps
*  lineage         structure; only 1 for all sweeps
*  n_swp psd arrays of n_lev x n_sen size
*  n_swp pa arrays  of n_lev x n_sen size IF do_pa is QPTRUE
*
*  also sets n_lev and n_sen inside all sweep structures.
*
*  */

int iswp;
float *thlo, *thhi, *elo, *ehi;

int bad_malloc = 0;

PEACE_SWEEP *this_swpp;
IDFS_LINEAGE *lngp;

thlo = thhi = elo = ehi = NULL;
spinp->sweep = NULL;
lngp = NULL;

spinp->sweep = (PEACE_SWEEP *)malloc (spinp->n_swp * sizeof(  PEACE_SWEEP ));
if (spinp->sweep == NULL){
    printf("PEACE3D: Sweep Malloc failed\n"); bad_malloc++;
}

thlo = (float *)malloc (n_sen * sizeof (float));
if (thlo == NULL){
   printf("PEACE3D: thho malloc failed\n"); bad_malloc++;
}

thhi = (float *)malloc (n_sen * sizeof (float));
if (thhi == NULL){
   printf("PEACE3D: thhi malloc failed\n"); bad_malloc++;
}

elo  = (float *)malloc (n_lev * sizeof (float));
if (elo == NULL){
   printf("PEACE3D: elo malloc failed\n"); bad_malloc++;}

ehi  = (float *)malloc (n_lev * sizeof (float));
if (ehi == NULL){
   printf("PEACE3D: ehi malloc failed\n"); bad_malloc++;
}

lngp  = (IDFS_LINEAGE *)malloc (sizeof (IDFS_LINEAGE));
if (lngp == NULL){
   printf("PEACE3D: lngp malloc failed\n"); bad_malloc++;
}


if (bad_malloc == 0){  /* only proceed if so far so good */
  for (iswp = 0; iswp < spinp->n_swp; iswp++){
    this_swpp = &(spinp->sweep[iswp]);
    this_swpp->n_sen   = n_sen;          /* all sweeps use same arrays */
    this_swpp->n_lev   = n_lev;
    this_swpp->thlo    = thlo;
    this_swpp->thhi    = thhi;
    this_swpp->elo     = elo;
    this_swpp->ehi     = ehi;
    this_swpp->lineage = lngp;
    this_swpp->psd   = (float *)malloc (n_sen * n_lev * sizeof (float) );
    if (this_swpp->psd == NULL){
      printf("PEACE3D: sweep psd malloc failed\n"); bad_malloc++;
    }
    if (do_pa){
       this_swpp->pa   = (float *)malloc (n_sen * n_lev * sizeof (float) );
       if (this_swpp->pa == NULL){
	 printf("PEACE3D: sweep pa malloc failed\n"); bad_malloc++;
       }
    } else { 
       this_swpp->pa = NULL;   /* set to NULL so can trap before free */
    }
  }
}

if (bad_malloc == 0) return ALL_OKAY;

/* if arrive here then a malloc failed; free all successful bits */
if  (spinp->sweep != NULL){
     for (iswp = 0; iswp < spinp->n_swp; iswp++){
        if (spinp->sweep[iswp].psd != NULL) free (spinp->sweep[iswp].psd);
        if (spinp->sweep[iswp].pa != NULL) free (spinp->sweep[iswp].pa);
     }
     free (spinp->sweep);
}

if  (thlo         != NULL) free (thlo);
if  (thhi         != NULL) free (thhi);
if  (elo          != NULL) free (elo);
if  (ehi          != NULL) free (ehi);
if  (lngp         != NULL) free (lngp);

return PEACE3D_ERROR;
}

/********************************************************************/
PEACE_SPIN3D *copy_spin (PEACE_SPIN3D *spin){
/********************************************************************/
/* make a full copy of a spin; mallocs the spin structure and internals */

PEACE_SPIN3D *spincopy;
PEACE_SWEEP *sweep, *swpcpy;
T_OR_F do_pa;
SDDAS_SHORT ret;

int n_sen, n_lev, n_swp, n_psd, i,j;

spincopy = (PEACE_SPIN3D *)malloc(sizeof(PEACE_SPIN3D));
if (spincopy == NULL) return NULL;

if (spin->sweep[0].pa == NULL) do_pa = QPFALSE; else do_pa = QPTRUE;
*spincopy = *spin;  /* copy btime, etime,tspin, n_swp */

n_sen = spin->sweep[0].n_sen;
n_lev = spin->sweep[0].n_lev;
n_swp = spin->n_swp;
n_psd = n_lev*n_sen;


/* malloc spin + sweeps; sets n_sen and n_lev inside sweep structures */
ret = malloc_spin_sweep( spincopy, n_sen, n_lev, do_pa);
if (ret == PEACE3D_ERROR) return NULL;  
/* copy contents - non-ptr stuff in particular and ptrs to bins */
/* can't do with a single line spincopy->sweep[i] = spin->sweep[i] */
/* because this would also reset array pts to pt to wrong place */
for (i=0; i<n_swp; i++) {
     spincopy->sweep[i].btime = spin->sweep[i].btime;
     spincopy->sweep[i].etime = spin->sweep[i].etime;
     spincopy->sweep[i].phi_lo = spin->sweep[i].phi_lo;
     spincopy->sweep[i].phi_hi = spin->sweep[i].phi_hi;
     spincopy->sweep[i].phi_frame = spin->sweep[i].phi_frame;
     spincopy->sweep[i].bin_units = spin->sweep[i].bin_units;
     spincopy->sweep[i].psd_units = spin->sweep[i].psd_units;
     spincopy->sweep[i].missing_value = spin->sweep[i].missing_value;
}

/* now fill arrays;  everything except psd and pa are shared so do once */
for (i=0; i< n_sen; i++){
   spincopy->sweep[0].thlo[i] = spin->sweep[0].thlo[i];
   spincopy->sweep[0].thhi[i] = spin->sweep[0].thhi[i];
}
for (i=0; i<n_lev; i++){
   spincopy->sweep[0].elo[i] = spin->sweep[0].elo[i];
   spincopy->sweep[0].ehi[i] = spin->sweep[0].ehi[i];
}
copy_lineage(spincopy->sweep[0].lineage , spin->sweep[0].lineage);
   
for (i=0; i<n_swp; i++){
   sweep = &(spin->sweep[i]);
   swpcpy = &(spincopy->sweep[i]);
   for (j=0; j<n_psd; j++){ swpcpy->psd[j] = sweep->psd[j]; }
   if (do_pa){
      for (j=0; j<n_psd; j++){ swpcpy->pa[j] = sweep->pa[j]; }
   }
}

return spincopy;
}
      



/*************************************************************/
SDDAS_SHORT free_spin_data (PEACE_SPIN3D *spinp){
/*************************************************************/
/* free up all arrays+lineage malloced by malloc_spin_sweep routine */

int iswp;

if  (spinp->sweep != NULL){
  iswp = 0;    /* use first sweep to free theta,energy arrays + lineage */
  if  (spinp->sweep[iswp].thlo    != NULL) free (spinp->sweep[iswp].thlo);
  if  (spinp->sweep[iswp].thhi    != NULL) free (spinp->sweep[iswp].thhi);
  if  (spinp->sweep[iswp].elo     != NULL) free (spinp->sweep[iswp].elo);
  if  (spinp->sweep[iswp].ehi     != NULL) free (spinp->sweep[iswp].ehi);
  if  (spinp->sweep[iswp].lineage != NULL) free (spinp->sweep[iswp].lineage);

  for (iswp = 0; iswp < spinp->n_swp; iswp++){
     if (spinp->sweep[iswp].psd != NULL) free (spinp->sweep[iswp].psd);
     if (spinp->sweep[iswp].pa != NULL) free (spinp->sweep[iswp].pa);
  }
  free (spinp->sweep);
}

return ALL_OKAY;
}

/****************************************************************/
SDDAS_SHORT free_n_spins(PEACE_SPIN3D **spinpp, int nspins){
/****************************************************************/
/* free up n PEACE spin structures, the pointers to which are in the
*  array spinpp[]. For safety, the initial pointers should have been
*  initialised to NULL. Note that malloc_spin_sweep is structured
*  to ensure that the bins and psd arrays are not NULL. Note also that
*  this uses free_spin_data which assumes that there is only 1 copy of
*  the energy/angle bins per spin, shared by all the sweeps, and that
*  each spin has a separate copy of these bins. This routine does
*  NOT free the array of pointers, although all pointers are set to
*  NULL
*/

int i;
        
for (i=0; i<nspins; i++){
  if(spinpp[i]!= NULL){
      free_spin_data(spinpp[i]);
      free(spinpp[i]);
  }
}       
return ALL_OKAY;
}

/*************************************************/
void print_peace_spin_info(PEACE_SPIN3D *spinp){
/***************************************************/

printf("\n\nPEACE Spin 3D info:\n");
printf("Spin Start time:  "); print_idfs_time(&spinp->btime);printf("\n");
printf("Spin End time:    "); print_idfs_time(&spinp->etime); printf("\n");
printf("Spin period:      %12g secs\n", spinp->tspin);
printf("%d sweeps per spin\n", spinp->n_swp);
return;
}

/*************************************************************************/
int find_az_sweep ( PEACE_SPIN3D *spinp, float az ){
/*************************************************************************/
/* search peace spin structure to find sweep which contains az */

int i, iroll, inext, ns;

float plo, plo_next, phi, azroll;

ns = spinp->n_swp;

if (ns == 1){    /* only 1 sweep, so return it!  [for SPINPAD] */
   return 0;
}

for (i=0; i<ns; i++){
   plo = spinp->sweep[i].phi_lo;
   inext = (i+1) % ns;
   plo_next = spinp->sweep[inext].phi_lo;
   phi = spinp->sweep[i].phi_hi;
   if (DIAG_PRINT){
       printf("az = %g, i = %d, bin phi lo, hi %g, %g, next bin phi lo %g\n",
                    az, i , plo, phi, plo_next);
   }
   if (phi < plo) phi += 360.;      /* roll over at 360 */
   if (plo_next < plo) plo_next += 360.;
   for (iroll=0; iroll <2; iroll++){  /* check without and with az rolled */
      azroll = az +iroll*360.;
      if (plo <= azroll && phi >= azroll) return i; /* yes, within this sweep*/
      if (phi <= azroll && plo_next >= azroll){ /* between this sweeps & nxt*/
         if (  azroll - phi < 0.5*(plo_next - phi)) return i;
         else return inext;
      }
   }
}

return -1;  /* can't seem to find azimuth */

}

/************************************************************************/
SDDAS_SHORT rebin_sweep_to_pa(PEACE_SWEEP **swpp, int nswps,
       PEACE_PITCH_ANGLE_OPTS *paopts, PEACE_SWEEP **paswpp){
/************************************************************************/
/* Create a new sweep based on the pitch angle info in nswps swp's
*  
*  Inputs:
*      swpp      An array of sweeps including filled pa arrays
*      nswps     Number of such sweeps; assumed sequential in time
*      paopts    Structure with nbins, mu or theta, etc.
*      paswpp    ptr to Sweep ptr; this routine mallocs sweep struct + arrays
*      
*      ASSUME missing value for pa is NOT in range 0->180. cf IDFS -3.4e38
*
*  Outputs:
*      *paswpp     Fully malloced sweep; pa array holds weights used
*/

int i, isen, k, kpi, ktry, itest, iswp, inb;
int nlev, nsenin, n_pabins, npsdin, npsdout;
float missing, temp;
float *lobins, *hibins, *pas, *psdin, *wt, *psdout;  /* local copies */
SDDAS_SHORT ret_val;
PEACE_SWEEP *paswp;
float *A, *psdwt, *Aout;

ret_val = ALL_OKAY;

for (i=0; i<nswps -1; i++){
  if ((itest=test_sweeps_congruent(swpp[i], swpp[i+1]))!=CONGRUENT ){
    printf("Rebin to pa got incongruent sweeps; err %d, and failed\n",itest);
    return SWEEP_TO_PA_ERROR;
  }
}

A = Aout = psdwt = NULL;

n_pabins = paopts->n_pabins;
nsenin = swpp[0]->n_sen;
nlev = swpp[0]->n_lev;
npsdin = nlev*nsenin;

paswp = (PEACE_SWEEP *)malloc(sizeof(PEACE_SWEEP));
*paswpp = paswp;
*paswp = *swpp[0];    /* copy contents; arrays will be re-malloced */
paswp->etime = swpp[nswps-1]->etime;   /* full  time and az spanned*/
paswp->phi_hi = swpp[nswps-1]->phi_hi;
paswp->n_sen = n_pabins;
npsdout = nlev*n_pabins;

ret_val = malloc_sweep_arrays(paswp,QPTRUE);  /* use pa array for weights */
if (ret_val != ALL_OKAY){
   printf("Failed to malloc sweep arrays for rebin_sweep_to_pa\n");
   free(paswp);
   return ret_val;
}
copy_lineage(paswp->lineage,swpp[0]->lineage);
missing= paswp->missing_value;        /* use missing val from input */

for (i=0; i<nlev; i++){
   paswp->elo[i] = swpp[0]->elo[i];
   paswp->ehi[i] = swpp[0]->ehi[i];
}

wt     = paswp->pa;
psdout = paswp->psd;
for (i=0; i<npsdout; i++){
   psdout[i]= missing;   /* initialise all to missing; weight = 0 held in pa */
   wt[i] = 0.0;
}

lobins = paswp->thlo;  /* local copies of pointers */
hibins = paswp->thhi;

if (paopts->mu_bins == QPTRUE){/* create mu-bins backward so incr in theta */
   set_equal_bins( lobins, hibins, n_pabins, 1.0, -1.0);
   for (i=0; i<n_pabins; i++){
      lobins[i] = RAD2DEG*acos( lobins[i]);
      hibins[i] = RAD2DEG*acos( hibins[i]);
   }
   lobins[0] = 0.0;  /* force to elim rounding error */
   hibins[n_pabins -1] = 180.0;
} else {
   set_equal_bins (lobins, hibins, n_pabins, 0.0, 180.0);
}

A = (float *)malloc(npsdin * sizeof(float));
psdwt = (float *)malloc(npsdin * sizeof(float));
if (A != NULL && psdwt != NULL){
   fill_sweep_bin_area(A,swpp[0]);
   switch (diff_or_integral(swpp[0]->psd_units)){
     case INTEGRAL:   /* eg counts */
       for (i=0; i<npsdin; i++) {psdwt[i]=1.0;} break; /*already integ over A*/
     case DIFFERENTIAL: default: /* eg psd */
       for (i=0; i<npsdin; i++) {psdwt[i]=A[i];} break;
   }
   for (iswp=0; iswp<nswps; iswp++){
     pas    = swpp[iswp]->pa;
     psdin  = swpp[iswp]->psd;

     if (pas == NULL) continue;  /* don't use sweeps with no pa */
			     /* ready to do the real work; */
			     /*  recall psd[jth][ilev] = psd[n_lev*jth+ilev] */

     for (isen = 0; isen<nsenin; isen++){  /* go through all input sensors */
	 k = nlev*isen;
	 for (i=0; i<nlev; i++){
	     kpi = k+i;
             inb = inbin(pas[kpi],lobins,hibins, n_pabins);
	     temp = psdin[kpi];
	     if (inb >= 0 && temp && temp!=missing){ /* only add non-zero psd*/
		ktry = nlev*inb+i;
		wt[ktry] +=A[kpi];
		if (psdout[ktry] == missing) {
		   psdout[ktry]=temp*psdwt[kpi];
		} else {
		   psdout[ktry] +=temp*psdwt[kpi];
		}
	     }
	  }
     }
   }
   for (i=0; i<npsdout; i++){
      if (psdout[i] != missing && (temp = wt[i]) != 0.0){
         psdout[i] /= temp;
      } else {
         psdout[i]  = missing;
      }
   }
   ret_val = ALL_OKAY;
   if (diff_or_integral(paswp->psd_units)==INTEGRAL){
     Aout = (float *)malloc(npsdout*sizeof(float));
     if (Aout == NULL){
        printf("Failed to malloc Aout in rebin sweep to pa\n");
	ret_val = SWEEP_TO_PA_ERROR;
     } else {
        fill_sweep_bin_area(Aout,paswp);
	for (i=0; i<npsdout; i++)psdout[i] *=Aout[i];
     }
   }
   if (strlen(paswp->lineage->virtual_inst)<LINEAGE_MAX_STR -4){
      strcat(paswp->lineage->virtual_inst,"_pa");
   }

} else {
   printf("Error mallocing A or psdwt in rebin_sweep_to_pa\n");
   ret_val = SWEEP_TO_PA_ERROR;
}

if (A     != NULL) free (A);
if (Aout  != NULL) free (Aout);
if (psdwt != NULL) free (psdwt);

return ret_val;
}

/**********************************************************************/
void set_pa_opts_defaults(PEACE_PITCH_ANGLE_OPTS *paopts){
/*********************************************************************/
paopts->do_pa = QPTRUE;
paopts->n_pabins = 13;
paopts->mu_bins  = QPFALSE;
return;
}

/**********************************************************************/
SDDAS_SHORT Rebin_n_sweeps_to_pa(PEACE_SWEEP **swppp, int nswps, 
    int nswpgrp, PEACE_PITCH_ANGLE_OPTS *paopts, 
    PEACE_SWEEP ***paswppp_p, int *npaswp){
/**********************************************************************/
/* Take nswps sweeps and group nswpgrp at a time, rebin the group in pitch
*  angle, and return npaswp pitch angle distributions. This routine mallocs
*  the array paswpp, while the internal contents which holds the npaswp 
*  results are malloced in the rebin_sweep_to_pa routine which this calls
*/

int iswp, igrp, ngrp;
SDDAS_SHORT ret_val;
PEACE_SWEEP **paswppp;

ngrp = nswps / nswpgrp;  /* integer arith; truncates */

paswppp = (PEACE_SWEEP **)malloc(ngrp * sizeof(PEACE_SWEEP *));
if (paswppp == NULL){
     printf("paswpp malloc failed in Rebin_n_sweeps_to_pa\n");
     return SWEEP_TO_PA_ERROR;
}
*paswppp_p = paswppp;  /* pass back pointer to newly malloced swp ptr array */
for (igrp = 0; igrp<ngrp; paswppp[igrp++]=NULL);
iswp = 0;
for (igrp = 0; igrp < ngrp; igrp++, iswp +=nswpgrp){
   ret_val = rebin_sweep_to_pa(&swppp[iswp],nswpgrp,paopts, &paswppp[igrp]);
   if (ret_val != ALL_OKAY){
      printf("rebin_sweep_to_pa in Rebin_n_sweeps_to_pa failed\n");
      free_n_sweeps(ngrp, paswppp);
      free(paswppp);
      return ret_val;
   }
}
*npaswp = ngrp;
return ALL_OKAY;

}

/**********************************************************************/
SDDAS_SHORT Rebin_n_spins_to_pa(PEACE_SPIN3D **spinpp, int nspins, 
     PEACE_PITCH_ANGLE_OPTS *paopts, PEACE_SWEEP ***paswppp_p){
/**********************************************************************/
/* Take nspins spins and rebin each onto a pitch angle distributions. 
*  This routine mallocs the array paswpp, while the internal contents which 
*  holds the npaswp results are malloced in the rebin_sweep_to_pa routine 
*  which this calls. 
*/

/* NB Spin structure contains an array of sweep structures, but rebin...
*  routine expects an array of ptrs to sweep structures, so we need to
*  copy the pointers. These are done tediously for safety */ 

int ispin, iswp, nswpinsp,nswpnew;
SDDAS_SHORT ret_val;
PEACE_SWEEP **paswppp, **swppp;

paswppp = (PEACE_SWEEP **)malloc(nspins * sizeof(PEACE_SWEEP *));
if (paswppp == NULL){
     printf("paswpp malloc failed in Rebin_n_spins_to_pa\n");
     return SWEEP_TO_PA_ERROR;
}
*paswppp_p = paswppp; /* pass swp array back */

for (ispin = 0; ispin<nspins; paswppp[ispin++]=NULL);
nswpinsp = spinpp[0]->n_swp;  /* initial number */
swppp = (PEACE_SWEEP **)malloc( nswpinsp * sizeof (PEACE_SWEEP*));

for (ispin = 0; ispin < nspins; ispin++){
   if (DIAG_SWEEP_TO_PA) {
      printf("*******************\nSpin %d\n********************\n",ispin);
      print_peace_spin_info(spinpp[ispin]);
   }
   nswpnew = spinpp[ispin]->n_swp;
   if (nswpnew != nswpinsp){  /* if number changed re-malloc */
      free(swppp);
      nswpinsp = nswpnew;
      swppp = (PEACE_SWEEP **)malloc( nswpinsp * sizeof (PEACE_SWEEP*));
   }
   for (iswp = 0; iswp < nswpinsp; iswp++) { /* the tedious way here */
        swppp[iswp] = &(spinpp[ispin]->sweep[iswp]); /* copy ptrs */
	if (DIAG_SWEEP_TO_PA){
	    printf("Sweep %d in spin is\n",iswp);
	    print_sweep_structure(swppp[iswp]);
	}
   }
   ret_val = rebin_sweep_to_pa(swppp,nswpinsp,
                       paopts, &paswppp[ispin]);
   if (ret_val != ALL_OKAY){
      printf("rebin_sweep_to_pa in Rebin_n_spins_to_pa failed\n");
      free_n_sweeps(nspins, paswppp);
      free(paswppp);
      free(swppp);
      return ret_val;
   }
   if(DIAG_SWEEP_TO_PA){
      printf("Rebinned pa sweep is:\n");
      print_sweep_structure(paswppp[ispin]);
   }
}

if (DIAG_SWEEP_TO_PA){
 printf("Finished Rebin_n_spins_to_pa\n");
}
free(swppp);
return ALL_OKAY;
}

/***********************************************************************/
int inbin(float value, float *lobins, float *hibins, int nbins) {
/***********************************************************************/
int i, nm1;

nm1=nbins-1;

if (value < lobins[0] || value > hibins[nm1]) return -1;
i=0;
while (value >= hibins[i] && i < nm1){i++;}
if (lobins[i]<=value) return i;
else return -1;

}


#ifdef __cplusplus
}
#endif
