/* Qsweep_spin_read.c ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*
*  May 2002 separate out from Qsweep_spin.c
*  June 2002 Resort polar bins to be flow and cope with SPINPAD reverse order
*  July 2002 Complete flow <-> look adjustments
*  Marc 2004 Duplicate promote loop to update database and data
*
*  Routines for reading PEACE sweeps and spins from idfs
*
*/

#ifdef __cplusplus
extern "C"{
#endif

#include "Qidfs_utils.h"
#include "Qsweep_spin.h"
#include "Qpeace_pad.h"

/*********************************************************************/
SDDAS_SHORT fill_sweep(PEACE_SWEEP *swpp, IDFS_LINEAGE *lngp,
              struct idf_data *idfp, SDDAS_ULONG dkey, 
	      T_OR_F have_bins, T_OR_F do_pa){
/*********************************************************************/

/* fill a sweep of PEACE data for all sensors. If already have the
*  energy/theta bin boundaries and lineage, don't get them again.
*  (in practice, for spins this routine is passed the sweep number, 
*  zero being the 1st bin).
*
*  Need to set bin/psd units and lineage prior to calling this routine
*  and position file to start of spin. Also need to set options in paopts
*  if convert to pitch angles sought.
*/

int  isen, isen_spad, nsen, is_spinpad;
SDDAS_CHAR fwd, full_swp;
SDDAS_SHORT ret_val;
static int swp_step0, swp_step1;
IDFS_TIME dummytime;
float tspin;

is_spinpad = strncmp("SPINPAD",lngp->instrument,7);
if (is_spinpad==0) is_spinpad=1; else is_spinpad = 0;

nsen = swpp->n_sen;
full_swp = 1;

for (isen = 0, isen_spad = nsen -1; isen < nsen; isen++, isen_spad--){
   fwd =  (isen < nsen -1 ) ?   0 : 1 ;
   ret_val = read_drec (dkey,lngp->exten, lngp->vsn, idfp, isen,fwd, full_swp);
   if (ret_val < 0 || !idfp->filled_data){
      printf ("\nError %d from read_drec.\n or no data for sensor %d",
             ret_val, isen);
      return ret_val;   /* changed 19/6/2001 from PEACE3D_ERROR */
                        /* for GetNPeaceSweeps to trap NEXT_FILE_STATUS */
   }
   if (DIAG_SWEEP_IDF){print_idf_data(idfp);}
   if (idfp->num_sample != swpp->n_lev){
      printf("PEACE3D: n_level check fail\n");
      printf("   idfp->num_sample = %d, n_lev expected = %d\n",
                 idfp->num_sample, swpp->n_lev);
      return SWEEP_NLEVEL_CHANGE;
   }
   if (!have_bins && isen == 0){    /* need  e-bin arrays+lineage; use sen 0*/
      swp_step0 = idfp->swp_data[0];  /* static vars to hold 1st couple of */
      swp_step1 = idfp->swp_data[1];  /* sweep steps as mode change check */
      ret_val = e_bin_edges (dkey, idfp, lngp->vsn, swpp->elo, swpp->ehi,
                              swpp->bin_units, swpp->n_lev);
      if (ret_val != ALL_OKAY){
          printf("PEACE3D: e_bin_edge fail\n");return PEACE3D_ERROR;
      }
      ret_val = invert_array(swpp->elo, swpp->n_lev);
      if(ret_val!=ALL_OKAY){
          printf("PEACE3D: invert e_bin fail\n");return PEACE3D_ERROR;
      }
      ret_val = invert_array(swpp->ehi, swpp->n_lev);
      if(ret_val!=ALL_OKAY){
          printf("PEACE3D: invert e_bin fail\n");return PEACE3D_ERROR;
      }
      ret_val = check_bins (swpp->elo, swpp->ehi, swpp->n_lev);/*finite bins*/
      if (ret_val != ALL_OKAY){
          printf("PEACE3D: check_bins failed\n"); return PEACE3D_ERROR;
      }
      copy_lineage(swpp->lineage,lngp);
   }
      /***************************
      * use sensor 0 to fill btime, etime and azimuths
      ****************************/
   if (isen == 0){
      if (is_spinpad){  /* for Spinpad use spin start/stop */
	 set_times_from_idf(&dummytime,&dummytime,&(swpp->btime),idfp);
	 swpp->etime = swpp->btime;
         tspin = idfp->spin_rate/1000.;
         swpp->etime.sec  = swpp->btime.sec + 1.e-9*swpp->btime.nano + tspin ; 
	                                                /*trunc tspin to int*/
         swpp->etime.nano = 1000000000*(swpp->btime.sec +
	               1.e-9*swpp->btime.nano+tspin -swpp->etime.sec);                  
      } else {
	 set_times_from_idf(&(swpp->btime),&(swpp->etime),&dummytime,idfp);
      }
         /* az range for full swp; change from look to flow and mod 360 */
      swpp->phi_lo = mod360(idfp->start_az[0]+180.0);
      swpp->phi_hi = mod360(idfp->stop_az[swpp->n_lev -1]+180.0);  
   }
   if (!have_bins){         /* fill theta arrays sensor by sensor           */
      if (is_spinpad){      /* SPINPAD returns flow directions, ordered     */
                            /* 180->0, so fill sensors in reverse order.    */
			    /* These will then work with                    */
			    /* re-binning, etc. which assumes ascending bins*/
         swpp->thlo[isen_spad] = idfp->start_theta;
	 swpp->thhi[isen_spad] = idfp->stop_theta;
      } else {              /* other Virtuals return LOOK direction, ordered*/
                            /* 180->0, so fill sensors in order but flip    */
			    /* angles; NB start -> 180 - stop, etc.                                       */
       	 swpp->thlo[isen] = 180.0-idfp->stop_theta;
	 swpp->thhi[isen] = 180.0-idfp->start_theta;
      }
   }
   /*************************************
   * fill sweep psd
   *****************************************/
   if (swp_step0 != idfp->swp_data[0] ||
       swp_step1 != idfp->swp_data[1] ||
       swpp->n_lev != idfp->num_sample){
       printf("PEACE3D: fill sweep psd lev check fail\n");
       return PEACE3D_ERROR;
   }
   ret_val = fill_single_sen_sweep(swpp,isen,dkey,idfp, do_pa);
       /* NB fill_single_sen_sweep uses isen to fill nsen-1-isen for spinpad */
   if (ret_val!= ALL_OKAY){
       printf("PEACE3D: fill_single_sweep error\n");
       return PEACE3D_ERROR;
   }
}                       /* end for loop over sensors */
return ALL_OKAY;
}

/**********************************************************************/
SDDAS_SHORT fill_single_sen_sweep(PEACE_SWEEP *swpp, int isen,
               SDDAS_ULONG dkey, struct idf_data *idfp, 
	       T_OR_F do_pa){
/**********************************************************************/

/* Get sweep data and, if requested, pa data from sensor isen
*  Fill swpp[isen_to_fill] where isen_to_fill is isen for most 
*      virtual instruments, but reverse (nsen -1 -isen) for SPINPAD */

SDDAS_SHORT tbl_oper[MAX_TABLES_TO_APPLY];
SDDAS_CHAR num_tbls, tbls_to_apply[MAX_TABLES_TO_APPLY];

SDDAS_SHORT ret_val;

float conv_data[MAX_LEVELS];
int j,n, nlev, isen_to_fill;

if(strncmp("SPINPAD",swpp->lineage->instrument,7)==0) {
   isen_to_fill = swpp->n_sen -1 - isen;
} else {
   isen_to_fill = isen;
}
nlev = swpp->n_lev;

ret_val = set_psd_tables(swpp->lineage,swpp->psd_units, &num_tbls,
                     tbls_to_apply, tbl_oper);
if (ret_val != ALL_OKAY){
   printf("PEACE3D: Couldn't set psd tables in fill_single_sweep\n");
   return PEACE3D_ERROR;
}

ret_val = convert_to_units (dkey,swpp->lineage->exten,
              swpp->lineage->vsn, idfp, isen, SENSOR, 0,
              num_tbls, tbls_to_apply, tbl_oper, conv_data, 0, 0);
if (ret_val != ALL_OKAY){
    printf("PEACE3D: convert_to_units returned error %d\n", ret_val);
    return PEACE3D_ERROR;
}

           /***********************************************
           * copy data in inverse order into psd array. Note
           * swpp->psd is [n_sen][n_lev] array so for sensor isen, data is
           * at psd[n_lev * isen + 0] thru [n_lev * isen + (n_lev -1)]
           *************************************************/

for (j=nlev*isen_to_fill, n=nlev -1; n>=0; j++,n--){
   swpp->psd[j] = conv_data[n];
}

if (do_pa){/* if pitch angles requested, do now. Just like psd really */
  if (idfp->num_pitch != nlev){
    printf("Pitch angles requested but idfp->num_pitch = %d != n_lev = %d\n",
                     idfp->num_pitch, nlev);
    return QPEACE_PITCH_ANGLE_ERROR;
  } 
  ret_val = convert_to_units (dkey,swpp->lineage->exten,
              swpp->lineage->vsn, idfp, isen, PITCH_ANGLE, 0,
              0, tbls_to_apply, tbl_oper, conv_data, 0, 0);
  if (ret_val != ALL_OKAY){
      printf("PEACE3D: convert_to_units returned error %d\n", ret_val);
      return PEACE3D_ERROR;
  }
  for (j=nlev*isen_to_fill, n=nlev -1; n>=0; j++,n--){
      swpp->pa[j] = conv_data[n];
  }
}
return ALL_OKAY;
}


/******************************************************************/
SDDAS_SHORT GetPeaceSpin(IDFS_LINEAGE *lineage, IDFS_TIME *btime,
   BIN_UNITS bin_units, PSD_UNITS psd_units, T_OR_F do_pa, 
   PEACE_SPIN3D *spinp){
/******************************************************************/
/*  Get a complete spin of PEACE 3D data. This module is self-contained
*   and deals with all IDFS calls. It needs to be passed the pointer
*   to a declared spin structure, but it mallocs all the necessary
*   internal storage required. free_spin_data will free all such
*   memory.
*
*   Inputs are self-evident from their declarations
*   Output is filled spin structure
*/

void *idf_creator, *promote_struct;
struct idf_data *idfp;
SDDAS_ULONG dkey,pa_dkey;
SDDAS_SHORT ret_val,pa_status;

int n_sen, n_lev,i;
int ipromote;

dbInitialize ();  /* new version of IDFS with libdbSQL... needed */
CfgInit ();
init_idfs();
get_version_number (&(lineage->vsn));
ret_val = create_idf_data_structure(&idf_creator);
if (ret_val != ALL_OKAY){
     printf("Error %d in create idf struct\n", ret_val);
     return PEACE3D_ERROR;
}
idfp = (struct idf_data *)idf_creator;

ret_val = get_data_key(lineage->project, lineage->mission,
                lineage->experiment, lineage->instrument,
                lineage->virtual_inst, &dkey);
if (ret_val != ALL_OKAY){
       printf("Error getting data key\n");
       free_experiment_info();
       return PEACE3D_ERROR;
}
/* kludge here, because pitch_angle_source_status requires parent file_open,
*  so can't easily then do promote */
if(do_pa){
   if(strncmp(lineage->project,"CLUST",5)==0){
      ret_val = get_data_key(lineage->project, lineage->mission,
                      "FGM", "VEC","CFUNIT", &pa_dkey);
   } else if (strncmp(lineage->project,"DOUBLEST",8)==0){
      ret_val = get_data_key(lineage->project, lineage->mission,
                      "FGM", "VEC","DFUNIT", &pa_dkey);
   } else {
      ret_val = PEACE3D_ERROR;
   } 	   
   if (ret_val != ALL_OKAY){
	  printf("Error getting pa_data key\n");
	  free_experiment_info();
	  return PEACE3D_ERROR;
   }
}

/* backup 1 spin so init_spin3d can advance */
spinp->btime = *btime;
add_secs_to_idfs_time( &(spinp->btime), -PEACE_SPIN_PERIOD);
spinp->etime = *btime;
add_secs_to_idfs_time( &(spinp->etime), 1.0);

if (DIAG_PRINT_3D){
  printf("\nRequested time:     ");print_idfs_time(btime);printf("\n");
  printf("Backed up 1 spin to:");print_idfs_time(&(spinp->btime));printf("\n");
}

/* do a promote if needed; twice */

for(ipromote=0; ipromote<2; ipromote++){
  promote_struct = pro_InitPromote (
     spinp->btime.year, spinp->btime.doy, 1000*spinp->btime.sec, 
     spinp->etime.year, spinp->etime.doy, 1000*spinp->etime.sec, 
     SHOW_PROMOTE_STATUS_XWIN);
  pro_AddSource (promote_struct, dkey);
  /* kludge here, because pitch_angle_source_status requires parent file_open,
  *  so can't easily do promote afterwards. Only other option I think is
  *  to read pa_inst lineage directly from VIDF; this would be more generic.*/
  if(do_pa){ pro_AddSource (promote_struct, pa_dkey);}

  if (pro_NeedToPromote (promote_struct) == sTrue) {
      printf("Trying to promote data: Pass %d\n",ipromote);
      pro_PromoteAll (promote_struct);
  }
  pro_FreePromote (promote_struct);
}

ret_val = init_idf_file (spinp->btime, spinp->etime, 
                           dkey, idfp, lineage->vsn);
if(ret_val!=ALL_OKAY){
   printf("Error init_idf_file\n");
   free_experiment_info();
   return PEACE3D_ERROR;
}

if(do_pa){  /* check status for pa data if requested; must do after file_open */
   ret_val = pitch_angle_source_status(
           dkey,lineage->exten,lineage->vsn,&pa_status, &pa_dkey);
   if ( ret_val!= ALL_OKAY || pa_status != PA_READY ){
	printf("Pitch angle source error: return val = %d; pa_status = %d",
        	ret_val, pa_status);
	printf(" Ignoring pitch angle request\n");
	do_pa = QPFALSE;
   }
} 

/* set spin start, period, n_swp, find n_sen, n_lev */
ret_val = init_spin3d(spinp,lineage,idfp,dkey, &n_sen, &n_lev, do_pa);
if(ret_val!=ALL_OKAY){
   printf("Error init_spin3d\n");
   free_experiment_info();
   return PEACE3D_ERROR;
}

if (DIAG_PRINT_3D){
  print_peace_spin_info(spinp);
  printf("%d energy levels per sweep\n",n_lev);
}

ret_val = malloc_spin_sweep(spinp, n_sen, n_lev, do_pa);
if (ret_val != ALL_OKAY){
       printf("Error malloc spin sweep\n");
       free_experiment_info();
       return PEACE3D_ERROR;
}


ret_val = fill_spin3d(spinp,lineage,idfp,dkey, bin_units,
                                     psd_units,PEACE_TZERO, do_pa);
if (ret_val != ALL_OKAY){
   printf("Error fill_spin3d returned %d \n",ret_val);
   free_spin_data(spinp);
   free_experiment_info();
   return PEACE3D_ERROR;
}
free_experiment_info();
if (DIAG_GET_SPIN){
   printf("\nGetPeaceSpin returning with\n");
   print_peace_spin_info(spinp);
   for (i=0; i<spinp->n_swp; i++){
      printf("Sweep %d in spin is\n",i);
      print_sweep_structure(&(spinp->sweep[i]));
   }
}
return ret_val;

}

/******************************************************************/
SDDAS_SHORT GetN1PeaceSpins(
   IDFS_LINEAGE *lineage, 
   IDFS_TIME *btime,
   IDFS_TIME *etime,
   BIN_UNITS bin_units, 
   PSD_UNITS psd_units, 
   T_OR_F do_pa,
   int *nspins, 
   PEACE_SPIN3D ***spinppp){
/******************************************************************/

/*  Get all PEACE data within times specified from ONE IDFS file
*   Put each spin in a PEACE_SPIN3D structure and return an array of
*   pointers which points to them.
*
*   This module is self-contained
*   and deals with all IDFS calls. It returns the pointer
*   to an array of spin structure pointers, and mallocs all the necessary
*   internal storage required, including the pointer array.
*
*   Inputs are self-evident from their declarations
*   Output is malloced and filled pointer array (and the nspins spin
*   structures to which they point!).
*/

void *idf_creator, *promote_struct;
struct idf_data *idfp;
SDDAS_ULONG dkey,pa_dkey;
SDDAS_SHORT ret_val,pa_status;
PEACE_SPIN3D spin4init, *s4ip, *cursp, **spinpp;
IDFS_TIME bt,et,lasttime,dummytime;

int n_sen, n_lev, nsp, isp, nswpinsp,i;
int ipromote;

dbInitialize ();  /* new version of IDFS with libdbSQL... needed */
CfgInit ();
init_idfs();
get_version_number (&(lineage->vsn));
ret_val = create_idf_data_structure(&idf_creator);
if (ret_val != ALL_OKAY){
     printf("Error %d in create idf struct\n", ret_val);
     return PEACE3D_ERROR;
}
idfp = (struct idf_data *)idf_creator;

ret_val = get_data_key(lineage->project, lineage->mission,
                lineage->experiment, lineage->instrument,
                lineage->virtual_inst, &dkey);
if (ret_val != ALL_OKAY){
       printf("Error getting data key\n");
       free_experiment_info();
       return PEACE3D_ERROR;
}
/* kludge here, because pitch_angle_source_status requires parent file_open,
*  so can't easily then do promote */
if(do_pa){
   if(strncmp(lineage->project,"CLUST",5)==0){
      ret_val = get_data_key(lineage->project, lineage->mission,
                      "FGM", "VEC","CFUNIT", &pa_dkey);
   } else if (strncmp(lineage->project,"DOUBLEST",8)==0){
      ret_val = get_data_key(lineage->project, lineage->mission,
                      "FGM", "VEC","DFUNIT", &pa_dkey);
   } else {
      ret_val = PEACE3D_ERROR;
   } 	   
   if (ret_val != ALL_OKAY){
	  printf("Error getting pa_data key\n");
	  free_experiment_info();
	  return PEACE3D_ERROR;
   }
}

s4ip = &spin4init;

s4ip->btime             = *btime;
/* back up 1 spin, so init_spin3d can advance 1 spin */
add_secs_to_idfs_time(&(s4ip->btime), -PEACE_SPIN_PERIOD);
s4ip->etime = lasttime  = *etime;

/* do a promote if needed; twice */

for(ipromote=0; ipromote<2; ipromote++){
  promote_struct = pro_InitPromote (
     s4ip->btime.year, s4ip->btime.doy, 1000*s4ip->btime.sec, 
     s4ip->etime.year, s4ip->etime.doy, 1000*s4ip->etime.sec, 
     SHOW_PROMOTE_STATUS_XWIN);
  pro_AddSource (promote_struct, dkey);
  /* kludge here, because pitch_angle_source_status requires parent file_open,
  *  so can't easily do promote afterwards. Only other option I think is
  *  to read pa_inst lineage directly from VIDF; this would be more generic.*/
  if(do_pa){   pro_AddSource (promote_struct, pa_dkey); }

  if (pro_NeedToPromote (promote_struct) == sTrue) {
      printf("Trying to promote data: Pass %d\n",ipromote);
      pro_PromoteAll (promote_struct);
  }
  pro_FreePromote (promote_struct);
}

ret_val = init_idf_file (s4ip->btime, s4ip->etime, dkey, idfp, lineage->vsn);
if(ret_val!=ALL_OKAY){
   printf("Error init_idf_file\n");
   free_experiment_info();
   return PEACE3D_ERROR;
}

if(do_pa){  /* check status for pa data if requested; must do after file_open */
   ret_val = pitch_angle_source_status(
           dkey,lineage->exten,lineage->vsn,&pa_status, &pa_dkey);
   if ( ret_val!= ALL_OKAY || pa_status != PA_READY ){
	printf("Pitch angle source error: return val = %d; pa_status = %d",
        	ret_val, pa_status);
	printf(" Ignoring pitch angle request\n");
	do_pa = QPFALSE;
   }
} 

/* set spin start, period, n_swp, find n_sen, n_lev */
ret_val = init_spin3d(s4ip,lineage,idfp,dkey, &n_sen, &n_lev, do_pa);
if(ret_val!=ALL_OKAY){
   printf("Error init_spin3d\n");
   free_experiment_info();
   return PEACE3D_ERROR;
}

if (DIAG_PRINT_3D){
  print_peace_spin_info(s4ip);
  printf("%d energy levels per sweep\n",n_lev);
}
nswpinsp = s4ip->n_swp;

nsp = (int)((idfs_time_to_epoch_secs(etime) - 
                idfs_time_to_epoch_secs(btime))/(idfp->spin_rate/1000.));
nsp++;  /* expect at most nsp spins */

spinpp = (PEACE_SPIN3D**)malloc(nsp*sizeof(PEACE_SPIN3D*));
if (spinpp == NULL){
     printf("GetN1PeaceSpins failed to malloc spinpp with %d spin ptrs\n",nsp);
     free_experiment_info();  return PEACE3D_ERROR;
}

*spinppp = spinpp;  /* pass the address of the ptr array back */

for (i=0; i<nsp; spinpp[i++]=NULL);  /* initialise ptrs to NULL */
bt = *btime;
et = *etime;
isp = 0;

while (compare_idfs_time(&bt,&lasttime)!=T1_LATER &&
     ret_val == ALL_OKAY && isp < nsp){
               /* prepare spin; set n_swp, spin start/stop */
     cursp = spinpp[isp]=(PEACE_SPIN3D *)malloc(sizeof(PEACE_SPIN3D));
     if (cursp == NULL){
         printf("GetN1Spins failed to malloc spin structure\n");
	 ret_val = PEACE3D_ERROR;
	 break;
     }
      
     cursp->n_swp = nswpinsp;
     cursp->tspin      = idfp->spin_rate/1000.;
     set_times_from_idf(&dummytime, &dummytime, &(cursp->btime), idfp);
     cursp->etime = cursp->btime;
     cursp->etime.sec  = cursp->btime.sec + 1.e-9*cursp->btime.nano
                            + cursp->tspin;  /* truncates tspin to int */
     cursp->etime.nano = 1000000000*(
        cursp->btime.sec + 1.e-9*cursp->btime.nano +cursp->tspin
        - cursp->etime.sec);

     ret_val = malloc_spin_sweep(spinpp[isp], n_sen, n_lev, do_pa);
     if (ret_val != ALL_OKAY){
	printf("Error malloc spin sweep\n");
	ret_val = PEACE3D_ERROR;
	break;
     }
          
     ret_val = fill_spin3d(cursp,lineage,idfp,dkey, bin_units,
                                	  psd_units,PEACE_TZERO, do_pa);
     if(DIAG_PRINT_3D){
       printf("spin %d: ",isp);
       printf("\nsun_sen %ld spin_rate %ld\n", idfp->sun_sen, idfp->spin_rate);
     }
     if (ret_val == PEACE3D_INCOMPLETE_SPIN){
	   printf("Incomplete spin found; trying to get next one\n"); 
           free_spin_data(cursp);  /* undoes malloc_spin_sweep;  */
	                           /* skip to end gathering while */
	   ret_val = ALL_OKAY;     /* reset so while will run!    */
     } else if (ret_val == NEXT_FILE_STATUS){
           printf("NEXT_FILE_STATUS returned from fill_spin3d");
	   printf(" remove current spin and return\n");
	   free_spin_data(cursp);
	   break;
     } else  if (ret_val != ALL_OKAY){
	printf("Error fill_spin3d returned %d \n",ret_val);
	ret_val = PEACE3D_ERROR;
	break;          /* other errs not tolerated  */
     } else {
	/* set spin stop time to stop of last sweep in spin; go to next spin */
	cursp->etime = cursp->sweep[nswpinsp-1].etime;
	ret_val = advance_idf_file_spin_start(dkey,idfp,lineage,&bt,&et,0); 
                                                            /*use sen 0*/
	if(ret_val != ALL_OKAY){
	     printf("Advance failed returning %d\n",ret_val);
	     break;
	}
	if (idfp->num_sample != n_lev){
	     printf("Number of levels)  changed from %d to %d.  ",
	        	n_lev,idfp->num_sample);
	     printf("Mode change not supported\n");
	     ret_val = PEACE3D_ERROR;
	     break;
	}
	isp++;   /* doesn't incr if hit incomplete spin */
     }
} /* end while gathering spins */

if (ret_val != ALL_OKAY || isp <= 0 || isp > nsp){
    printf("GetN1Spins Error after gathering while: Returned %d with isp =%d",
                ret_val, isp);
    printf(" out of %d expected\n",nsp);
    printf("Current spin start time = ");
    print_idfs_time(&(cursp->btime));printf("\n");
    if (ret_val == NEXT_FILE_STATUS && isp > 0 && isp <=nsp){
        printf("NEXT_FILE_STATUS reached; will try to fetch more");
    } else {
        free_n_spins(spinpp,nsp);
        free(spinpp); *spinppp=NULL;   /* pass back NULL ptr for safety */
        return PEACE3D_ERROR;
    }
}
*nspins = isp;  /* set to actual number of spins filled; this is
                   also the number of structures malloced, 
		   although the size of the pointer array could be
		   bigger */
free_experiment_info();
return ret_val;

}

/********************************************************************/
SDDAS_SHORT GetNPeaceSpins(
   IDFS_LINEAGE *lineage, 
   IDFS_TIME *btime,
   IDFS_TIME *etime,
   BIN_UNITS bin_units, 
   PSD_UNITS psd_units, 
   T_OR_F do_pa,
   int *nspins, 
   PEACE_SPIN3D ***spinppp){
/******************************************************************/

/*  Get all PEACE data within times specified.
*   Put each spin in a PEACE_SPIN3D structure and return an array of
*   pointers which points to them.
*
*   This module is self-contained
*   and deals with all IDFS calls. It returns the pointer
*   to an array of spin structure pointers, and mallocs all the necessary
*   internal storage required, including the pointer array.
*
*   Inputs are self-evident from their declarations
*   Output is malloced and filled pointer array (and the nspins spin
*   structures to which they point!).
*   
*   This routine makes multiple calls as necessary to GetN1PeaceSpins 
*   which returns all the spins up to the end of a single IDFS file.
*/

int nfiles,i,j,nspin1[N_MAX_IDFS_FILES], nspintot,isp;

PEACE_SPIN3D **spin1file[N_MAX_IDFS_FILES];
/* spin1file is an array[N_...] of ptrs to ptrs to SPIN3D structs
*  so spin1file[i] is a PEACE_SPIN3D ** the address of which is what 
*  GetN1PeaceSpins expects so it can malloc an array of ptrs to SPIN3D
*  structs. These are accessed as (spin1file[i])[j] which is a ptr to the
*  jth spin from file i */


IDFS_TIME start_in_file_time;

SDDAS_SHORT ret;

nfiles = 0;
nspintot = 0;

start_in_file_time = *btime;

ret = NEXT_FILE_STATUS;   /* start */
while (ret == NEXT_FILE_STATUS && nfiles < N_MAX_IDFS_FILES){
   printf("ready\n");
   ret = GetN1PeaceSpins(lineage,&start_in_file_time,etime,bin_units,psd_units,
      do_pa, &(nspin1[nfiles]), &(spin1file[nfiles]));
   if (ret == PEACE3D_ERROR) break;
   printf("keep going\n");
   start_in_file_time = (spin1file[nfiles])[nspin1[nfiles]-1]->etime;
   add_secs_to_idfs_time(&start_in_file_time,
       2.0*(spin1file[nfiles])[nspin1[nfiles]-1]->tspin);/*add a bit;goto next file */
   printf("start time is ");print_idfs_time(&start_in_file_time); printf("\n");
   nspintot += nspin1[nfiles];
   printf("Finished file number %d\n",nfiles);
   nfiles++;
}


if (nfiles == 0) {
    return PEACE3D_ERROR;
 
} else if (nfiles ==1){   /* if only 1 file must be good or nfiles ==0 */
    *spinppp = spin1file[0];
    *nspins  = nspin1[0];
    return ALL_OKAY;
   
} else {     /* copy all the bits from each file into a single array */
    *spinppp = (PEACE_SPIN3D**)malloc(nspintot*sizeof(PEACE_SPIN3D*));
    if (*spinppp ==  NULL){   /* trap failed malloc and clean up */
       for (i=0; i<nfiles; i++){
	   free_n_spins(spin1file[i],nspin1[i]);
	   free(spin1file[i]); spin1file[i]=NULL;
	   return PEACE3D_ERROR;
       }
    }
    isp=0;
    for (i=0; i<nfiles; i++){
      for (j=0; j<nspin1[i]; j++){
	 (*spinppp)[isp] = copy_spin ((spin1file[i])[j]); 
	 if ((*spinppp)[isp]!=NULL)  isp++;  /* else ignore */
      }
      free_n_spins(spin1file[i],nspin1[i]);
      printf("spins in spin1file[%d] freed; ",i);fflush(NULL);
      free(spin1file[i]); spin1file[i]= NULL;
      printf("spin1file[%d] freed\n",i);fflush(NULL); 
    }
    *nspins = isp;
    printf("GetNPeaceSpins returning\n");fflush(NULL);
    return ALL_OKAY;
}

}



/***************************************************************************/
PEACE_SWEEP *PeaceGetAzSlice(
       IDFS_LINEAGE *lineage, IDFS_TIME *btime, float az,
       BIN_UNITS bin_units,
       PSD_UNITS psd_units,
       PEACE_PITCH_ANGLE_OPTS *paopts){
/**************************************************************************/
/* peace_az_slice routine to grab an az slice.
*
*      Inputs:
*        idfs lineage for sweeps [incl ext but not vsn]
*        time sweeps
*        azimuth for sweeps
*
*      Returns pointer to complete sweep structure
*      If PAD requested, return that in a sweep structure
*/

PEACE_SPIN3D spin_data, *spin_datap;
SDDAS_SHORT ret_val;
PEACE_SWEEP *swp=NULL, *swppa = NULL, **swppap=NULL;

int iswp;

if (strcmp(lineage->instrument,"PAD") == 0){
  swp = (PEACE_SWEEP *)malloc(sizeof(PEACE_SWEEP));
  if (swp == NULL){
     printf("Failed to malloc sweep structure\n");
     return NULL;
  }
  nullify_sweep_arrays(swp);
  ret_val = GetPeacePAD(lineage,btime,bin_units,psd_units,paopts->do_pa, swp);
  if (ret_val != ALL_OKAY){
     printf("GetPeacePAD returned error %d\n",ret_val);
     if (swp != NULL) {free_sweep_arrays(swp);free(swp);}
     return NULL;
  }
  if (paopts->do_pa){
      ret_val=rebin_sweep_to_pa(&swp,1,paopts,&swppa);
      if (DIAG_PLOTAZ_REBIN_PA){
	 printf("GetAzSlice PAD got\n");
	 print_sweep_structure(swp);
	 printf("Rebinned in pitch angle to\n");
	 print_sweep_structure(swppa);
      }
      if (ret_val != ALL_OKAY || swppa == NULL){
         printf("Rebinning PAD to pa failed; returning orig. PAD\n");
	 if (swppa!=NULL){ free_sweep_arrays(swppa); free(swppa);}
	 return swp;
      } else {
         free_sweep_arrays(swp); free(swp);
	 return swppa;
      }      
  } else {
      return swp;
  } 
} else {
  ret_val = GetPeaceSpin(lineage,btime,bin_units,psd_units,
                                     paopts->do_pa, &spin_data);
  if (ret_val != ALL_OKAY){
    printf("Failed to get 1st spin\n");
    return NULL;
  }
  if (paopts->do_pa){
     spin_datap = &spin_data; /* rebin called to fill 1 swp */
     ret_val = Rebin_n_spins_to_pa(&spin_datap,1,paopts, &swppap);
     if (ret_val != ALL_OKAY || *swppap == NULL){
         printf("Rebinning spin sweeps to pa failed\n!");
	 if (swppap!=NULL){ free_sweep_arrays(*swppap); free(*swppap);}
	 return NULL;
     } else {
         free_spin_data(&spin_data);
	 return *swppap;
     }
  } else { 
      iswp = find_az_sweep(&spin_data, az);
      if (iswp < 0){
	printf("requested az %g not found in spin\n", az);
	free_spin_data(&spin_data);
	return NULL;
      }
      swp = create_sweep_full_copy(&spin_data.sweep[iswp]);
      free_spin_data(&spin_data);
      return swp;
  }
}
/* outermost braces for routine */
/* shouldn't get here !*/
printf("ERROR: Control got to end of PEACEGetAzSlice without returning\n");
return NULL;

}

/******************************************************************/
SDDAS_SHORT GetN1PeaceSweeps(
        IDFS_LINEAGE *lngp, 
	IDFS_TIME *btimep,
        IDFS_TIME *etimep,
	BIN_UNITS bin_units, 
	PSD_UNITS psd_units,
	T_OR_F do_pa,
	int *nswps, 
	PEACE_SWEEP ***swpppp){
/******************************************************************/
/*  Get all PEACE sweep data within times specified. 
*   Put them each in a PEACE_SWEEP structure and return an array of
*   pointers which points to them. Each sweep structure includes its
*   own copies of bin boundaries and lineage.
*   
*   This module is self-contained
*   and deals with all IDFS calls. It returns the pointer
*   to an array of sweep structure pointers, and mallocs all the necessary
*   internal storage required, including the pointer array. 
*
*   Inputs are self-evident from their declarations
*   Output is malloced and filled pointer array (and the nsweeps, sweep 
*   structures to which they point!). 
* 
*   If do_pa = QPTRUE, pitch angle array is filled in sweep structure.
*   This routine does NOT call rebin_sweep_to_pa
*/

void *idf_creator, *promote_struct;
struct idf_data *idfp;
SDDAS_ULONG dkey, pa_dkey;
SDDAS_SHORT ret_val,pa_status;
IDFS_TIME etime, btime, lasttime;
PEACE_SWEEP *curswp, **swppp;
PEACE_SPIN3D spin3d;
int i, iswp, nswp;   /* nswp is local copy of *nswps */
int nsen, nlev;
int ipromote;

btime = *btimep;

if (strcmp(lngp->instrument,"PAD")==0){
   return GetNPeacePADs(lngp,btimep,etimep,
            bin_units,psd_units,do_pa, nswps,swpppp);
}

if(DIAG_PRINT_3D){
  printf("3D bin units requested: %d <=> %s\n", 
             bin_units,BIN_UNITS_STRING(bin_units));
  printf("3D psd units requested: %d <=> %s\n", 
             psd_units,PSD_UNITS_STRING(psd_units));
	     
}

dbInitialize ();   /* I need a newer version of IDFS with libdbSQL...*/
CfgInit ();
init_idfs();
get_version_number (&(lngp->vsn));

ret_val = create_idf_data_structure(&idf_creator);
if (ret_val != ALL_OKAY){
     printf("Error %d in create idf struct for sweep\n", ret_val);
     free_experiment_info(); return ret_val;
}
idfp = (struct idf_data *)idf_creator;

ret_val = get_data_key(lngp->project, lngp->mission,
                lngp->experiment, lngp->instrument,
                lngp->virtual_inst, &dkey);
if (ret_val != ALL_OKAY){
       printf("Error getting data key\n");
       free_experiment_info(); return ret_val; 
}
/* kludge here, because pitch_angle_source_status requires parent file_open,
*  so can't easily then do promote */
if(do_pa){
   if(strncmp(lngp->project,"CLUST",5)==0){
      ret_val = get_data_key(lngp->project, lngp->mission,
                      "FGM", "VEC","CFUNIT", &pa_dkey);
   } else if (strncmp(lngp->project,"DOUBLEST",8)==0){
      ret_val = get_data_key(lngp->project, lngp->mission,
                      "FGM", "VEC","DFUNIT", &pa_dkey);
   } else {
      ret_val = PEACE3D_ERROR;
   } 	   
   if (ret_val != ALL_OKAY){
	  printf("Error getting pa_data key\n");
	  free_experiment_info();
	  return PEACE3D_ERROR;
   }
}


/* backup 1 spin, so init_spin3d can advance 1 spin; kludge, but works? */
add_secs_to_idfs_time( &btime, -PEACE_SPIN_PERIOD); 
etime=btime;
add_secs_to_idfs_time( &etime, 1.0);  /* just for file initialisation */

/* do a promote if needed; twice */

for(ipromote=0; ipromote<2; ipromote++){
  promote_struct = pro_InitPromote (
     btime.year, btime.doy, 1000*btime.sec, 
     etime.year, etime.doy, 1000*etime.sec, 
     SHOW_PROMOTE_STATUS_XWIN);
  pro_AddSource (promote_struct, dkey);
  /* kludge here, because pitch_angle_source_status requires parent file_open,
  *  so can't easily do promote afterwards. Only other option I think is
  *  to read pa_inst lineage directly from VIDF; this would be more generic.*/
  if(do_pa){  pro_AddSource (promote_struct, pa_dkey);}

  if (pro_NeedToPromote (promote_struct) == sTrue) {
      printf("Trying to promote data: Pass %d\n",ipromote);
      pro_PromoteAll (promote_struct);
  }
  pro_FreePromote (promote_struct);
}

ret_val = init_idf_file (btime,etime, dkey, idfp, lngp->vsn);
if(ret_val!=ALL_OKAY){
   printf("Error init_idf_file\n");
   free_experiment_info(); return ret_val;
}

if(do_pa){  /* check status for pa data if requested; must do after file_open */
   ret_val = pitch_angle_source_status(
           dkey,lngp->exten,lngp->vsn,&pa_status, &pa_dkey);
   if ( ret_val!= ALL_OKAY || pa_status != PA_READY ){
	printf("Pitch angle source error: return val = %d; pa_status = %d",
        	ret_val, pa_status);
	printf(" Ignoring pitch angle request\n");
	do_pa = QPFALSE;
   }
} 
/* initialise nsen and nlev using spin3d routines;
*  also spin3d.n_swp will get the number of sweeps/spin*/

ret_val = init_spin3d( &spin3d, lngp, idfp, dkey, &nsen, &nlev, do_pa);

/* get several sweeps. Expect nswp = (swps/spin)*dt/tspin */

nswp = (int)(spin3d.n_swp*(idfs_time_to_epoch_secs(etimep) - 
                idfs_time_to_epoch_secs(btimep))/(idfp->spin_rate/1000.));
nswp++;  /* expect at most nswp sweeps */

if (strncmp("SPINPAD",lngp->instrument,7)== 0)nswp *=3;  /*overkill for HAR*/

if(DIAG_PRINT_3D){printf("Expecting at most %d sweeps\n",nswp);}
if (nswp < 1){
   printf("GetNPeaceSweeps nsweeps = %d out of range\n", nswp);
   free_experiment_info(); return PEACE3D_ERROR;
}
swppp= (PEACE_SWEEP**)malloc(nswp*sizeof(PEACE_SWEEP*));
if (swppp == NULL){
   printf("GetN1PeaceSweeps failed to malloc swppp with %d sweep ptrs\n",nswp);
   free_experiment_info();  return PEACE3D_ERROR;
}

*swpppp = swppp;  /* pass the address of the ptr array back */

for (i=0; i<nswp; swppp[i++]=NULL);  /* initialise ptrs to NULL */

btime = *btimep;
lasttime = *etimep;
iswp = 0;

while ((compare_idfs_time(&btime,&lasttime)!=T1_LATER )&&
     (ret_val == ALL_OKAY && iswp < nswp)){

     curswp = swppp[iswp] = (PEACE_SWEEP*)malloc(sizeof(PEACE_SWEEP));
     if (curswp == NULL){
        printf("GetN1PeaceSweeps failed to malloc sweep structure\n");
	break;
     }
     nullify_sweep_arrays(curswp);  /* set ptrs inside swp to NULL for safety*/
     
     curswp->n_sen = nsen;
     curswp->n_lev = nlev;
     curswp->bin_units = bin_units;
     curswp->psd_units = psd_units;
     curswp->phi_frame = PEACE_TZERO;    

     ret_val = malloc_sweep_arrays(curswp, do_pa);
     if (ret_val != ALL_OKAY){
	  printf("malloc_sweep_arrays failed\n");
	  break;
     }
     
     curswp->missing_value = QP_DEFAULT_MISSING;

     /* fill sweep. penult. arg is have_bins: QPFALSE (0) gets bins & lineage*/
     ret_val=fill_sweep(curswp, lngp, idfp, dkey, QPFALSE, do_pa);
     if (ret_val != ALL_OKAY){
          /* try to cope with mode change which changes n_lev */
	  if (ret_val != SWEEP_NLEVEL_CHANGE){
            printf("fill_sweep returned %d\n", ret_val);
	    break;
	  } else {  /* clean up present sweep and try to continue */
	     free_sweep_arrays(curswp); free(curswp);iswp--;
	     /* attempt to reinitialise; spin3d not needed??? */
	     printf("Trying to reinitialise and continue\n If mode change ");
	     printf(" is SPINPAD MAR -> HAR may run out of alloc.\n");
             ret_val = init_spin3d( &spin3d, lngp, idfp, dkey, &nsen, 
	             &nlev, do_pa);   
	     continue;
	  }
	  
     }
     
     btime = curswp->btime;
 
     if(DIAG_PRINT_3D){
        printf("GetN1PeaceSweeps iswp = %d\n",iswp);
        print_sweep_structure(curswp);
     }
     iswp++; 
     
}  /* end while loop gathering sweeps */


if (ret_val != ALL_OKAY || iswp <= 0 || iswp > nswp){
    printf("GetN1PeaceSweeps Error after gathering while: ");
    printf(" Returned %d with iswp =%d out of %d expected\n", 
          ret_val, iswp,nswp);
    printf("Current spin start time = ");
    print_idfs_time(&btime);printf("\n");
    if (ret_val == NEXT_FILE_STATUS && iswp > 0 && iswp <= nswp){ 
        printf("NEXT_FILE_STATUS reached\n");
    } else {  
        free_n_sweeps(nswp, swppp); *swpppp = NULL;
        return PEACE_PAD_ERROR;
    }
}
*nswps = iswp;  /* set to actual number of sweeps filled; this is
                   also the number of sweep structures malloced, 
		   although the size of the pointer array could be
		   bigger */

free_experiment_info();
return ret_val;

}

/******************************************************************/
SDDAS_SHORT GetNPeaceSweeps(
        IDFS_LINEAGE *lngp, 
	IDFS_TIME *btimep,
        IDFS_TIME *etimep,
	BIN_UNITS bin_units, 
	PSD_UNITS psd_units,
	T_OR_F do_pa,
	int *nswps, 
	PEACE_SWEEP ***swpppp){
/******************************************************************/
/*  Get all PEACE sweep data within times specified. Extend across
*   several IDFS files.  
*   Put them each in a PEACE_SWEEP structure and return an array of
*   pointers which points to them. Each sweep structure includes its
*   own copies of bin boundaries and lineage.
*   
*   This module is self-contained
*   and deals with all IDFS calls. It returns the pointer
*   to an array of sweep structure pointers, and mallocs all the necessary
*   internal storage required, including the pointer array. 
*
*   Inputs are self-evident from their declarations
*   Output is malloced and filled pointer array (and the nsweeps, sweep 
*   structures to which they point!). 
* 
*   If do_pa = QPTRUE, pitch angle array is filled in sweep structure.
*   This routine does NOT call rebin_sweep_to_pa
*
*   This routine makes multiple calls as necessary to GetN1PeaceSweeps 
*   which returns all the sweeps up to the end of a single IDFS file.
*/

int nfiles,i,j,nswp1[N_MAX_IDFS_FILES], nswptot,iswp, tries;

PEACE_SWEEP **swp1file[N_MAX_IDFS_FILES];
/* swp1file is an array[N_...] of ptrs to ptrs to SWEEP structs
*  so swp1file[i] is a PEACE_SWEP ** the address of which is what 
*  GetN1PeaceSweeps expects so it can malloc an array of ptrs to SWEEP
*  structs. These are accessed as (swp1file[i])[j] which is a ptr to the
*  jth swp from file i */


IDFS_TIME start_in_file_time;

SDDAS_SHORT ret;

nfiles  = 0;
nswptot = 0;
tries   = 0;
start_in_file_time = *btimep;

ret = NEXT_FILE_STATUS;   /* start */
while (ret    == NEXT_FILE_STATUS && 
       nfiles <  N_MAX_IDFS_FILES && 
       tries  <  N_MAX_IDFS_START_NUDGES){
   printf("ready\n");
   nswp1[nfiles]=0;
   ret = GetN1PeaceSweeps(lngp,&start_in_file_time,etimep,bin_units,psd_units,
      do_pa, &(nswp1[nfiles]), &(swp1file[nfiles]));
   if (ret == PEACE3D_ERROR  || 
       ret == QUTILS_ERROR   || 
       ret == PEACE_PAD_ERROR ||
	  nswp1[nfiles]==0) { 
	  free_n_sweeps(nswp1[nfiles],swp1file[nfiles]);
	  if (nswp1[nfiles] > 0){
              free(swp1file[nfiles]); swp1file[nfiles] = NULL;
	  }
	  printf("GetN1PeaceSweeps returned error %d, nswp1 = %d\n",
                                                    ret,nswp1[nfiles]);
	  if (nfiles == 1){  /* we have at least 1 good file */
             add_secs_to_idfs_time(&start_in_file_time, 2.0*PEACE_SPIN_PERIOD);
	     tries++;
	     printf("Trying to nudge start time\n");
	     continue;
	  } else {
             printf("GetN1PeaceSweeps failed to find any data\n");
             break;
	  }
   }
   printf("keep going\n");
   start_in_file_time = (swp1file[nfiles])[nswp1[nfiles]-1]->etime;
   add_secs_to_idfs_time(&start_in_file_time,
    2.0*PEACE_SPIN_PERIOD); /*add a bit;goto next file */
   printf("start time is ");print_idfs_time(&start_in_file_time); printf("\n");
   nswptot += nswp1[nfiles];
   printf("Finished file number %d\n",nfiles);
   nfiles++;
}


if (nfiles == 0) {
    return PEACE3D_ERROR;
 
} else if (nfiles ==1){   /* if only 1 file must be good or nfiles ==0 */
    *swpppp = swp1file[0];
    *nswps  = nswp1[0];
    return ALL_OKAY;
   
} else {     /* copy all the bits from each file into a single array */
    *swpppp = (PEACE_SWEEP**)malloc(nswptot*sizeof(PEACE_SWEEP*));
    if (*swpppp == NULL){  /* trap bad malloc */
        for (i=0; i<nfiles; i++){
	   free_n_sweeps(nswp1[i],swp1file[i]);
	   free(swp1file[i]); swp1file[i] = NULL;
	   return PEACE3D_ERROR;
	}
    }
    iswp=0;
    for (i=0; i<nfiles; i++){
      for (j=0; j<nswp1[i]; j++){
	 (*swpppp)[iswp] = create_sweep_full_copy ((swp1file[i])[j]); 
	 if ((*swpppp)[iswp]!= NULL) iswp++; /* else ignore */
      }
      free_n_sweeps(nswp1[i],swp1file[i]);
      printf("sweeps in swp1file[%d] freed; ",i);fflush(NULL);
      free(swp1file[i]); swp1file[i]= NULL;
      printf("swp1file[%d] freed\n",i);fflush(NULL);
    }
    *nswps = iswp;
    printf("GetNPeaceSweeps returning\n");fflush(NULL);
    return ALL_OKAY;
}




}


/*********************************************************************/
SDDAS_SHORT init_spin3d(PEACE_SPIN3D *spin_dp,
                        IDFS_LINEAGE  *lngp,
                        struct idf_data *idfp,
                        SDDAS_ULONG dkey, int *n_sen, int *n_lev,
			T_OR_F do_pa){
/********************************************************************/

/* initialise info in spin3d structure by reading some data. In particular:
*
*  set spin_dp->tspin    spin period in secs
*      spin_dp->n_swp    sweeps per spin
*      spin_dp->btime    to actual start of spin
*      n_sen             to number of sensors
*      n_lev             to number of levels
*
*  advance file to spin start of NEXT spin!!! 
*/

SDDAS_SHORT ret_val, num_sensor;
SDDAS_LONG ret_vallong;
SDDAS_CHAR fwd, full_swp;
float dt_swp;                 /* I work in seconds!!!*/
int isen, is_spinpad;
IDFS_TIME dummytime, swpstart[NSWEEPS_4_SWEEPS_IN_SPIN];
int i, nswps_to_try;

if (strncmp("SPINPAD",lngp->instrument,7)== 0)is_spinpad = 1;
else is_spinpad = 0;

nswps_to_try = NSWEEPS_4_SWEEPS_IN_SPIN;

ret_vallong = read_idf (dkey,lngp->exten, lngp->vsn,
                     (SDDAS_CHAR *) &num_sensor, _SEN, 0, 0, 1);
if (ret_vallong < 0) {
      printf ("\n Error %d from read_idf routine.\n", ret_val);
      return PEACE3D_ERROR;
    }
*n_sen = num_sensor;

idfp->filled_data = 0;          /* trick while loop into executing */
fwd = 0; full_swp = 1;isen = 0; /* read until find filled sen; don't adva t */
while (!idfp->filled_data && isen < num_sensor){
   ret_val = read_drec(dkey,lngp->exten,lngp->vsn, idfp,isen,fwd,full_swp);
   if (ret_val < 0){
            printf ("\nError %d from read_drec.\n", ret_val);
            return PEACE3D_ERROR;
   }
   isen++;
   if(DIAG_PRINT_3D){
      printf("read_drec'ed sen %d. ret_val = %d, filled_data = %d\n",
                      isen-1, ret_val, idfp->filled_data);
   }
}
if (isen == num_sensor  && !idfp->filled_data){
      printf("Failed to initialise; all sensors read but no data filled\n");
      return PEACE3D_ERROR;
}
*n_lev = idfp->num_sample;
isen -= 1;         /* sensor found; counter incremented once too many times */

if(DIAG_PRINT)printf("Initialise used sensor %d; there are %d n_samples\n",
            isen, *n_lev);


/* set spin period; DON'T reposition file*/

spin_dp->tspin      = idfp->spin_rate/1000.;
spin_dp->btime.year = idfp->byear;
spin_dp->btime.doy  = idfp->bday;
msec_nano_to_sec_nano(idfp->sun_sen,0L,    /* sun sens in msec, 0 nano */
                  &(spin_dp->btime.sec),&(spin_dp->btime.nano));
spin_dp->etime      = spin_dp->btime;
add_secs_to_idfs_time( &(spin_dp->etime), spin_dp->tspin);

if (DIAG_PRINT_3D){
  printf("Spin Start time:    "); print_idfs_time(&spin_dp->btime);printf("\n");
  printf("Spin End time:      "); print_idfs_time(&spin_dp->etime);printf("\n");
}
if (DIAG_PRINT)print_idf_data(idfp);

/* use sweep we just got and then get some more; re-read to advance t */

if (is_spinpad){
   dt_swp = spin_dp->tspin;  /* kludge for spinpad;  1/spin for MAR */
  /* at most 3 per spin in HAR, so this is safe (only used to malloc*/
  /* sweep pointers, so not too wasteful and copes with mode changes */
  /* Alternative is to detect mode, but that doesn't cope MAR->HAR */
  /* leave here as 1/spin, so plotaz will gather only 1, but modify in */
  /* GetNSweeps so can allocate enough pointers */
} else {
  fwd = 1; full_swp = 1; /* read 1st sweep of spin for isen and advance t*/
  for(i=0; i<nswps_to_try; i++){
    ret_val = read_drec (dkey,lngp->exten, lngp->vsn, idfp, isen,fwd, full_swp);
    set_times_from_idf(&(swpstart[i]), &dummytime, &dummytime,idfp);
  }

  dt_swp = sweep_duration_calculator(swpstart,nswps_to_try);

  if ( dt_swp < 0.0){
     printf("PEACE3D: Error dt_swp < 0.0.\n");
     return PEACE3D_ERROR;
  }

}

/* add 0.1 to spin period to allow for error prior to int truncate */
spin_dp->n_swp = (int)((spin_dp->tspin/dt_swp)+0.1); 

if (DIAG_PRINT_3D){
   for (i=0; i<nswps_to_try; i++){
     printf("Sweep %d start time: ", i);
     print_idfs_time(&(swpstart[i]));
     printf("\n");
   }
   printf("dt_swp = %g; sweeps/spin = %d\n",dt_swp, spin_dp->n_swp);
}   

/* KLUDGE!!! I can't get pitch angle data to behave, so here's a kludge
*  which skips through to the NEXT spin; note that above code makes
*  no attempt to re-position files.
*/
ret_val = advance_idf_file_spin_start(dkey, idfp, lngp, 
                          &(spin_dp->btime), &(spin_dp->etime),num_sensor-1);
/* 
*  advance_idf... returns etime of SWEEP, so calculate spin_etime:
*  add spin period to spin start and put in as spin end
*/
spin_dp->etime = spin_dp->btime;
add_secs_to_idfs_time(&(spin_dp->etime), spin_dp->tspin);
			  
if (ret_val != ALL_OKAY){
   printf("Error advance_idf_file_spin_start in init_spin3d\n");
   return PEACE3D_ERROR;
}

return ALL_OKAY;
}

/*******************************************************************/
float sweep_duration_calculator(IDFS_TIME *swpstart, int n){
/*******************************************************************/
/* use n swpstart times to calculate sweep duration. Do this by
*  removing  min and max values and averaging remainder. This removal
*  is necessary to cope with end of spin jitter
*/

int nmax, i;
float dt[NSWEEPS_4_SWEEPS_IN_SPIN];   /* one more than used */
float dtmin, dtmax, sum;

nmax = NSWEEPS_4_SWEEPS_IN_SPIN;
if ( n > nmax) n=nmax;
if (n<2) return -1.0;
if (n == 2) return (idfs_time_to_epoch_secs(&(swpstart[1])) - 
                       idfs_time_to_epoch_secs(&(swpstart[0])));
for (i=0; i<n-1; i++){
  dt[i] = idfs_time_to_epoch_secs(&(swpstart[i+1])) - 
                       idfs_time_to_epoch_secs(&(swpstart[i]));
}
dtmin = dtmax = dt[0];
for (i=1, sum = dt[0]; i<n-1; i++){
   if (dt[i] < dtmin) dtmin = dt[i];
   if (dt[i] > dtmax) dtmax = dt[i];
   sum  += dt[i];
}
if (n==3) return (dt[0]+dt[1])/2.0;  /* not much else to do! */
sum -= (dtmin + dtmax);
return sum / (float)( n-3);
}
   

/*******************************************************************/
SDDAS_SHORT fill_spin3d (PEACE_SPIN3D *spinp,
                        IDFS_LINEAGE  *lngp,
                        struct idf_data *idfp,
                        SDDAS_ULONG dkey,
                        BIN_UNITS bin_units,
                        PSD_UNITS psd_units,
                        PHI_ZERO_FRAME phi_frame,
			T_OR_F do_pa ){
/*****************************************************************/
/* idfs file has been positioned to start of spin
*  spinp has btime and etime set to start and stop of spin
*  malloc_spin_sweep takes care of:
*
*    n_swp sweep structures and all internal arrays+lineage have been malloced
*    sweep structures have n_sen and n_lev set
*
*  ALL sweeps within spin assumed to have same energy levels
*  ALL sweeps within spin assumed to have same theta bins for sensors
*  ALL sweeps within spin assumed to have same lineage
*
*  fill spin structure by filling sweeps
*  use 1st sweep in spin to fill energy level and theta bins and lineage
*  [all done in fill_sweep]
*
*  Returns one of:

*    ret_val of fill_sweep     normal return
*    PEACE3D_INCOMPLETE_SPIN   if data gap detected via tswp_end > spin_end
*
*  Issues warning if sweep straddles spin end time in case calculation of
*  spin end using spin period is slightly off. For a real data gap > sweep
*  time, the next sweep will still be needed to complete spin and will
*  throw an INCOMPLETE_SPIN error.
*/

SDDAS_SHORT ret_val = ALL_OKAY;
int iswp;
PEACE_SWEEP *swpp;

for (iswp = 0; iswp < spinp->n_swp && ret_val == ALL_OKAY; iswp++){
   swpp = &(spinp->sweep[iswp]); /* local copy of sweep addr */
   swpp->bin_units      = bin_units;
   swpp->psd_units      = psd_units;
   swpp->phi_frame      = phi_frame;
   swpp->missing_value  = QP_DEFAULT_MISSING;
   ret_val = fill_sweep(swpp, lngp, idfp, dkey, iswp, do_pa); 
                                                    /*iswp=0 gets bins*/
   if (ret_val != ALL_OKAY){
      printf("PEACE3D ERROR: Fill sweep returned %d for sweep %d\n",
               ret_val, iswp);
   }
   if ( compare_idfs_time(&spinp->etime, &swpp->etime) == T2_LATER){
      if (compare_idfs_time(&spinp->etime, &swpp->btime) == T1_LATER){
        printf("PEACE3D Warning: sweep %d time straddles end of spin",iswp);
        printf("\nSweep start:");print_idfs_time(&swpp->btime);
        printf("\nSpin start: ");print_idfs_time(&spinp->btime);
        printf("\nSpin end:   ");print_idfs_time(&spinp->etime);
        printf("\nSweep end:  ");print_idfs_time(&swpp->etime);
        printf("\nsun_sen %ld spin_rate %ld\n", idfp->sun_sen, idfp->spin_rate);
        printf("Proceeding ...\n");
      } else {
        printf("PEACE3D ERROR: Sweep %d later than spin end.", iswp);
        printf(" Apparent incomplete spin");
        printf("\nSpin start: ");print_idfs_time(&spinp->btime);
        printf("\nSpin end:   ");print_idfs_time(&spinp->etime);
        printf("\nSweep start:");print_idfs_time(&swpp->btime);
        printf("\nSweep end:  ");print_idfs_time(&swpp->etime);
        printf("\nsun_sen %ld spin_rate %ld\n", idfp->sun_sen, idfp->spin_rate);
        ret_val = PEACE3D_INCOMPLETE_SPIN;
      }
   }
}

return ret_val;

}


#ifdef __cplusplus
}
#endif
