/* plotaz.cc ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  26 November 2000
*  last up-date 16 July 2001
*  30 Oct 2001  Adding Liouville Mapping and 1D cuts
*  March 2002   Implemented zero as missing value.
*
*  Utility routines for generic used, including:
*
*  Setting strings, idfs lineages, etc.
*  Setting conversion tables, getting energy bin limits, ...
*  Opeining and positioning idfs file
*  Manipulating time into format needed by and returned from idfs routines
*
*/


#include <stdio.h>

#include "plotaz.h"


 /*
 * Constructor
 *
 * Creates child widgets of the main PlotAzBox widget
 */

PlotAzBox::PlotAzBox( QWidget *parent, const char *name )
    : QVBox( parent, name )
{
    swp1=NULL;  // initialize to NULL so safe against free
    swp2=NULL;  // these now freed at beginning of RUN loop so swps kept
    swpar=NULL; // for Slice to use
    set_pa_opts_defaults(&paopts1);
    set_pa_opts_defaults(&paopts2);
    
    setMargin(10);
    setSpacing(2);

    QString str;

    QHBox *row1 = new QHBox( this );
    row1->setSpacing( 2 );

    // Create a LineageBox which has comboboxes for the various lineage components
    lineage = new LineageBox( row1 );

    // Column for transfering lineage to sweep boxes
    QVBox *lineage_transfer_col = new QVBox(row1);
    // add some space at top
    new QLabel(" ", lineage_transfer_col);

    // Create two pushbuttons to transfer lineage...
    QPushButton *setprimary = new QPushButton( "Primary Lineage -> ",
                                                lineage_transfer_col );
    // ...and connect the clicked SIGNAL with the SLOT
    connect( setprimary, SIGNAL( clicked() ),
             this, SLOT( slotSetPrimarySweepLineage()  ));
    QHBox *copy_time_box  = new QHBox(lineage_transfer_col);
    QLabel *copy_time_label = new QLabel("Copy Time", copy_time_box);
       copy_time_label->setAlignment(AlignRight | AlignVCenter);
    QPushButton *copy_time = new QPushButton ("|\nV",copy_time_box);
    connect (copy_time, SIGNAL (clicked() ),
             this , SLOT (slotCopyTime() ) );

    QPushButton *setsecondary = new QPushButton( "Secondary Lineage -> ",
                                                lineage_transfer_col );
    connect( setsecondary, SIGNAL( clicked() ),
             this, SLOT( slotSetSecondarySweepLineage()  ));
    // add some space at bottom
    new QLabel(" ", lineage_transfer_col);

    QVBox *sweeps = new QVBox (row1);
       sweeps->setSpacing(2);
    prim_sweep  = new SweepBox(1, Horizontal, sweeps);
      prim_sweep->setFrameStyle(QFrame::Box | QFrame::Sunken);
      prim_sweep->setTitle("Primary Slice - Controls Autoscaling ...");
    sec_sweep  = new SweepBox(1, Horizontal, sweeps);
      sec_sweep->setFrameStyle(QFrame::Box | QFrame::Sunken);
      sec_sweep->setTitle("Secondary Slice - plotted on right");

    QHBox *scale_opts = new QHBox (this);
    scale_opts->setSpacing(2);
    QVBox *scale_labels       = new QVBox (scale_opts);
    	                        scale_labels->setSpacing(2);
    QVBox *scale_units        = new QVBox (scale_opts);
    	                        scale_units->setSpacing(2);
    QVBox *scale_linlog       = new QVBox (scale_opts);
    	                        scale_linlog->setSpacing(2);
    QVBox *scale_auto         = new QVBox (scale_opts);
    	                        scale_auto->setSpacing(2);
    QVBox *scale_minmax_label = new QVBox (scale_opts);
    	                        scale_minmax_label->setSpacing(2);
    QVBox *scale_min          = new QVBox (scale_opts);
    	                        scale_min->setSpacing(2);
    QVBox *scale_max          = new QVBox (scale_opts);
     	                        scale_max->setSpacing(2);

// Energy/velocity options
    new QLabel ("Energy/Velocity Options", scale_labels);
    energy_units = new QComboBox(FALSE,scale_units);
       for (int ibu = 0; ibu < N_BIN_UNITS; ibu++){
          energy_units->insertItem (BIN_UNITS_STRING((BIN_UNITS)ibu), -1);
       }
    energy_linlog = new QComboBox(FALSE,scale_linlog);
       energy_linlog->insertItem("Linear",-1);
       energy_linlog->insertItem("Log",-1);
       energy_linlog->setCurrentItem(1);
   energy_scale = new QComboBox(FALSE,scale_auto);
       energy_scale->insertItem("Autoscale",-1);
       energy_scale->insertItem("Userscale->",-1);
    new QLabel ("min, max", scale_minmax_label);
    emin = new QLineEdit ("1", scale_min);
    emax = new QLineEdit ("10", scale_max);
// PSD Options
    new QLabel ("Phase Space Options", scale_labels);
    psd_units = new QComboBox(FALSE,scale_units);
       for (int ipsu = 0; ipsu < N_PSD_UNITS; ipsu++){
          psd_units->insertItem (PSD_UNITS_STRING((PSD_UNITS)ipsu), -1);
       }
    psd_linlog = new QComboBox(FALSE,scale_linlog);
       psd_linlog->insertItem("Linear",-1);    /* same order as PGSpectro.h */
       psd_linlog->insertItem("Log",-1);
       psd_linlog->setCurrentItem(1);
    psd_scale = new QComboBox(FALSE,scale_auto);
       psd_scale->insertItem("Autoscale",-1); /* same order as PGSpectro.h */
       psd_scale->insertItem("Userscale->",-1);
    new QLabel ("min, max", scale_minmax_label);
    psdmin = new QLineEdit ("1", scale_min);
    psdmax = new QLineEdit ("10", scale_max);
// Analysis Options
    new QLabel("Analysis", scale_labels);
    anal_proc = new QComboBox(FALSE,scale_units);
       anal_proc->insertItem("None",-1);
       anal_proc->insertItem("Add Prim+Sec",-1);
       anal_proc->insertItem("Subtract Prim-Sec",-1);
       anal_proc->insertItem("Multiply Prim*Sec",-1);
       anal_proc->insertItem("Divide Prim/Sec", -1);
    anal_linlog = new QComboBox(FALSE,scale_linlog);
       anal_linlog->insertItem("Linear",-1);    /* same order as PGSpectro.h */
       anal_linlog->insertItem("Log",-1);
       anal_linlog->setCurrentItem(1);
    anal_scale = new QComboBox(FALSE,scale_auto);
       anal_scale->insertItem("Autoscale",-1); /* same order as PGSpectro.h */
       anal_scale->insertItem("Userscale->",-1);
    new QLabel ("min, max", scale_minmax_label);
    analmin = new QLineEdit ("1", scale_min);
    analmax = new QLineEdit ("10", scale_max);

// Plot Options
    QHBox *plot_opts = new QHBox (this);
    plot_opts->setSpacing(2);
    new QLabel ("Plot Options", plot_opts);
    plot_col = new QComboBox(FALSE,plot_opts);
       plot_col->insertItem("Rainbow",-1);   /* same order as PGSpectro.h */
       plot_col->insertItem("Gray White Low",-1);
       plot_col->insertItem("Gray White High", -1);
    plot_dev = new QComboBox(FALSE,plot_opts);
       plot_dev->insertItem("Xwindow",-1);
       plot_dev->insertItem("Landscape GIF File ->",-1);
       plot_dev->insertItem("Portrait GIF File ->",-1);
       plot_dev->insertItem("Landscape PS File ->",-1);
       plot_dev->insertItem("Portrait PS File ->",-1);
    new QLabel ("Xwin Number", plot_opts);
    plot_win_edit = new QLineEdit ("3", plot_opts);
    plot_win_edit->setMaxLength(3); plot_win_edit->setFixedWidth(45);
    new QLabel ("Plot File Stem", plot_opts);
    plot_file_edit = new QLineEdit ("plotaz_plot", plot_opts);

/*  for the future - copied from qjas
// Export Options
    QHBox *exp_opts_box = new QHBox (this);
    exp_opts_box->setSpacing(5);
    new QLabel ("Export to File Options", exp_opts_box);
    exp_opt = new QComboBox(FALSE,exp_opts_box);
       exp_opt->insertItem("Ascii Cluster Exchange",-1);
       exp_opt->insertItem("Ascii Tabular",-1);
       exp_opt->insertItem("CDF", -1);
    new QLabel ("Export File Stem", exp_opts_box);
    exp_file_edit = new QLineEdit ("qjas_data", exp_opts_box);
*/

// Action Buttons
    QHBox *actions = new QHBox (this);
    QPushButton *run = new QPushButton( "Run",actions );
    connect( run,  SIGNAL(clicked()) , this, SLOT (slotRun()) );
    // slicenmap Runs plotaz (to fetch/plot data) AND slice'n'map GUI
    QPushButton *slicenmap = new QPushButton( "Run then Slice'n'Map",actions );
    connect( slicenmap,  SIGNAL(clicked()) , this, SLOT (slotRun()) );
    connect( slicenmap,  SIGNAL(clicked()) , this, SLOT (slotSlice()) );
    QPushButton *help = new QPushButton( "Help", actions);
    connect (help,     SIGNAL(clicked()) , this, SLOT (slotHelp()) );
    quit = new QPushButton( "Quit",actions);

// create help window and slicnmap
    helpbwin = new Helpbwin();
    slicewin = new Slicewin(this);

}


/*
*     Constructors for LineageBox and SweepBox
*/

LineageBox::LineageBox( QWidget *parent, const char *name )
    : QVBox( parent, name )
{
    proj = new  QComboBox (FALSE, this);
    proj->insertStrList (project_strs);
    miss = new  QComboBox (FALSE, this);
    for (int isc = 0; isc < N_SC; isc++){
          miss->insertItem (SC_NAME((SC)isc), -1);
    }
    exp  = new  QComboBox (FALSE, this);
    exp->insertStrList (exp_strs);
    inst = new  QComboBox (FALSE, this);
    inst->insertStrList (inst_strs);
    virins = new  QComboBox (FALSE, this);
    virins->insertStrList (virins_strs);
// more to follow
}

SweepBox::SweepBox (int strips, Orientation orientation,
                          QWidget *parent, const char *name )
     : QGroupBox (strips, orientation, parent, name)
{
QHBox *row_lin = new QHBox (this );
// row_lin->setMaximumHeight(20);      // this works if needed
row_lin->setSpacing(2);
lineage_lbl = new QLabel ("Lineage",row_lin);
project = new QLabel ("CLUSTERII", row_lin);
project->setFrameStyle(QFrame::Box | QFrame::Raised);
mission = new QLabel ("CLUSTER-2", row_lin);
mission->setFrameStyle(QFrame::Box | QFrame::Raised);
exp = new QLabel ("PEACE", row_lin);
exp->setFrameStyle(QFrame::Box | QFrame::Raised);
inst = new QLabel ("PAD", row_lin);
inst->setFrameStyle(QFrame::Box | QFrame::Raised);
virins = new QLabel ("CPPADL", row_lin);
virins->setFrameStyle(QFrame::Box | QFrame::Raised);
pachoice = new QComboBox (FALSE, row_lin);
       pachoice->insertItem("no pa binning", -1);
       pachoice->insertItem("pa: angles", -1);
       pachoice->insertItem("pa: cosines", -1);    
npa = new QLineEdit("13", row_lin);
      QToolTip::add(npa, "Number of pitch angle bins for rebinning");

QHBox *row_time_prompt = new QHBox(this);
new QLabel("Time close to but earlier than spin start:",
                                  row_time_prompt);

QHBox *row_time = new QHBox (this);
row_time->setSpacing(2);
time_lbl   = new QLabel ("yyyy doy hh:mm:ss.sss", row_time);
year = new QLineEdit ("2001", row_time);
year->setMaxLength(4);year->setFixedWidth(60);
doy = new QLineEdit ("090", row_time);
doy->setMaxLength(3);doy->setFixedWidth(45);
hr = new QLineEdit ("18", row_time);
hr->setMaxLength(2);hr->setFixedWidth(30);
min = new QLineEdit ("0", row_time);
min->setMaxLength(2);min->setFixedWidth(30);
sec = new QLineEdit ("12.345", row_time);
sec->setMaxLength(6);sec->setFixedWidth(75);

QHBox *row_slice = new QHBox (this);
row_slice->setSpacing(2);
slice_lab  = new QLabel ("Azimuthal Slice (deg)",row_slice);
azslice = new QLineEdit ("0", row_slice);

}

void SweepBox::SetLineage (LineageBox *lineagebox){

project->setText(lineagebox->proj->currentText());
mission->setText(lineagebox->miss->currentText());
exp->setText(lineagebox->exp->currentText());
inst->setText(lineagebox->inst->currentText());
virins->setText(lineagebox->virins->currentText());
}

/* Constructor and destructor for help window */

Helpbwin::Helpbwin(QWidget *parent, const char *name)
   : QVBox(parent,name){

resize(600,600);
setCaption("plotaz help");

QHBox *buttons = new QHBox(this);
QPushButton *backbutton, *forwardbutton, *topbutton;

backbutton = new QPushButton( "<- Back", buttons, 0 );
forwardbutton = new QPushButton ("Forward ->",buttons, 0);
topbutton = new QPushButton ("Top",buttons,0);

QTextBrowser *helpbrowser = new QTextBrowser(this);

connect( backbutton, SIGNAL( clicked() ), helpbrowser, SLOT( backward() ) );
connect( forwardbutton, SIGNAL( clicked() ), helpbrowser, SLOT( forward() ) );
connect( topbutton, SIGNAL( clicked() ), helpbrowser, SLOT( home() ) );

QString homepath = QString(getenv("QPEACE_HOME")) + "/doc/";

helpbrowser->mimeSourceFactory()->setFilePath(homepath);
helpbrowser->setSource("plotaz.html");

}

Helpbwin::~Helpbwin(void){

hide();
}

/* Constructor and destructor for slice window */

Slicewin::Slicewin(QWidget *parent, const char *name)
   : QDialog(parent,name){

// initialize to NULL so safe to throw away on RUN

for (int i=0; i<N_POSSIBLE_CUT_SOURCES; i++)swposs[i]  = NULL;
for (int i=0; i<N_MAX_CUTS; i++) swpp[i] = NULL;

QVBox * slicevbox = new QVBox(this);
//slicevbox->setFixedSize(800,500);   // helps

slicevbox->setSpacing(2);
slicevbox->setMargin(10);
 
setCaption("PEACE Sweep Shifting, Liouville Mapping and Slicing");

new QLabel("Primary => [Parallel Velocity Shift] => Liouville Mapping => Predicted Secondary",slicevbox);

QHBox *shiftbox = new QHBox(slicevbox);
    shiftbox->setSpacing(2);
    new QLabel ("Add parallel velocity (km/s) to v//(primary)", shiftbox);
    v1 = new QLineEdit ("0", shiftbox);
      QToolTip::add(v1, 
       "Parallel velocity to SHIFT all bins BEFORE mapping (if SHIFTED source selected)");
    new QLabel ("   Add parallel velocity (km/s) to v//(secondary)", shiftbox);
    v2 = new QLineEdit ("0", shiftbox);
      QToolTip::add(v2, 
       "Parallel velocity to SHIFT all bins BEFORE mapping (if SHIFTED source selected)");
    shiftbox->setFrameStyle(QFrame::Box | QFrame::Sunken);

QHBox *mapbox = new QHBox(slicevbox);
    mapbox->setSpacing(2);
    new QLabel ("Primary |B1| (arb units)", mapbox);
    b1 = new QLineEdit ("1", mapbox);
    new QLabel ("  Secondary |B2| (arb units)", mapbox);
    b2 = new QLineEdit ("1", mapbox);
    new QLabel ("  Potential Phi2-Phi1 (eV)", mapbox);
    pot = new QLineEdit ("0.0", mapbox);
    mapbox->setFrameStyle(QFrame::Box | QFrame::Sunken);
    QToolTip::add(mapbox,"NOTE: Liouville Mapping requires f(v) phase space density, not Counts or Differential Energy Flux");

new QLabel("\nSelect Sources for Wheel Plots (if desired)",slicevbox);

QHBox *wheel_select = new QHBox(slicevbox);
    wheel_select->setSpacing(2);
    new QLabel ("Left Half", wheel_select);
    lwheel = new Sources(wheel_select);
    new QLabel ("Right Half", wheel_select);
    rwheel = new Sources(wheel_select);
    wheel_select->setFrameStyle(QFrame::Box | QFrame::Sunken);

new QLabel("\nSelect Sources, Pitch Angles (FLOW direction), and Options for 1D Cuts (if desired)",
                                 slicevbox);

QHBox *cut_select = new QHBox(slicevbox);
    cut_select->setSpacing(2);
    cut_select->setFrameStyle(QFrame::Box | QFrame::Sunken);
    for (int icut = 0; icut < N_MAX_CUTS; icut++){
        cuts[icut] = new Cuts_Selector(icut, cut_select);
	cuts[icut]->setSpacing(2);
    }    

// scaling options cut and pasted from PlotAzBox; would be better
// to put this code once in a set of classes, but I'm in a hurry
QHBox *scale_opts = new QHBox (slicevbox);
    scale_opts->setSpacing(2);
    scale_opts->setMargin(5);
    QVBox *scale_labels       = new QVBox (scale_opts);
    	                        scale_labels->setSpacing(2);
    QVBox *scale_units        = new QVBox (scale_opts);
    	                        scale_units->setSpacing(2);
    QVBox *scale_linlog       = new QVBox (scale_opts);
    	                        scale_linlog->setSpacing(2);
    QVBox *scale_auto         = new QVBox (scale_opts);
    	                        scale_auto->setSpacing(2);
    QVBox *scale_minmax_label = new QVBox (scale_opts);
    	                        scale_minmax_label->setSpacing(2);
    QVBox *scale_min          = new QVBox (scale_opts);
    	                        scale_min->setSpacing(2);
    QVBox *scale_max          = new QVBox (scale_opts);
     	                        scale_max->setSpacing(2);

// Energy/velocity options 
    new QLabel ("Energy/Velocity Options", scale_labels);
    energy_units = new QComboBox(FALSE,scale_units); /* can't change */
    energy_units->insertItem ("As Main Win", -1);
    energy_linlog = new QComboBox(FALSE,scale_linlog);
       energy_linlog->insertItem("Linear",-1);
       energy_linlog->insertItem("Log",-1);
       energy_linlog->setCurrentItem(1);
   energy_scale = new QComboBox(FALSE,scale_auto);
       energy_scale->insertItem("Autoscale",-1);
       energy_scale->insertItem("Userscale->",-1);
    new QLabel ("min, max", scale_minmax_label);
    emin = new QLineEdit ("1", scale_min);
    emax = new QLineEdit ("10", scale_max);
// PSD Options
    new QLabel ("Phase Space Options", scale_labels); /* can't change */
    psd_units = new QComboBox(FALSE,scale_units);
    psd_units->insertItem ("As Main Win", -1);
    psd_linlog = new QComboBox(FALSE,scale_linlog);
       psd_linlog->insertItem("Linear",-1);    /* same order as PGSpectro.h */
       psd_linlog->insertItem("Log",-1);
       psd_linlog->setCurrentItem(1);
    psd_scale = new QComboBox(FALSE,scale_auto);
       psd_scale->insertItem("Autoscale",-1); /* same order as PGSpectro.h */
       psd_scale->insertItem("Userscale->",-1);
    new QLabel ("min, max", scale_minmax_label);
    psdmin = new QLineEdit ("1", scale_min);
    psdmax = new QLineEdit ("10", scale_max);

// Plot Options
    QHBox *plot_opts = new QHBox (slicevbox);
    plot_opts->setSpacing(2);
    new QLabel ("Plot Options", plot_opts);
    plot_col = new QComboBox(FALSE,plot_opts);
       plot_col->insertItem("Rainbow",-1);   /* same order as PGSpectro.h */
       plot_col->insertItem("Gray White Low",-1);
       plot_col->insertItem("Gray White High", -1);
    plot_dev = new QComboBox(FALSE,plot_opts);
       plot_dev->insertItem("Xwindow",-1);
       plot_dev->insertItem("Landscape GIF File ->",-1);
       plot_dev->insertItem("Portrait GIF File ->",-1);
       plot_dev->insertItem("Landscape PS File ->",-1);
       plot_dev->insertItem("Portrait PS File ->",-1);
    new QLabel ("Xwin Number", plot_opts);
    plot_win_edit = new QLineEdit ("3", plot_opts);
    plot_win_edit->setMaxLength(3); plot_win_edit->setFixedWidth(45);
    new QLabel ("Plot File Stem", plot_opts);
    plot_file_edit = new QLineEdit ("slice_map_plot", plot_opts);

/*  for the future - copied from qjas
// Export Options
    QHBox *exp_opts_box = new QHBox (this);
    exp_opts_box->setSpacing(5);
    new QLabel ("Export to File Options", exp_opts_box);
    exp_opt = new QComboBox(FALSE,exp_opts_box);
       exp_opt->insertItem("Ascii Cluster Exchange",-1);
       exp_opt->insertItem("Ascii Tabular",-1);
       exp_opt->insertItem("CDF", -1);
    new QLabel ("Export File Stem", exp_opts_box);
    exp_file_edit = new QLineEdit ("qjas_data", exp_opts_box);
*/

// Action Buttons
    QHBox *actions = new QHBox (slicevbox);
    QPushButton *plotbut = new QPushButton( "Run",actions );
/*    QPushButton *exprtbut = new QPushButton( "Export", actions); */
    QPushButton *help = new QPushButton( "Help", actions);
/*    connect(exprtbut,  SIGNAL(clicked()),  this, SLOT ( slotExp()) ); */
    connect( plotbut,  SIGNAL(clicked()) , this, SLOT (slotRun()) );
    connect (help,     SIGNAL(clicked()) , parent, SLOT (slotHelp()) );
    QPushButton *dismissbut = new QPushButton( "Dismiss",actions);
    connect (dismissbut, SIGNAL(clicked()), this, SLOT (slotDismiss()) );

slicevbox->adjustSize();

}

Slicewin::~Slicewin(void){
// throw away old sweeps - but NOT Primary & Secondary from main plotaz
for (int i = 0; i<N_POSSIBLE_CUT_SOURCES; i++){
   if (i != PRIMARY && i != SECONDARY){
      if (swposs[i] != NULL){
         free_sweep_arrays(swposs[i]); 
         free(swposs[i]);
	 swposs[i]=NULL;
      }
   }
}         

hide();
}


/* Constructor for cut selection combination */
Cuts_Selector::Cuts_Selector(int icut, QWidget *parent, const char *name)
   : QVBox(parent,name){
    srclab = new QLabel(this);
    srclab->setNum(icut);
    source = new Sources(this);
    pangle = new Pangles(this);
    linesty = new Linestys(this);
    colour  = new Colours(this);
}    

/* Constructor for Sources, etc.  pull downs */
Sources::Sources(QWidget *parent, const char *name)
    : QComboBox(FALSE,parent,name){
   this->insertItem("None",0);    // same order as enum !!
   this->insertItem("Primary", -1);
   this->insertItem("Shifted Primary", -1);
   this->insertItem("Mapped Primary", -1);
   this->insertItem("Sh then Map Prim", -1);
   this->insertItem("Secondary", -1);
   this->insertItem("Shift Secondary", -1);
   this->insertItem("Mapped/Secondary",-1);
   this->insertItem("ShMap/ShSec'ry",-1);
   this->insertItem("Mapped-Secondary",-1);
   this->insertItem("ShMap - ShSec'ry",-1);
}

Pangles::Pangles(QWidget *parent, const char *name)
   : QLineEdit("0", parent, name){
}

Linestys::Linestys(QWidget *parent, const char *name)
    : QComboBox(FALSE,parent,name){
/* these are in enum order for cpgsls calls; there is no zero */
   this->insertItem("Not-used",0);  // can't pass zero to cpgsls so must trap!!
   this->insertItem("Solid",1);
   this->insertItem("Dashed", 2);
   this->insertItem("DashDot", 3);
   this->insertItem("Dotted", 4);
   this->insertItem("DashDotDotDot",5);
   this->setCurrentItem(1);
}

Colours::Colours(QWidget *parent, const char *name)
    : QComboBox(FALSE,parent,name){
/* these are in enum order from 0 for cpgsci calls */
   this->insertItem("White",0);   
   this->insertItem("Black", -1);
   this->insertItem("Red", -1);
   this->insertItem("Green", -1);
   this->insertItem("Blue",-1);
   this->insertItem("Cyan",-1);
   this->insertItem("Magenta",-1);
   this->insertItem("Yellow",-1);
   this->insertItem("Orange",-1);
   this->insertItem("YellowGreen",-1);
   this->setCurrentItem(1);
}

void PlotAzBox::slotSetPrimarySweepLineage(void){
prim_sweep->SetLineage(lineage);
}

void PlotAzBox::slotSetSecondarySweepLineage(void){
sec_sweep->SetLineage(lineage);
}

void PlotAzBox::slotCopyTime(void){
sec_sweep->year->setText( prim_sweep->year->text() );
sec_sweep->doy->setText( prim_sweep->doy->text() );
sec_sweep->hr->setText( prim_sweep->hr->text() );
sec_sweep->min->setText( prim_sweep->min->text() );
sec_sweep->sec->setText( prim_sweep->sec->text() );
}

void PlotAzBox::slotRun(void){

// throw away old sweeps
if (swp1 != NULL){free_sweep_arrays(swp1); free(swp1);swp1=NULL;}
if (swp2 != NULL){free_sweep_arrays(swp2); free(swp2);swp2=NULL;}
if (swpar != NULL){free_sweep_arrays(swpar); free(swpar); swpar = NULL;}

transfer_lineage(prim_sweep, &lng1);
transfer_lineage(sec_sweep, &lng2);

int pach = prim_sweep->pachoice->currentItem();
if (pach == NO_BINNING){
   paopts1.do_pa = QPFALSE;
} else {
   paopts1.do_pa = QPTRUE;
   if (pach == PA_ANGLES) paopts1.mu_bins = QPFALSE;
   else paopts1.mu_bins = QPTRUE;
   paopts1.n_pabins = prim_sweep->npa->text().toInt();
}

pach = sec_sweep->pachoice->currentItem();
if (pach == NO_BINNING){
   paopts2.do_pa = QPFALSE;
} else {
   paopts2.do_pa = QPTRUE;
   if (pach == PA_ANGLES) paopts2.mu_bins = QPFALSE;
   else paopts2.mu_bins = QPTRUE;
   paopts2.n_pabins = sec_sweep->npa->text().toInt();
}    
transfer_sweep_time(prim_sweep, &idft1);
transfer_sweep_time(sec_sweep,  &idft2);

az1 = prim_sweep->azslice->text().toFloat();
az2 = sec_sweep->azslice->text().toFloat();

bin_units = (BIN_UNITS)energy_units->currentItem();
psdu = (PSD_UNITS)psd_units->currentItem();

Set_sweep_wheel_options_default(&swoptions);

swoptions.color_table = (PGSpectro_col)plot_col->currentItem();
swoptions.zlinlog     = (PGSpectro_linlog)psd_linlog->currentItem();
swoptions.rlinlog     = (PGSpectro_linlog)energy_linlog->currentItem();
swoptions.zautoscale  = (Yes_or_No)psd_scale->currentItem();
swoptions.rautoscale  = (Yes_or_No)energy_scale->currentItem();
swoptions.zmin        = psdmin->text().toFloat();
swoptions.zmax        = psdmax->text().toFloat();
swoptions.rmin        = emin->text().toFloat();
swoptions.rmax        = emax->text().toFloat();
swoptions.skip_missing = YES;
    // z=0.0 used to be treated as missing by  Sweep_wheel_plot
    
QString pgfile, pgfileexten;
QString pgdev, pgdevstr;

pgfile=plot_file_edit->text(); // get file name; if xwin, reset below

switch(plot_dev->currentItem()){
   case 1:
      pgdev.setLatin1("/GIF",-1);
      pgfileexten.setLatin1(".gif", -1);
      break;
   case 2:
      pgdev.setLatin1("/VGIF",-1);
      pgfileexten.setLatin1(".gif", -1);
      break;
   case 3:
      pgdev.setLatin1("/CPS",-1);
      pgfileexten.setLatin1(".ps", -1);
      break;
   case 4:
      pgdev.setLatin1("/VCPS",-1);
      pgfileexten.setLatin1(".ps", -1);
      break;
   case 0:  default:
      bool ok;
      pgfile = plot_win_edit->text();
      int ixwin = pgfile.toInt( &ok, 10);  // see if text in pgfile ok xwin
      if (!ok || ixwin < 1 ||ixwin > 255){ // if not open default xwin
          pgfile.setLatin1(PLOTAZ_DEFAULT_XWIN_STR, -1);
      }
      pgdev.setLatin1("/Xserv", -1);
      pgfileexten.setLatin1("");
      break;
}
char pg_dev_cstr[200];
pgdevstr = pgfile.append(pgfileexten);
pgdevstr.append(pgdev);
strcpy(pg_dev_cstr, pgdevstr.latin1());
// swoptions.pg_device =  pg_dev_cstr;

if (cpgopen(pg_dev_cstr) <=0){
 printf("Failed to open pgplot device %s successfully\n", pg_dev_cstr);
 printf("If file check write permissions; if Xwindow, check pgxwin server\n");
 return;       // early exit, before getting sweeps and malloc memory
}

// get the two sweeps - routine below mallocs them
// use free_sweep_arrays and then free(swp) to cleanup
swp1=NULL; swp2=NULL; swpar=NULL; /* initialise */
swp1 = PeaceGetAzSlice (&lng1,&idft1,az1,bin_units,psdu, &paopts1);
if (swp1==NULL){
   printf("Failed to get 1st sweep\n");
   cpgclos();
   return;}
if (DIAG_PRINT_AZ){printf("swp1:\n"); print_sweep_structure(swp1);}
swp2 = PeaceGetAzSlice (&lng2,&idft2,az2,bin_units,psdu, &paopts2);
if (swp2==NULL){
   printf("Failed to get 2nd sweep\n");
   if (swp1 != NULL){free_sweep_arrays(swp1); free(swp1);swp1=NULL;}
   cpgclos();
   return;
}

if (anal_proc->currentItem() != NONE)cpgsubp(1,2);
cpgshls(0,0,1,0); /* background white */
cpgshls(1,0,0,0); /* default colour black */
cpgpage();                                    

pg_ret_val = PlotTwoSweeps(swp1,swp2, &swoptions);
QString s;  /* return actual min/max values to line edit boxes */
emin->setText(s.setNum(swoptions.rmin));
emax->setText(s.setNum(swoptions.rmax));
psdmin->setText(s.setNum(swoptions.zmin));
psdmax->setText(s.setNum(swoptions.zmax));

if (anal_proc->currentItem() != NONE){
   swpar = sweep_arithmetic(swp1,swp2,(ARITH_OPER)anal_proc->currentItem());
   if (swpar == NULL){
      printf("Sweep_arithmetic returned NULL pointer\n");
   } else {
       swoptions.zlinlog     = (PGSpectro_linlog)anal_linlog->currentItem();
       swoptions.zautoscale  = (Yes_or_No)anal_scale->currentItem();
       swoptions.zmin        = analmin->text().toFloat();
       swoptions.zmax        = analmax->text().toFloat();
       strcpy(swpar->lineage->project,"Analysis: ");
       strcpy(swpar->lineage->mission,"Primary");
       switch(anal_proc->currentItem()){
          case ADD:
	     strcpy(swpar->lineage->experiment," + "); break;
	  case SUBTRACT:
 	     strcpy(swpar->lineage->experiment," - "); break;
	  case MULTIPLY:
 	     strcpy(swpar->lineage->experiment," x ");
	     swpar->psd_units = UNKNOWN_PSD_UNITS;  /* for wedge label*/
	     break;
	  case DIVIDE:
 	     strcpy(swpar->lineage->experiment," / ");
	     swpar->psd_units = UNKNOWN_PSD_UNITS;  /* for wedge label*/
	     break;
	  default:
 	     strcpy(swpar->lineage->experiment," ??? ");
	     swpar->psd_units = UNKNOWN_PSD_UNITS;  /* for wedge label*/
	     break;
        }
       strcpy(swpar->lineage->instrument,"Secondary");
       strcpy(swpar->lineage->virtual_inst,"");
       cpgpage();
       pg_ret_val = PlotTwoSweeps(swpar,swpar,&swoptions);
       analmin->setText(s.setNum(swoptions.zmin));
       analmax->setText(s.setNum(swoptions.zmax));
   }
        
}

cpgiden();

/* make sure pgplot dev is properly closed */
cpgclos();

}


void PlotAzBox::transfer_lineage(SweepBox *swpbx, IDFS_LINEAGE *lng){

strcpy(lng->project,swpbx->project->text().latin1());
strcpy(lng->mission,swpbx->mission->text().latin1());
strcpy(lng->experiment,swpbx->exp->text().latin1());
strcpy(lng->instrument,swpbx->inst->text().latin1());
strcpy(lng->virtual_inst,swpbx->virins->text().latin1());
strcpy(lng->exten,"");
if (DIAG_PRINT_3D)print_idfs_lineage(lng);

}

void PlotAzBox::transfer_sweep_time(SweepBox *swpbx, IDFS_TIME *idft){

int year, doy, hr, min, isec;
long nano;
float secs;

year = swpbx->year->text().toInt();
doy  = swpbx->doy->text().toInt();
hr   = swpbx->hr->text().toInt();
min  = swpbx->min->text().toInt();
secs = swpbx->sec->text().toFloat();

isec = (int)secs;
nano = (long)(1000000000*(secs-isec));

to_idfs_time(year, doy, hr, min, isec, nano, idft);
if (DIAG_PRINT_3D)print_idfs_time(idft);

}


void PlotAzBox::slotSlice(void){
   slicewin->show();
}

void PlotAzBox::slotHelp(void){
if ( QApplication::desktop()->width() < 600
            || QApplication::desktop()->height() < 600 ){
   helpbwin->showMaximized();
} else {
   helpbwin->show();
}
}

void Slicewin::slotDismiss(void){
   this->hide();
}

void Slicewin::slotRun(void){

PlotAzBox* plazbx = dynamic_cast<PlotAzBox*>(this->parentWidget());

// throw away old sweeps - but NOT Primary & Secondary from main plotaz
for (int i = 0; i<N_POSSIBLE_CUT_SOURCES; i++){
   if (i != PRIMARY && i != SECONDARY){
      if (swposs[i] != NULL){
         free_sweep_arrays(swposs[i]); 
         free(swposs[i]);
	 swposs[i]=NULL;
      }
   }
}         


bprim      = b1->text().toFloat();
bsec       = b2->text().toFloat();
potential  = pot->text().toFloat();
vprim      = v1->text().toFloat();
vsec       = v2->text().toFloat();

scop = &sweep_cut_options;
swop = &sweep_wheel_options;
Set_Sweep_cut_options_default(scop);
Set_sweep_wheel_options_default(swop);
swop->color_table = (PGSpectro_col)plot_col->currentItem();

scop->xlinlog    =  swop->rlinlog    = 
                        (PGSpectro_linlog)energy_linlog->currentItem();
scop->ylinlog    =  swop->zlinlog    = 
                        (PGSpectro_linlog)psd_linlog->currentItem();
scop->xautoscale =  swop->rautoscale = (Yes_or_No)energy_scale->currentItem();
scop->yautoscale =  swop->zautoscale = (Yes_or_No)psd_scale->currentItem();
scop->xmin       =  swop->rmin       = emin->text().toFloat();
scop->xmax       =  swop->rmax       = emax->text().toFloat();
scop->ymin       =  swop->zmin       = psdmin->text().toFloat();
scop->ymax       =  swop->zmax       = psdmax->text().toFloat();

QString pgfile, pgfileexten;
QString pgdev, pgdevstr;

pgfile=plot_file_edit->text(); // get file name; if xwin, reset below

switch(plot_dev->currentItem()){
   case 1:
      pgdev.setLatin1("/GIF",-1);
      pgfileexten.setLatin1(".gif", -1);
      break;
   case 2:
      pgdev.setLatin1("/VGIF",-1);
      pgfileexten.setLatin1(".gif", -1);
      break;
   case 3:
      pgdev.setLatin1("/CPS",-1);
      pgfileexten.setLatin1(".ps", -1);
      break;
   case 4:
      pgdev.setLatin1("/VCPS",-1);
      pgfileexten.setLatin1(".ps", -1);
      break;
   case 0:  default:
      bool ok;
      pgfile = plot_win_edit->text();
      int ixwin = pgfile.toInt( &ok, 10);  // see if text in pgfile ok xwin
      if (!ok || ixwin < 1 ||ixwin > 255){ // if not open default xwin
          pgfile.setLatin1(PLOTAZ_DEFAULT_XWIN_STR, -1);
      }
      pgdev.setLatin1("/Xserv", -1);
      pgfileexten.setLatin1("");
      break;
}
char pg_dev_cstr[200];
pgdevstr = pgfile.append(pgfileexten);
pgdevstr.append(pgdev);
strcpy(pg_dev_cstr, pgdevstr.latin1());
// swoptions.pg_device =  pg_dev_cstr;

if (cpgopen(pg_dev_cstr) <=0){
 printf("Failed to open pgplot device %s successfully\n", pg_dev_cstr);
 printf("If file check write permissions; if Xwindow, check pgxwin server\n");
 return;       // early exit, before getting sweeps and malloc memory
}

for (int i=0; i<N_POSSIBLE_CUT_SOURCES; i++)swpcount[i]=0;
lwhsrc = (SWEEP_SOURCE)lwheel->currentItem();
rwhsrc = (SWEEP_SOURCE)rwheel->currentItem();
if (lwhsrc != NO_SOURCE)swpcount[lwhsrc]++;
if (rwhsrc != NO_SOURCE)swpcount[rwhsrc]++;

nswps = 0;
for (int i=0; i< N_MAX_CUTS; i++){
   SWEEP_SOURCE src_holder;
   if ((src_holder = (SWEEP_SOURCE)cuts[i]->source->currentItem()) 
                                                        != NO_SOURCE){
      cutsrc[nswps]  = src_holder;
      thetas[nswps]  = cuts[i]->pangle->text().toFloat();
      cutlsty[nswps] = (QPGLinestyle)cuts[i]->linesty->currentItem();
      cutcol[nswps]  = (QPGLine_colour)cuts[i]->colour->currentItem();
      swpcount[src_holder]++;
      nswps++;
   }
}

scop->nswps = nswps;
scop->ncuts = 1;  /* ignored by Sweep_n_cuts...*/
scop->linestyles = cutlsty;
scop->colours    = cutcol;

swposs[PRIMARY] = plazbx->swp1;  // get primary and secondary sweeps from main win
swposs[SECONDARY]  = plazbx->swp2;


// make needed sweeps; see if either needs shifting first
if (swpcount[SHIFTED_PRIMARY]>0 
           ||  swpcount[SHIFTED_THEN_MAPPED_PRIMARY]  >0 
           ||  swpcount[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]  >0 
	   ||  swpcount[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY] >0 ){
   // need shifted primary sweep
   swposs[SHIFTED_PRIMARY] = PADvpara_boost(swposs[PRIMARY], vprim, 
               PLOTAZ_SHIFT_THETA_FAC, PLOTAZ_SHIFT_LEVEL_FAC);
   if (swposs[SHIFTED_PRIMARY] == NULL){
      printf("Failed to shift primary sweep\n"); cpgclos(); return;
   }
}
if (swpcount[SHIFTED_SECONDARY]>0 
           ||  swpcount[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]  >0 
	   ||  swpcount[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY] >0 ){
   // need shifted secondary sweep
   swposs[SHIFTED_SECONDARY] = PADvpara_boost(swposs[SECONDARY], vsec, 
            PLOTAZ_SHIFT_THETA_FAC, PLOTAZ_SHIFT_LEVEL_FAC);
   if (swposs[SHIFTED_SECONDARY] == NULL){
      printf("Failed to shift secondary sweep\n"); cpgclos(); return;
   }
}

// now unshifted mapping
char holdstr[500];
if (swpcount[MAPPED_PRIMARY]>0 
           ||  swpcount[MAPPED_OVER_SECONDARY]  >0 
	   ||  swpcount[MAPPED_MINUS_SECONDARY] >0 ){
   // need mapped sweep
   swposs[MAPPED_PRIMARY] = 
              PADLiouvilleMap(swposs[PRIMARY], bprim, bsec, potential,
	      PLOTAZ_LIOUVILLE_UNSHIFT_DOUBLE);
   if (swposs[MAPPED_PRIMARY] == NULL){
      printf("Failed to map sweep\n"); cpgclos(); return;
   }  // see what sweeps derived from mapped are necessary
   if (swpcount[MAPPED_OVER_SECONDARY]> 0){   // need mapped/sec
      swposs[MAPPED_OVER_SECONDARY] = 
         sweep_arithmetic(swposs[MAPPED_PRIMARY], swposs[SECONDARY], DIVIDE);
      if (swposs[MAPPED_OVER_SECONDARY] == NULL){
         printf("Failed to take map/sec sweep\n"); cpgclos(); return;
      }
      sprintf(holdstr, "%s/%s",
           swposs[MAPPED_PRIMARY]->lineage->mission,
           swposs[SECONDARY]->lineage->mission);
      strncpy(swposs[MAPPED_OVER_SECONDARY]->lineage->mission,holdstr,
          LINEAGE_MAX_STR -1);
      sprintf(holdstr,"%s,%s",
        swposs[MAPPED_PRIMARY]->lineage->instrument,
        swposs[SECONDARY]->lineage->instrument);
      strncpy(swposs[MAPPED_OVER_SECONDARY]->lineage->instrument,
           holdstr,LINEAGE_MAX_STR -1 );
       sprintf(holdstr,"%s,%s",
        swposs[MAPPED_PRIMARY]->lineage->virtual_inst,
        swposs[SECONDARY]->lineage->virtual_inst);
       strncpy(swposs[MAPPED_OVER_SECONDARY]->lineage->virtual_inst,
           holdstr,LINEAGE_MAX_STR -1);
      swposs[MAPPED_OVER_SECONDARY]->psd_units = UNKNOWN_PSD_UNITS;
      swposs[MAPPED_OVER_SECONDARY]->etime = swposs[SECONDARY]->btime; 
   }
   if (swpcount[MAPPED_MINUS_SECONDARY] > 0){   // need mapped - sec
      swposs[MAPPED_MINUS_SECONDARY] = 
        sweep_arithmetic(swposs[MAPPED_PRIMARY], swposs[SECONDARY], SUBTRACT);
      if (swposs[MAPPED_MINUS_SECONDARY] == NULL){
         printf("Failed to take map-sec sweep\n"); cpgclos(); return;
      }
      sprintf(holdstr, "%s - %s",
           swposs[MAPPED_PRIMARY]->lineage->mission,
           swposs[SECONDARY]->lineage->mission);
      strncpy(swposs[MAPPED_MINUS_SECONDARY]->lineage->mission,holdstr,
         LINEAGE_MAX_STR -1);
      sprintf(holdstr,"%s,%s",
        swposs[MAPPED_PRIMARY]->lineage->instrument,
        swposs[SECONDARY]->lineage->instrument);
      strncpy(swposs[MAPPED_MINUS_SECONDARY]->lineage->instrument,
           holdstr,LINEAGE_MAX_STR -1);
      sprintf(holdstr,"%s,%s",
        swposs[MAPPED_PRIMARY]->lineage->virtual_inst,
        swposs[SECONDARY]->lineage->virtual_inst);
      strncpy(swposs[MAPPED_MINUS_SECONDARY]->lineage->virtual_inst,
           holdstr,LINEAGE_MAX_STR -1);
      swposs[MAPPED_MINUS_SECONDARY]->etime = swposs[SECONDARY]->btime; 
   }
}

// repeat mapping with shifted pads
if (swpcount[SHIFTED_THEN_MAPPED_PRIMARY]>0 
           ||  swpcount[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]  >0 
	   ||  swpcount[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY] >0 ){
   // need shifted mapped sweep
   swposs[SHIFTED_THEN_MAPPED_PRIMARY] = 
              PADLiouvilleMap(swposs[SHIFTED_PRIMARY], bprim, bsec, potential,
	      PLOTAZ_LIOUVILLE_SHIFTED_DOUBLE);
   if (swposs[SHIFTED_THEN_MAPPED_PRIMARY] == NULL){
      printf("Failed to map sweep\n"); cpgclos(); return;
   } // now see if sweeps derived from shifted_mapped
   if (swpcount[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]> 0){   // need mapped/sec
      swposs[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY] = 
         sweep_arithmetic(swposs[SHIFTED_THEN_MAPPED_PRIMARY], 
	        swposs[SHIFTED_SECONDARY], DIVIDE);
      if (swposs[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY] == NULL){
         printf("Failed to take shifted map/sec sweep\n"); cpgclos(); return;
      }
      sprintf(holdstr, "%s/%s",
           swposs[SHIFTED_THEN_MAPPED_PRIMARY]->lineage->mission,
           swposs[SHIFTED_SECONDARY]->lineage->mission);
      strncpy(swposs[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]->lineage->mission,holdstr,
             LINEAGE_MAX_STR -1 );
      sprintf(holdstr,"%s,%s",
        swposs[SHIFTED_THEN_MAPPED_PRIMARY]->lineage->instrument,
        swposs[SHIFTED_SECONDARY]->lineage->instrument);
      strncpy(swposs[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]->lineage->instrument,
           holdstr,LINEAGE_MAX_STR -1);
       sprintf(holdstr,"%s,%s",
        swposs[SHIFTED_THEN_MAPPED_PRIMARY]->lineage->virtual_inst,
        swposs[SHIFTED_SECONDARY]->lineage->virtual_inst);
       strncpy(swposs[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]->lineage->virtual_inst,
           holdstr,LINEAGE_MAX_STR -1);
     swposs[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]->psd_units = UNKNOWN_PSD_UNITS; 
     swposs[SHIFTED_THEN_MAPPED_OVER_SHIFTED_SECONDARY]->etime = swposs[SHIFTED_SECONDARY]->btime; 
   }
   if (swpcount[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY] > 0){// need mapped - sec
      swposs[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY] = 
        sweep_arithmetic(swposs[SHIFTED_THEN_MAPPED_PRIMARY], 
	    swposs[SHIFTED_SECONDARY], SUBTRACT);
      if (swposs[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY] == NULL){
         printf("Failed to take shifted map-sec sweep\n"); cpgclos(); return;
      }
      sprintf(holdstr, "%s - %s",
           swposs[SHIFTED_THEN_MAPPED_PRIMARY]->lineage->mission,
           swposs[SHIFTED_SECONDARY]->lineage->mission);
      strncpy(swposs[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY]->lineage->mission,
             holdstr, LINEAGE_MAX_STR -1);
      sprintf(holdstr,"%s,%s",
        swposs[SHIFTED_THEN_MAPPED_PRIMARY]->lineage->instrument,
        swposs[SHIFTED_SECONDARY]->lineage->instrument);
      strncpy(swposs[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY]->lineage->instrument,
           holdstr,LINEAGE_MAX_STR -1);
      sprintf(holdstr,"%s,%s",
        swposs[SHIFTED_THEN_MAPPED_PRIMARY]->lineage->virtual_inst,
        swposs[SHIFTED_SECONDARY]->lineage->virtual_inst);
      strncpy(swposs[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY]->lineage->virtual_inst,
           holdstr,LINEAGE_MAX_STR -1);
      swposs[SHIFTED_THEN_MAPPED_MINUS_SHIFTED_SECONDARY]->etime = swposs[SHIFTED_SECONDARY]->btime; 
   }
}
// now should have all the sweeps we need for the plots
// put their pointers into the array
for (int i=0; i<nswps; i++)swpp[i] = swposs[cutsrc[i]];

if ( (lwhsrc != NO_SOURCE || rwhsrc != NO_SOURCE) && nswps > 0)cpgsubp(1,2);

cpgshls(0,0,1,0); /* background white */
cpgshls(1,0,0,0); /* default colour black */

// plot wheels 1st; if only 1 source, plot in both halves
if (lwhsrc != NO_SOURCE || rwhsrc != NO_SOURCE){ // at least 1 exists
   if (lwhsrc == NO_SOURCE)      lwhsrc = rwhsrc;
   else if (rwhsrc == NO_SOURCE) rwhsrc = lwhsrc;
   cpgpage();                                    
   pg_ret_val = PlotTwoSweeps(swposs[lwhsrc], swposs[rwhsrc],swop);
   if (pg_ret_val != PGSPECTRO_OK){
      printf("Slice'n'Map: Wheel plot failed\n");
      cpgclos();
      return;
   }
   QString s;  /* return actual min/max values to line edit boxes */
   emin->setText(s.setNum(swop->rmin));
   emax->setText(s.setNum(swop->rmax));
   psdmin->setText(s.setNum(swop->zmin));
   psdmax->setText(s.setNum(swop->zmax));
}

// now cuts
if (nswps > 0){
   cpgpage();                                    
   pg_ret_val = Sweep_n_cuts_plot(swpp,nswps,thetas, scop);
   if (pg_ret_val != PGSPECTRO_OK){
      printf("Slice'n'Map: Slices plot failed\n");
      cpgclos();
      return;
   }
   QString s;  /* return actual min/max values to line edit boxes */
   emin->setText(s.setNum(scop->xmin));
   emax->setText(s.setNum(scop->xmax));
   psdmin->setText(s.setNum(scop->ymin));
   psdmax->setText(s.setNum(scop->ymax));
}

cpgiden();
cpgclos();


}

